/*

calc_wavelet

description: Calculates an analytical wavelet. Such wavelets may be used
   in a numerical simulation.  This C function is based on the IDL
   procedure "analyze_src".

input:
                                                        Wavelets for which the
   Variable     Explanation                             variable must be set
   --------     -----------                             ----------------------
   type         The type of analytical wavelet          ****
                (e.g., Kelly) 

   time_shift   Time shift that makes the wavelet       all
                causal

   center_freq  Approximate center frequency            Gaussian, Kelly, Ricker,
                                                        Tubman, Gabor,
                                                        Derivative-of-Gabor,
                                                        Berlage, Tsang-Rader,
                                                        Chan-Tsang

   coef_exp_e   Coefficient used to scale the           Gabor, Derivative-of-
                exponent in e^()                        Gabor, Wright, Berlage,
                                                        Tsang-Rader

   exp_t        exponent in t^()--- time ramp           Wright, Berlage

   phase_shift  Phase shift in sinusiod                 Gabor, Derivative-
                                                        of-Gabor, Berlage

   strength     Linear scaling applied to the wavelet   all

output:
   wavelet      The analytical wavelet

required functions: none

references: none

*/

#include <stdio.h>
#include <math.h>
#include <string.h>

void calc_wavelet( long int nt, float dt, char type[], float time_shift, 
		  float center_freq, float coef_exp_e, float exp_t, 
		  float phase_shift, float strength, float *wavelet, 
		  long int *status )

{
	long int i ;

	float w0, t, a ;

	w0 = 2. * 3.14159 * center_freq ;

	phase_shift *= ( 3.14159 / 180. ) ;

	if( strcmp( type, "Gaussian" ) == 0 ) {
		
		for( i = 0 ; i < nt ; ++ i ) {
			t = i * dt - time_shift ;
			a = w0 * t ;
			wavelet[i] = exp( -0.25 * a * a ) ;
		}
	}
	else if( strcmp( type, "Kelly" ) == 0 ) {

		for( i = 0 ; i < nt ; ++ i ) {
			t = i * dt - time_shift ;
			a = w0 * t ;
			wavelet[i] = a * exp( -0.25 * a * a ) ;
		}
	}
	else if( strcmp( type, "Ricker" ) == 0 ) {

		for( i = 0 ; i < nt ; ++ i ) {
			t = i * dt - time_shift ;
			a = w0 * t ;
			wavelet[i] = ( 1. - 0.5 * a * a ) * exp( -0.25 * a * a ) ;
		}
	}
	else if( strcmp( type, "Tubman" ) == 0 ) {

		for( i = 0 ; i < nt ; ++ i ) {
			t = i * dt - time_shift ;
			a = w0 * t ;
			wavelet[i] = a * ( 3. - 0.5 * a * a ) * exp( -0.25 * a * a ) ;
		}
	}
	else if( strcmp( type, "Gabor" ) == 0 ) {

		for( i = 0 ; i < nt ; ++ i ) {
			t = i * dt - time_shift ;
			if( t > 0. ) {
				a = w0 * t / coef_exp_e ;
				wavelet[i] = exp( - a * a ) * cos( w0 * t + phase_shift ) ;
			}
			else
			  wavelet[i] = 0. ;
		}
	}
	else if( strcmp( type, "Derivative-of-Gabor" ) == 0 ) {

		for( i = 0 ; i < nt ; ++ i ) {
			t = i * dt - time_shift ;
			if( t > 0. ) {
				a = w0 * t / coef_exp_e ;
				wavelet[i] = 
				  exp( - a * a ) * 
				    ( sin( w0 * t + phase_shift ) +
				     2. * a * cos( w0 * t + phase_shift ) / coef_exp_e ) ;
			}
			else
			  wavelet[i] = 0. ;
		}
	}
	else if( strcmp( type, "Wright" ) == 0 ) {

		for( i = 0 ; i < nt ; ++ i ) {
			t = i * dt - time_shift ;
			if( t > 0. ) {
				a = t / coef_exp_e ;
				wavelet[i] = pow( t, exp_t ) * exp( - a ) ;
			}
			else
			  wavelet[i] = 0. ;
		}
	}
	else if( strcmp( type, "Berlage" ) == 0 ) {

		for( i = 0 ; i < nt ; ++ i ) {
			t = i * dt - time_shift ;
			if( t > 0. ) {
				a = t / coef_exp_e ;
				wavelet[i] = pow( t, exp_t ) * exp( - a ) * 
				  cos( w0 * t + phase_shift ) ;
			}
			else
			  wavelet[i] = 0. ;
		}
	}
	else if( strcmp( type, "Tsang-Rader" ) == 0 ) {

		for( i = 0 ; i < nt ; ++ i ) {
			t = i * dt - time_shift ;
			if( t > 0. ) {
				a = t / coef_exp_e ;
				wavelet[i] = t * exp( - a ) * sin( w0 * t ) ;
			}
			else
			  wavelet[i] = 0. ;
		}
	}
	else if( strcmp( type, "Chan-Tsang" ) == 0 ) {

		for( i = 0 ; i < nt ; ++ i ) {
			t = i * dt - time_shift ;
			if( t > 0. ) {
				a = w0 * t / sqrt(3.) ;
				wavelet[i] = t * t * exp( - a ) * sin( w0 * t ) ;
			}
			else
			  wavelet[i] = 0. ;
		}
	}
	else {
		fprintf( stderr, "ERROR: Function calc_wavelet.\n" ) ;
		fprintf( stderr, "   The specified wavelet --- %s ---\n", type ) ;
		fprintf( stderr, "   is not recognized.\n" ) ;
		*status = -1 ;
	}

	for( i = 0 ; i < nt ; ++ i )
	  wavelet[i] *= strength ;
}
