#define FLAG_ERROR         -1             /* Flag indicating that an error has occurred. */
#define FLAG_OK             1             /* Flag indicating that an error has not 
                                             occurred. */
#define STR_FLAG_MAX       10             /* Maximum number of characters in a flag,
					     which is a string. */
#define STR_MAX            80             /* Maximum number of characters in a 
					     description, which is a string. A description
					     describes, for example, an em region. */
#define VERTICES_MAX       30             /* Maximum number of vertices in the polygon
					     that outlines an em region. */
#define INDICES_MAX        400            /* Maximum number of indices for a snapshot.
					     To make a snapshot of the field, the value
					     of the field must be determined at regularly
					     spaced grid points in both the r and z 
					     directions. These grid points are specified
					     by their associated indices. */
#define GRID_PTS_MAX       5000           /* Maximum number of grid points in the main 
					     part of the model or in a coaxial cable. */
#define GRID_PTS_MIN       10             /* Minimum number of grid points in either the
					     r or z directions. */
#define AREA_ERR_MAX       1.e-9          /* Maximum relative difference between the total
					     area of the em regions and the total area
					     of the grid. */
#define GRID_DIF_MAX       1.e-4          /* Maximum relative difference between the grid 
					     spacing in the coaxial cable and the main 
					     grid, at the attachment point. */
#define REL_THICK_MAX      0.3            /* Maximum radial thickeness of a thin layer
					     relative to the grid spacing */
#define TIME_STEPS_MIN     50             /* Minimum number of time steps. */
#define SCAN_SAMPLES_MIN   50             /* Minimum number of samples in the scans. */
#define SNAPSHOTS_MIN      3              /* Minimum number of snapshots. */
#define SNAPSHOTS_MAX      25             /* Maximum number of snapshots. */
#define RATIO_MIN          1.e5           /* Minimum ratio between the maximum value
					     the initial value of a wavelet */
#define EPSILON_R_MIN      1.             /* Minimum relative dielectric permittivity. */
#define EPSILON_R_MAX      100.           /* Maximum relative dielectric permittivity. */
#define SIGMA_MIN          0.             /* Minimum electrical conductivity. */
#define SIGMA_MAX          10.            /* Maximum electrical conductivity. */
#define SIGMA_METAL        1.e6           /* Regions with conductivities greater than
                                             this are treated as perfect conductors. */
#define MU_R_MIN           0.97           /* Minimum relative magnetic permeability. */
#define MU_R_MAX           1.03           /* Maximum relative magnetic permeability. */
#define MU_0               1.257e-6       /* Magnetic permeability of free space. */
#define EPSILON_0          8.854e-12      /* Dielectric permittivity of free space. */


typedef struct {                          /* Structure used for a component of the
					     electromagnetic field (H, Ez, or Er). 
					     This structure is used both for the main
					     grid and the coaxial cables. */
        float val ;                       /* Value of the component. */
        float c1 ;                        /* 1st coefficient used to calculate the 
					     component. */
        float c2 ;                        /* 2nd coefficient used to calculate the 
					     component. */
} FIELD ;

typedef struct {                          /* Structure used for computation of the Mur
					     boundary. */
	float b_now ;                     /* Value of Er (at the boundary). */
	float b1_now ;                    /* Value of Er (at one grid point removed
					     from the boundary). */
	float b_prev ;                    /* Value of Er (at the boundary) for the 
					     previous time step. */
	float b1_prev ;                   /* Value of Er (at one grid point removed
					     from the boundary) for the previous
					     time step. */
	float coef_1 ;                    /* Coefficients used to calculate the boundary. */
	float coef_2 ;
	float coef_3 ;
} BOUNDARY ;

typedef struct {                          /* Structure used for computation of the
					     electromagnetic field in a coxial cable. */
	FIELD *hc ;                       /* Structure for H --- its value and two
					     coefficients needed to compute H. */
	FIELD *erc ;                      /* Structure for Er --- its value and two
					     coefficients needed to compute Er. */
	FIELD *ezc ;                      /* Structure for Ez --- its value and two
					     coefficients needed to compute Ez. */
	float *src_coef ;                 /* Coefficents used to apply the source
					     in a coaxial cable. The source is specified
					     as a voltage, which is applied as Er. */
	float v_coef ;                    /* Coefficent used to calculate the voltage
					     in a coaxial cable. */
	float i_coef ;                    /* Coefficent used to calculate the current
					     in a coaxial cable. */
	float impedance ;                 /* Impedance of the coaxial cable. Used to
					     apply the absorbing boundary. */
} CC ;

typedef struct {                          /* Structure with the parameters specifying
					     a coaxial cable. */
	char description[STR_MAX] ;       /* Short description of the coaxial cable. */
	long int nr ;                     /* Number of grid points in r direction. */
	long int nz ;                     /* Number of grid points in z direction. */
	float dr ;                        /* Grid spacing in the r direction. */
	float dz ;                        /* Grid spacing in the z direction. */
	float r[GRID_PTS_MAX] ;           /* Radial distance to grid points for H. The 
					     number of points is nr+2. These points are 
					     in the coordinate system for the main grid. */
	float z[GRID_PTS_MAX] ;           /* Axial distance to grid points for H. The 
					     number of points is nz+2. These points are 
					     in the coordinate system for the main grid. */
	float sigma ;                     /* Electrical conductivity. */
	float mu_r ;                      /* Relative magnetic permeability. */
	float mu ;                        /* Magnetic permeability. */
	float epsilon_r ;                 /* Relative dielectric permittivity. */
	float epsilon ;                   /* Dielectric permittivity. */
	float src_loc_z ;                 /* Location (in the z direction) at which the 
					     source is applied. This location is specified in
					     the coordinate system for the coaxial cable.
					     If the cable does not have a source, then 
					     src_loc_z must be negative. */
	long int src_loc_j ;              /* The index corresponding to src_loc_z. */
	float volt_cur_loc_z ;            /* Location (in the z direction) at which the
                                             voltage and the current are calculated. 
					     This location is specified in the coordinate 
					     system for the coaxial cable. */
	long int volt_cur_loc_j ;         /* The index corresponding to volt_cur_loc_z. */
	float grid_attach_r ;             /* r coordinate for the main grid at which
					     the coaxial cable is attached. This coordinate
					     pertains to the innermost grid point (for Er). */
	long int grid_attach_i ;          /* The index corresonding to grid_attach_r. */
	float grid_attach_z ;             /* z coordinate for the main grid at which
					     the coaxial cable is attached. This coordinate
					     pertains to the innermost grid point (for Er). */
	long int grid_attach_j ;          /* The index corresonding to grid_attach_z. */
	char coax_attach[STR_FLAG_MAX] ;  /* Flag indicating whether the top or the
					      bottom of the coaxial cable is attached to the
					      main grid. The flag may be either "top" or
					      "bottom". */
} COAX ;

typedef struct {                          /* Structure used to describe an em region. */
	char description[STR_MAX] ;       /* Short description of the em region. */
        long int n_vertices ;             /* Number of vertices in the polygon that outlines
					     the em region. */
        float r[VERTICES_MAX] ;           /* r coordinates of the vertices in the polygon. */
        float z[VERTICES_MAX] ;           /* z coordinates of the vertices in the polygon. */
	float sigma ;                     /* Electrical conductivity. */
	float eps_r ;                     /* Relative dielectric permittivity. */
	float epsilon ;                   /* Dielectric permittivity. */
	float mu_r ;                      /* Relative magnetic permeability. */
	float mu ;                        /* Magnetic permeability. */
} EM_REGION ;

typedef struct {                          /* Structure used to hold several 
					     different variables needed for the 
					     simulation. These variables do not
					     fit in the other structures. */
	long int nt ;                     /* Number of time steps. */
	float dt ;                        /* Time step. */
	char src_type[STR_FLAG_MAX] ;     /* Flag indicating the type of source. The flag
					      must be 'e', 'm', or 'c', which mean
					      electric conduction current, magnetic
					      conduction current, or coaxial cable. */
	float src_loc_r ;                 /* r coordinate for the main grid at which the
					      source is applied. This variable is used
					      only if the source type is 'e' or 'm'. */
	float src_loc_z ;                 /* z coordinate for the main grid at which the
					      source is applied. This variable is used
					      only if the source type is 'e' or 'm'. */
	long int src_loc_i ;              /* Index corresponding to src_loc_r. */
	long int src_loc_j ;              /* Index corresponding to src_loc_z. */
	float *source ;                   /* Source wavelet. The size of this vector is
					     nt. */
	float *source_c ;                 /* Additional source wavelet required for
					     a source in a coaxial cable. The size of 
					     this vector is nt. */
	char coax_filename[FILENAME_MAX] ;
                                          /* Name of the file with the information
					     on the coaxial cables. */
	long int n_coax ;                 /* Number of coaxial cables. */
	long int coax_source ;            /* Index of the coaxial cable that is the source. */
	char em_region_filename[FILENAME_MAX] ;
                                          /* Name of the file with the information
                                             on the em regions. */
	long int n_em_region ;            /* Number of em regions. */
	char thin_layer[STR_FLAG_MAX] ;   /* Flag indicating whether there are thin layers
					     (in addition to the em regions). The flag
					     must be either 'yes' or 'no'. */
	char thin_layer_filename[FILENAME_MAX] ;
	                                  /* Name of the file with the information
					     about the thin layers. */
	long int n_thin_layer ;           /* Number of thin layers. */
	char voltage_filename[FILENAME_MAX] ;
                                          /* Name of the file in which the voltages
					     will be printed. */
	char current_filename[FILENAME_MAX] ;
                                          /*  Name of the file in which the currents
					      will be printed. */
} GENERAL ;

typedef struct {                          /* Structure used to describe the main grid and
					     the variables related to the simulation. */
	char image_plane[STR_FLAG_MAX] ;  /* Flag indicating whether there is an image
					     plane. The flag must be either 'yes' or 
					     'no'. */
	char boundary[STR_FLAG_MAX] ;     /* Flag indicating the type of absorbing
					     boundary. The flag must be either 'mur' or
					     'impedance'. */
	char spacing_filename[FILENAME_MAX] ;
	                                   /* Name of the file with information on the 
					       spacing of the grid points. */
	long int nr_space ;                /* Number of distances (in the r direction). */
	float *r_distance ;                /* Distances (in the r direction) at which the
					      radial spacing between the grid points is 
					      specified. The size of this vector is 
					      nr_space. */
	float *r_step_size ;               /* Specified radial spacing between the grid 
					      points. The size of this vector is nr_space. */
	long int nz_space ;                /* Number of distances (in the z  direction). */
	float *z_distance ;                /* Distances (in the z direction) at which the
					      axial spacing between the grid points is
					      specified. The size of this vector is
					      nz_space. */
	float *z_step_size ;               /* Specified axial spacing between the grid
					      points. The size of this vector is nz_space. */
	long int nr ;                      /* Number of grid points in r direction. */
	float r[GRID_PTS_MAX] ;            /* Radial distance to grid points for H. 
					      The number of points is nr+2. */
	float *dr ;                        /* Radial distance between the grid points for 
					      H. The number of points is nr+1. */
	float *pr ;                        /* Radial distance between the grid points for
					      Er. The number of points is nr. */
	long int nz ;                      /* Number of grid points in z direction. */
	float z[GRID_PTS_MAX] ;            /* Axial distance to grid points for H. The
					      number of points is nz+2. */
	float *dz ;                        /* Axial distance between the grid points for
					      H. The number of points is nz+1. */
	float *pz ;                        /* Axial distance between the grid points for
					      Er. The number of points is nz. */
	short int *index_ez ;              /* Indices assigned to the grid points for
					      Ez. Each index indicates the em region 
					      in which the grid point is located. */
	short int *index_er ;              /* Indices assigned to the grid points for
					      Er. Each index indicates the em region
					      in which the grid point is located. */
	short int *index_h ;               /* Indices assigned to the grid points for
					      H. Each index indicates the em region
					      in which the grid point is located. */
} GRID ;

typedef struct {                           /* Structure for the normalization factors. */
	float time ;                       /* Factor to normalize time. */
	float distance ;                   /* Factor to normalize distance. */
	float charge ;                     /* Factor to normalize charge. */
	float mass ;                       /* Factor to normalize mass. */
	float epsilon ;                    /* Factor to normalize dielectric permittivity. */
	float mu ;                         /* Factor to normalize magnetic permeability. */
	float sigma ;                      /* Factor to normalize electrical conductivity. */
	float p ;                          /* Factor to normalize Hertz potential. This
					      factor is not used in fd_bh. */
	float e ;                          /* Factor to normalize electric field intensity. */
	float h ;                          /* Factor to normalize magnetic field intensity. */
	float voltage ;                    /* Factor to normalize voltage. */
	float current ;                    /* Factor to normalize current. */
} NORM ;

typedef struct {                           /* Structure for specifying the scans. */
	char flag[STR_FLAG_MAX] ;          /* Flag indicating whether scans should be 
					      written. The flag must be either 'yes' or
					      'no'. */
	char loc_filename[FILENAME_MAX] ;  /* Name of the file in which the locations of
					      the scans are listed. */
	long int n_scan ;                  /* Number of scans. */
	float time_start ;                 /* Starting time for the scans. */
	float time_step ;                  /* Time step between samples in the scan. This
					      must be an integer multiple of dt. */
	float time_end ;                   /* Ending time for the scans. */
	long int time_index_start ;        /* Index corresponding to time_start. */
	long int time_index_step ;         /* Index corresponding to time_step. */
	long int time_index_end ;          /* Index corresponding to time_end. */
	char filename[FILENAME_MAX] ;      /* Name of the file in which the scans will
					      be written. */
} SCAN ;

typedef struct {                           /* Structure used to describe the location of
					      a scan. */
        float r ;                          /* r coordinate. */
        float z ;                          /* z coordinate. */
	long int i_index ;                 /* Index corresponding to r */
	long int j_index ;                 /* Index corresponding to z */
	char field_comp[STR_FLAG_MAX] ;    /* Component of the electromagnetic field. */
} SCAN_LOC ;

typedef struct {                           /* Structure for specifying the snapshots
					      of the electromagnetic field. */
	char flag[STR_FLAG_MAX] ;          /* Flag indicating whether snapshots should be
					      written. The flag must be either 'yes' or
					      'no'. */
	float time_start ;                 /* Starting time for the snapshots. */
	float time_step ;                  /* Time step between snapshots. */
	float time_end ;                   /* Ending time for the snapshots. */
	float r_start ;                    /* Radial distance to the start of the snapshot
					      (its left edge). */
	float r_step ;                     /* Radial distance between points (pixels) in the 
					      snapshot. */
	float r_end ;                      /* Radial distance to the end of the snapshot
					      (its right edge). */
	float z_start ;                    /* Axial distance to the start of the snapshot
					      (its bottom edge). */
	float z_step ;                     /* Axial distance between points (pixels) in the
					      snapshot. */
	float z_end ;                      /* Axial distance to the end of the snapshot
					      (its top edge). */
	long int time_index_start ;        /* Index corresponding to time_start. */
	long int time_index_step ;         /* Index corresponding to time_step. */
	long int time_index_end ;          /* Index corresponding to time_end. */
	long int i_index[INDICES_MAX] ;    /* Indices of the grid points at which the
					      electomagnetic field is assigned to a pixel
					      in the snapshot. These indices pertain to the
					      r direction. */
	long int n_i ;                     /* Number of indices in i_index. */
	long int j_index[INDICES_MAX] ;    /* Indices of the grid points at which the
					      electomagnetic field is assigned to a pixel
					      in the snapshot. These indices pertain to the
					      z direction. */
	long int n_j ;                     /* Number of indices in j_index. */
	char field_comp[STR_FLAG_MAX] ;    /* Component of the electromagnetic field. */
	char filename[FILENAME_MAX] ;      /* Name of the file in which the snapshots 
					      will be written. */
} SNAP ;

typedef struct {                          /* Structure used to describe a thin layer. */
	char description[STR_MAX] ;       /* Short description of the layer. */
        float inner_radius ;              /* Inner radius of the layer. */
        float radial_thickness ;          /* Thickness of the layer in the radial direction. */
	float top ;                       /* z coordinate of the top edge. */
	float bottom ;                    /* z coordinate of the bottom edge. */
	float sigma ;                     /* Electrical conductivity. */
	float eps_r ;                     /* Relative dielectric permittivity. */
	float epsilon ;                   /* Dielectric permittivity. */
	float mu_r ;                      /* Relative magnetic permeability. */
	float mu ;                        /* Magnetic permeability. */
	long int i_index_h ;              /* i index of the grid points (for H and ER) 
					     that are associated with the thin layer. 
					     (Although there are many grid points
					     associated with the thin layer, they all
					     have the same i index.) */
	long int i_index_ez ;             /* i index of the grid points (for Ez)
					     that are associated with the thin layer. 
					     (Although there are many grid points
					     associated with the thin layer, they all
					     have the same i index.) */
	long int j_index ;                /* j index of the grid point (for H) that is
					     associated with the bottom of the thin layer. */
	long int n_index ;                /* Number of grid points (for H) that are 
					     associated with the thin layer. */
	FIELD *erth ;                     /* Er inside the thin layer. */
	FIELD *hth ;                      /* H associated with the thin layer. */
	float *c3 ;                       /* extra coefficent needed to calculate Hth. */
} THIN_LAYER ;

typedef struct {
	char option[STR_FLAG_MAX] ;        /* Flag indicating how the wavelet is 
					      included in the simulations. The flag
					      must be either 'specified' or 'analytic'.
					      'specified' means that the wavelet is
					      listed is a file. 'analytic'
					      means that an analytical wavelet is 
					      calculated. */
	char filename[FILENAME_MAX] ;      /* Name of the file with the specified wavelet. */
	char type[100] ;                   /* Name of the analytical wavelet. */
	float time_shift ;                 /* Time shift that makes the wavelet causal. */
	float center_freq ;                /* Approximate center frequency. */
	float coef_exp_e ;                 /* Coefficient used to scale the exponent 
					      in e^(). */
	float exp_t ;                      /* Exponent in t^() --- generalized form of a
					      time ramp. */
	float phase_shift ;                /* Phase shift in sinusiod. */
	float strength ;                   /* Scaling applied to the wavelet. */
} WAVELET ;

#define R_VERTICES(i,j)    r_vertices[(i) * ( VERTICES_MAX + 1 ) + (j)]  
#define Z_VERTICES(i,j)    z_vertices[(i) * ( VERTICES_MAX + 1 ) + (j)]

#define H(i,j)             h[(j)*(grid->nr)+(i)]              
#define EZ(i,j)            ez[(j)*(grid->nr+1)+(i)]
#define ER(i,j)            er[(j)*(grid->nr)+(i)]

#define HC(i,j)            cc[k].hc[(j)*coax[k].nr+(i)]             
#define EZC(i,j)           cc[k].ezc[(j)*(coax[k].nr+1)+(i)]
#define ERC(i,j)           cc[k].erc[(j)*coax[k].nr+(i)]

#define INDEX_H(i,j)       grid->index_h[(j)*(grid->nr)+(i)]
#define INDEX_EZ(i,j)      grid->index_ez[(j)*(grid->nr+1)+(i)]
#define INDEX_ER(i,j)      grid->index_er[(j)*(grid->nr)+(i)]

#define MAX(a,b)           ( ( (a) > (b) ) ? (a) : (b) )
#define MIN(a,b)           ( ( (a) < (b) ) ? (a) : (b) )

