/*
  fd_bh

  dates:

     00.10 written by Karl J. Ellefsen
  
  description:

     fd_bh is used to simulate electromagnetic wave propagation for a 
     borehole radar antenna or a dielectric logging tool.

     If the code is compiled on a PC, a slight modification must be made to
     function write_idl_file. The modification is fully described in the function.

  execution:

     fd_bh [-idl_file filename] <infile >& listing &

     -idl_file = optional command line argument. If this argument
         is included, then the borehole model is written in file 
	 "filename", which can be plotted by IDL procedure "check_model". 
	 After writing the file, the program ends.
     infile = file with the input parameters
     listing = file that lists the input parameters, the spacing of 
	the grid points, the regions of the model with different
	electromagnetic properties, the parts constituting the antenna.
        In addition, this file has error messages (if any), warning
	messages, and information on the execution of the simulation.

  output: 

  required functions: 

  references: Ellefsen, K. J., 2001, ?
  
*/

#include <stdio.h>
#include <stdlib.h> 
#include <string.h>
#include <math.h>
#include <float.h>
#include "defs.h"

FILE *file_open2 ( char *name, char *mode ) ;
void *mem_alloc2 ( size_t nbytes, char array_name[] ) ;
long int file_read_float( FILE *fp, float *array, long int n ) ;
long int file_write_float( FILE *fp, float *array, long int n ) ;
long int file_write_long( FILE *fp, long int *array, long int n ) ;
void poly_pts( long int *n, float *x, float *y, float *x0, float *y0, long int *status ) ;
long int argck( int argc, char **argv, char *arg ) ;
long int round_off( double x ) ;
double polygon_area ( float x[], float y[], long int n ) ;

void calc_wavelet( long int nt, float dt, char type[], float time_shift,
		  float center_freq, float coef_exp_e, float exp_t,
		  float phase_shift, float strength, float *wavelet,
		  long int *status ) ;
void propagate( NORM norm, GRID *grid, GENERAL general, COAX *coax,
	       SCAN scan, SCAN_LOC *scan_loc, SNAP snap, EM_REGION *em_region,
	       THIN_LAYER *thin_layer ) ;

static void read_param( GENERAL *general, GRID *grid, NORM *norm, WAVELET *wavelet,
		       SCAN *scan, SNAP *snap, long int *status ) ;
static void write_param( GENERAL general, GRID grid, NORM norm, WAVELET wavelet,
			SCAN scan, SNAP snap ) ;
static void check_param( GENERAL general, GRID grid, NORM norm, WAVELET wavelet,
			SCAN scan, SNAP snap, long int *status ) ;
static void count_space( GRID *grid, long int *status ) ;
static void read_space( GRID *grid ) ;
static void write_space( GRID *grid ) ;
static void count_em_regions( char em_region_filename[], long int *n_em_region, 
			     long int *status) ;
static void read_em_regions( char em_region_filename[], long int n_em_region,
			    EM_REGION *em_region ) ;
static void write_em_regions( long int n_em_region, EM_REGION *em_region ) ;
static void check_em_regions( long int n_em_region, EM_REGION *em_region, GRID *grid,
			     long int *status ) ;
static void count_thin_layers( char thin_layer_filename[], long int *n_thin_layer, 
			     long int *status) ;
static void read_thin_layers( char thin_layer_filename[], long int n_thin_layer,
			    THIN_LAYER *thin_layer ) ;
static void write_thin_layers( long int n_thin_layer, THIN_LAYER *thin_layer ) ;
static void check_thin_layers( long int n_thin_layer, THIN_LAYER *thin_layer, GRID *grid,
			     long int *status ) ;
static void count_coax( char coax_filename[], long int *n_coax, long int *status ) ;
static void read_coax( char coax_filename[], long int n_coax, COAX *coax ) ;
static void write_coax( long int n_coax, COAX *coax ) ;
static void check_coax( GENERAL *general, COAX *coax, GRID *grid, long int *status ) ;
static void count_scan_loc( char loc_filename[], long int *n_scan, long int *status ) ;
static void read_scan_loc( char loc_filename[], long int n_scan, SCAN_LOC *scan_loc ) ;
static void write_scan_loc( long int n_scan, SCAN_LOC *scan_loc ) ;
static void check_scan_loc( long int n_scan, SCAN_LOC *scan_loc, GRID *grid, 
			   long int *status ) ;
static void read_wavelet( char wavelet_filename[], long int nt, float *source, 
			 long int *status ) ;
static void check_wavelet( GENERAL general ) ;
static void calc_extra_wavelet( GENERAL *general, WAVELET wavelet, COAX *coax ) ;
static void calc_grid_pts( GRID *grid, long int *status ) ;
static void check_grid_pts( GENERAL general, GRID *grid, long int *status ) ;
static void calc_em_indices( GRID *grid, long int n_em_region, EM_REGION *em_region,
			      long int *status ) ;
static void find_em_region( float *r, float *z, long int n_em_region, long int *n_vertices,
			   float *r_vertices, float *z_vertices, long int *index,
			   long int *status ) ;
static void calc_tl_indices( GRID *grid, long int n_thin_layer, THIN_LAYER *thin_layer,
			     long int *status ) ;
static void calc_scan_indices( SCAN *scan, SCAN_LOC *scan_loc, GRID *grid, GENERAL general ) ;
static void calc_snap_indices( SNAP *snap, GRID *grid, GENERAL general, long int *status ) ;
static void calc_src_indices( GRID *grid, GENERAL *general, long int *status ) ;
static void calc_coax_grid( long int n_coax, COAX *coax, GRID *grid, long int *status ) ;
static void write_idl_file( char filename[], GRID *grid, GENERAL general, COAX *coax,
			   SCAN scan, SCAN_LOC *scan_loc, SNAP snap,
			   EM_REGION *em_region, THIN_LAYER *thin_layer ) ;
static void calc_norms( NORM *norm ) ;
static void apply_norms( NORM norm, GRID *grid, GENERAL *general, COAX *coax, SCAN *scan,
			SCAN_LOC *scan_loc, SNAP *snap, EM_REGION *em_region,
			THIN_LAYER *thin_layer ) ;


int main( int argc, char *argv[] )

{

	long int i, status ;

	COAX *coax ;

	EM_REGION *em_region ;

	GENERAL general ;

	GRID grid ;

	NORM norm ;

        SCAN scan ;

	SCAN_LOC *scan_loc ;

        SNAP snap ;

	THIN_LAYER *thin_layer ;

        WAVELET wavelet ;

	/*
	  ---------------------------------------------------------------
	  Read, write, and check the input parameters.
	  --------------------------------------------------------------- 
	  */

        read_param ( &general, &grid, &norm, &wavelet, &scan, &snap, &status ) ;

	if ( status == FLAG_ERROR ) exit(EXIT_FAILURE) ;

	write_param ( general, grid, norm, wavelet, scan, snap ) ;

	check_param ( general, grid, norm, wavelet, scan, snap, &status ) ;

	if ( status == FLAG_ERROR ) exit(EXIT_FAILURE) ;

	/* 
	  ---------------------------------------------------------------
	  Count the number of parameters that specify the spacing of the
	  grid points.
	  Read and write the parameters.
	  ---------------------------------------------------------------
	  */

	count_space( &grid, &status ) ;

	if ( status == FLAG_ERROR ) exit(EXIT_FAILURE) ;

	grid.r_distance = (float *) 
	  mem_alloc2( grid.nr_space * sizeof( float ), "r_distance" ) ;

	grid.r_step_size = (float *) 
	  mem_alloc2( grid.nr_space * sizeof( float ), "r_step_size" ) ;

	grid.z_distance = (float *) 
	  mem_alloc2( grid.nz_space * sizeof( float ), "z_distance" ) ;

	grid.z_step_size = (float *) 
	  mem_alloc2( grid.nz_space * sizeof( float ), "z_step_size" ) ;

	read_space ( &grid ) ;

	write_space ( &grid ) ;

	/* 
	  ---------------------------------------------------------------
	  Count the number of regions with different electromagnetic
	  properties.
	  Read, write, and check the em regions.
	  ---------------------------------------------------------------
	  */

	count_em_regions ( general.em_region_filename, &(general.n_em_region), &status ) ;

	if ( status == FLAG_ERROR ) exit(EXIT_FAILURE) ;

	em_region = (EM_REGION *)
	  mem_alloc2( general.n_em_region * sizeof(EM_REGION), "em_region" ) ;

	read_em_regions ( general.em_region_filename, general.n_em_region, em_region ) ;

	write_em_regions ( general.n_em_region, em_region ) ;

	check_em_regions ( general.n_em_region, em_region, &grid, &status ) ;

	if ( status == FLAG_ERROR ) exit(EXIT_FAILURE) ;

	/* 
	  ---------------------------------------------------------------
	  Count the number of thin layers. 
	  Read, write, and check the thin layers.
	  ---------------------------------------------------------------
	  */

	if( general.thin_layer[0] == 'y' ) {

		count_thin_layers ( general.thin_layer_filename, &(general.n_thin_layer), 
				   &status ) ;

		if ( status == FLAG_ERROR ) exit(EXIT_FAILURE) ;

		thin_layer = (THIN_LAYER *)
		  mem_alloc2( general.n_thin_layer * sizeof(THIN_LAYER), "thin_layer" ) ;

		read_thin_layers ( general.thin_layer_filename, general.n_thin_layer, 
				  thin_layer ) ;

		write_thin_layers ( general.n_thin_layer, thin_layer ) ;

		check_thin_layers ( general.n_thin_layer, thin_layer, &grid, &status ) ;

		if ( status == FLAG_ERROR ) exit(EXIT_FAILURE) ;
	}
	else
	  general.n_thin_layer = 0 ;

	/* 
	  ---------------------------------------------------------------
	  Count the number of scan locations.
	  Read, write, and check the scan locations.
	  ---------------------------------------------------------------
	  */
	
	if ( scan.flag[0] == 'y' ) {

		count_scan_loc ( scan.loc_filename, &(scan.n_scan), &status ) ;

		if ( status == FLAG_ERROR ) exit(EXIT_FAILURE) ;

		scan_loc = (SCAN_LOC *) 
		  mem_alloc2( scan.n_scan * sizeof( SCAN_LOC ), "scan_loc" ) ;

		read_scan_loc ( scan.loc_filename, scan.n_scan, scan_loc ) ;

		write_scan_loc ( scan.n_scan, scan_loc ) ;

		check_scan_loc ( scan.n_scan, scan_loc, &grid, &status ) ;

		if ( status == FLAG_ERROR ) exit(EXIT_FAILURE) ;

	}	
	else
	  scan.n_scan = 0 ;

	/* 
	  ---------------------------------------------------------------
	  Count the number of coaxial cables.
	  Read, write, and check the parameters specifying the coaxial 
	  cables. Determine which coaxial cable is the source. 
	  ---------------------------------------------------------------
	  */

	if ( general.src_type[0] == 'c' ) {

		count_coax( general.coax_filename, &(general.n_coax), &status ) ;

		if ( status == FLAG_ERROR ) exit(EXIT_FAILURE) ;

		coax = (COAX *) mem_alloc2( general.n_coax * sizeof( COAX ), "coax" ) ;

		read_coax ( general.coax_filename, general.n_coax, coax ) ;

		write_coax ( general.n_coax, coax ) ;

		check_coax ( &general, coax, &grid, &status ) ;

		if ( status == FLAG_ERROR ) exit(EXIT_FAILURE) ;

	}
	else
	  general.n_coax = 0 ;

	/* 
	  ---------------------------------------------------------------
	  If the source wavelet is stored in a file, read the wavelet.
	  Otherwise, calculate an analytical source wavelet.

	  Check the wavelet.
	  ---------------------------------------------------------------
	  */

	general.source = (float *) mem_alloc2( general.nt * sizeof( float ), "source" ) ;

        if ( wavelet.option[0] == 's' ) 

	  read_wavelet( wavelet.filename, general.nt, general.source, &status ) ;

	else 
	  calc_wavelet( general.nt, general.dt, wavelet.type, wavelet.time_shift, 
		       wavelet.center_freq, wavelet.coef_exp_e, wavelet.exp_t,
		       wavelet.phase_shift, wavelet.strength, general.source, &status ) ;

	if ( status == FLAG_ERROR ) exit(EXIT_FAILURE) ;

	check_wavelet( general ) ;

	/* 
	  ---------------------------------------------------------------
	  If there is a coaxial cable, then calculate the additional
	  source wavelet. 
	  ---------------------------------------------------------------
	  */

	if( general.src_type[0] == 'c' ) {

		general.source_c = (float *) mem_alloc2( general.nt * sizeof( float ), "source_c" ) ;

		calc_extra_wavelet( &general, wavelet, coax ) ;

	}

	/* 
	  ---------------------------------------------------------------
	  Calculate the locations of the grid points.
	  Check the grid points.
	  ---------------------------------------------------------------
	  */

	calc_grid_pts( &grid, &status ) ;

	if ( status == FLAG_ERROR ) exit(EXIT_FAILURE) ;

	check_grid_pts( general, &grid, &status ) ;

	if ( status == FLAG_ERROR ) exit(EXIT_FAILURE) ;

	/* 
	  ---------------------------------------------------------------
	  For each point in the grid, determine which em region it is in.
	  Assign the index (for the em region) to that grid point. 
	  ---------------------------------------------------------------
	  */

	grid.index_ez = (short int *) 
	  mem_alloc2( sizeof(short int) * ( ( grid.nr + 1 ) * grid.nz ), "index_ez" ) ;

	grid.index_er = (short int *) 
	  mem_alloc2( sizeof(short int) * ( grid.nr * ( grid.nz + 1 ) ), "index_er" ) ;

	grid.index_h = (short int *) 
	  mem_alloc2( sizeof(short int) * ( grid.nr * grid.nz ), "index_h" ) ;

	calc_em_indices( &grid, general.n_em_region, em_region, &status ) ;

	if ( status == FLAG_ERROR ) exit(EXIT_FAILURE) ;

	/* 
	  ---------------------------------------------------------------
	  For each thin layer, check that its thickness is small compared
	  to the spacing of the grid points. Count the number of grid 
	  points associated with the thin layer and determine the indices
	  of the grid points.
	  ---------------------------------------------------------------
	  */

	if( general.thin_layer[0] == 'y' ) {

		calc_tl_indices( &grid, general.n_thin_layer, thin_layer, &status ) ;

		if ( status == FLAG_ERROR ) exit(EXIT_FAILURE) ;

	}

	/* 
	  ---------------------------------------------------------------
	  Calculate the time and space (grid) indices for the scans.
	  ---------------------------------------------------------------
	  */

	if ( scan.flag[0] == 'y' ) 
	  calc_scan_indices( &scan, scan_loc, &grid, general ) ;

	/* 
	  ---------------------------------------------------------------
	  Check the location of the snapshots and the requested field
	  component. Calculate the time and space (grid) indices for the 
	  snapshots.
	  ---------------------------------------------------------------
	  */

	if ( snap.flag[0] == 'y' ) {

		calc_snap_indices( &snap, &grid, general, &status ) ;

		if ( status == FLAG_ERROR ) exit(EXIT_FAILURE) ;

	}

	/* 
	  ---------------------------------------------------------------
	  If the source is either an electric conduction current density
	  or a delta gap, check that its location is within the 
	  grid. Calculate the space (grid) indices for its location.
	  ---------------------------------------------------------------
	  */

	if ( general.src_type[0] == 'e' || general.src_type[0] == 'd' ) {

		calc_src_indices( &grid, &general, &status ) ;

		if ( status == FLAG_ERROR ) exit(EXIT_FAILURE) ;

	}

	/* 
	  ---------------------------------------------------------------
	  If there are coaxial cables, then calculate, for each cable, 
	  (1) the locations of the grid points, (2) the space indices for 
	  the leftmost attachment point to the main grid, (3) the space 
	  index for the location of the source (if it is a source cable), 
	  and (4) the space indices for the location at which its voltage 
	  and its current are calculated.

	  Check that, at the attachment point for each cable, the grid 
	  for the cable matches the main grid. 
	  ---------------------------------------------------------------
	  */

	if ( general.src_type[0] == 'c' ) {

		calc_coax_grid( general.n_coax, coax, &grid, &status ) ;

		if ( status == FLAG_ERROR ) exit(EXIT_FAILURE) ;

	}

	/* 
	  ---------------------------------------------------------------
	  Write information about the model to a file. (This file can be
	  read by the IDL procedure "check_model" and then plotted.)
	  ---------------------------------------------------------------
	  */

	if ( ( i = argck( argc, argv, "-idl_file" ) ) != 0 ) {

		if( argc != 3 ) {

			fprintf( stderr, "ERROR: To execute program fd_bh and generate\n" ) ;
			fprintf( stderr, "   a file that can be plotted with IDL,\n" ) ;
			fprintf( stderr, "   the command is:\n" ) ;
			fprintf( stderr, "       fd_bh -idl_file filename <in >& lst &\n" ) ;

			exit(EXIT_FAILURE) ;

		}

		write_idl_file( argv[i+1], &grid, general, coax, scan, scan_loc, snap, 
			       em_region, thin_layer ) ;

		exit(EXIT_SUCCESS) ;
	}

	/* 
	  ---------------------------------------------------------------
	  Calculate and then apply normalizations to the variables.
	  ---------------------------------------------------------------
	  */

	calc_norms( &norm ) ;

	apply_norms( norm, &grid, &general, coax, &scan, scan_loc, &snap,
		    em_region, thin_layer ) ;

	/* 
	  ---------------------------------------------------------------
	  Simulate the propagation of electromagnetic waves.
	  ---------------------------------------------------------------
	  */

	propagate( norm, &grid, general, coax, scan, scan_loc, snap, em_region, thin_layer ) ;

	free( grid.r_distance ) ;
	free( grid.r_step_size ) ;
	free( grid.z_distance ) ;
	free( grid.z_step_size ) ;
	free( grid.index_ez ) ;
	free( grid.index_er ) ;
	free( grid.index_h ) ;
	free( em_region ) ;
	free( general.source ) ;
	if( general.thin_layer[0] == 'y' )
	  free( thin_layer ) ;
	if( scan.flag[0] == 'y' ) 
	  free( scan_loc ) ;
	if ( general.src_type[0] == 'c' ) {
		free( coax ) ;
		free( general.source_c ) ;
	}

	return( 0 ) ;
}

/* Read the parameters from the input file. */

static void read_param ( GENERAL *general, GRID *grid, NORM *norm, WAVELET *wavelet, 
		 SCAN *scan, SNAP *snap, long int *status )
{

	/* read the parameters in the input file */

	*status = FLAG_OK ; 

	if( fscanf( stdin ,"spacing_filename=%s\n", &(grid->spacing_filename[0]) ) != 1 ) {

		fprintf( stderr, "Error in input file --- check \n" ) ;
		fprintf( stderr, "   the filename for the spacings of the grid points.\n" ) ;

		*status = FLAG_ERROR ;

	}

	if( fscanf( stdin ,"image_plane=%s\n", &(grid->image_plane[0]) ) != 1 ) {

		fprintf( stderr, "Error in input file --- check \n" ) ;
		fprintf( stderr, "   parameter for specifying the image plane.\n" ) ;

		*status = FLAG_ERROR ;

	}

	if( fscanf( stdin ,"boundary=%s\n", &(grid->boundary[0]) ) != 1 ) {

		fprintf( stderr, "Error in input file --- check \n" ) ;
		fprintf( stderr, "   parameter for specifying the boundary type.\n" ) ;

		*status = FLAG_ERROR ;

	}

	if( fscanf( stdin ,"source_type=%s\n", &(general->src_type[0]) ) != 1 ) {

		fprintf( stderr, "Error in input file --- check \n" ) ;
		fprintf( stderr, "   parameter for specifying the source type.\n" ) ;

		*status = FLAG_ERROR ;

	}

	if ( general->src_type[0] == 'e' || general->src_type[0] == 'd' ) {

		if( fscanf( stdin ,"source_loc_r=%f  source_loc_z=%f\n", 
			   &(general->src_loc_r), &(general->src_loc_z) ) != 2 ) {

			fprintf( stderr, "Error in input file --- check \n" ) ;
			fprintf( stderr, "   parameters for the source location.\n" ) ;

			*status = FLAG_ERROR ;
		}
	}
	else if ( general->src_type[0] == 'c' ) {
		if( fscanf( stdin ,"coax_filename=%s \n", 
			   &(general->coax_filename[0]) ) != 1 ) {
			fprintf( stderr, "Error in input file --- check \n" ) ;
			fprintf( stderr, "   the filename for the coaxial cable.\n" ) ;

			*status = FLAG_ERROR ;
		}
	}


	if( fscanf( stdin ,"nt=%ld  dt=%f\n", &(general->nt), &(general->dt) ) != 2 ) {

 		fprintf( stderr, "Error in input file --- check\n" ) ;
		fprintf( stderr, "   parameters for time steps and\n" ) ;
		fprintf( stderr, "   time step size.\n" ) ;

		*status = FLAG_ERROR ;

	}

	if( fscanf( stdin ,"time_norm=%f distance_norm=%f\n", 
		   &(norm->time), &(norm->distance) ) != 2 ) {

 		fprintf( stderr, "Error in input file --- check \n" ) ;
		fprintf( stderr, "   parameters for time and distance\n" ) ;
		fprintf( stderr, "   normalizations.\n" ) ;

		*status = FLAG_ERROR ;

	}

	if( fscanf( stdin ,"em_region_filename=%s\n", 
		   &(general->em_region_filename[0]) ) != 1 ) {

		fprintf( stderr, "Error in input file --- check \n" ) ;
		fprintf( stderr, "   filename for em regions.\n" ) ;

		*status = FLAG_ERROR ;

	}

	if( fscanf( stdin ,"thin_layer=%s\n", &(general->thin_layer[0]) ) != 1 ) {

		fprintf( stderr, "Error in input file --- check \n" ) ;
		fprintf( stderr, "   parameter for specifying whether there thin layers.\n" ) ;

		*status = FLAG_ERROR ;

	}

	if( general->thin_layer[0] == 'y' ) 
	  if( fscanf( stdin ,"thin_layer_filename=%s\n", 
		     &(general->thin_layer_filename[0]) ) != 1 ) {

		  fprintf( stderr, "Error in input file --- check \n" ) ;
		  fprintf( stderr, "   the filename for the thin layers. \n" ) ;

		  *status = FLAG_ERROR ;

	  }

        if( fscanf( stdin ,"wavelet_option=%s\n", &(wavelet->option[0])  ) != 1 ) {

                fprintf( stderr, "Error in input file --- check \n" ) ;
                fprintf( stderr, "   parameter for specifying\n" ) ;
                fprintf( stderr, "   the source wavelet.\n" ) ;

                *status = FLAG_ERROR ;

        }

	if( wavelet->option[0] == 's' ) {
		if( fscanf( stdin ,"wavelet_filename=%s \n", &(wavelet->filename[0]) ) != 1 ) {

			fprintf( stderr, "Error in input file --- check \n" ) ;
			fprintf( stderr, "   filename for source wavelet.\n" ) ;

			*status = FLAG_ERROR ;
		}
	}
	else if( wavelet->option[0] == 'a' ) {
		if( fscanf( stdin ,"type=%s time_shift=%f center_freq=%f\n", 
			   &(wavelet->type[0]), &(wavelet->time_shift), 
			   &(wavelet->center_freq) ) != 3 ) {

			fprintf( stderr, "Error in input file --- check \n" ) ;
			fprintf( stderr, "   parameters for temporal \n" ) ;
			fprintf( stderr, "   characteristics of wavelet\n" ) ;

			*status = FLAG_ERROR ;

		}

		if( fscanf( stdin ,"coef_exp_e=%f exp_t=%f phase_shift=%f \n", 
			   &(wavelet->coef_exp_e), &(wavelet->exp_t), 
			   &(wavelet->phase_shift) ) != 3 ) {

			fprintf( stderr, "Error in input file --- check \n" ) ;
			fprintf( stderr, "   parameters for temporal \n" ) ;
			fprintf( stderr, "   characteristics of wavelet\n" ) ;

			*status = FLAG_ERROR ;

		}
	}
	else {

		fprintf( stderr, "Error in input file --- check \n" ) ;
		fprintf( stderr, "   parameter wavelet_option.\n" ) ;

		*status = FLAG_ERROR ;
	}

	if( fscanf( stdin ,"strength=%f\n", &(wavelet->strength) )!= 1 ) {

 		fprintf( stderr, "Error in input file --- check \n" ) ;
		fprintf( stderr, "   parameter for wavelet strength.\n" ) ;

		*status = FLAG_ERROR ;

	}

	if( fscanf( stdin ,"scans=%s \n", &(scan->flag[0]) ) != 1 ) {

 		fprintf( stderr, "Error in input file --- check \n" ) ;
		fprintf( stderr, "   parameter for requesting \n" ) ;
		fprintf( stderr, "   scans \n" ) ;

		*status = FLAG_ERROR ;

	}

	if( scan->flag[0] == 'y' ) {

		if( fscanf( stdin , "time_start=%f time_step=%f time_end=%f\n", 
			   &(scan->time_start), &(scan->time_step), 
			   &(scan->time_end) ) != 3 ) {

			fprintf( stderr, "Error in input file --- check \n" ) ;
			fprintf( stderr, "   parameters for time steps \n" ) ;
			fprintf( stderr, "   of the scans \n" ) ;

			*status = FLAG_ERROR ;

		}

		if( fscanf( stdin , "location_filename=%s\n", &(scan->loc_filename[0]) ) != 1 ) {

			fprintf( stderr, "Error in input file --- check \n" ) ;
			fprintf( stderr, "   filename for scan locations.\n" ) ;

			*status = FLAG_ERROR ;

		}

		if( fscanf( stdin ,"scan_filename=%s\n", &(scan->filename[0]) ) != 1 ) {

			fprintf( stderr, "Error in input file --- check \n" ) ;
			fprintf( stderr, "   filename for scans.\n" ) ;

			*status = FLAG_ERROR ;

		}

	}

	if( fscanf( stdin ,"snapshots=%s\n", &(snap->flag[0])  ) != 1 ) {

 		fprintf( stderr, "Error in input file --- check \n" ) ;
		fprintf( stderr, "   parameter for requesting \n" ) ;
		fprintf( stderr, "   snapshots of the wavefield \n" ) ;

		*status = FLAG_ERROR ;

	}

	if( snap->flag[0] == 'y' ) {

		if( fscanf( stdin , "time_start=%f time_step=%f time_end=%f\n", 
			   &(snap->time_start), &(snap->time_step), 
			   &(snap->time_end) ) != 3 ) {

			fprintf( stderr, "Error in input file --- check \n" ) ;
			fprintf( stderr, "   parameters for time steps \n" ) ;
			fprintf( stderr, "   of the snapshots \n" ) ;

			*status = FLAG_ERROR ;

		}

		if( fscanf( stdin , "r_start=%f r_step=%f r_end=%f\n", 
			   &(snap->r_start), &(snap->r_step), 
			   &(snap->r_end) ) != 3 ) {

			fprintf( stderr, "Error in input file --- check \n" ) ;
			fprintf( stderr, "   parameters for r location \n" ) ;
			fprintf( stderr, "   of snapshot \n" ) ;

			*status = FLAG_ERROR ;

		}


		if( fscanf( stdin , "z_start=%f z_step=%f z_end=%f\n", 
			   &(snap->z_start), &(snap->z_step), 
			   &(snap->z_end) ) != 3 ) {

			fprintf( stderr, "Error in input file --- check \n" ) ;
			fprintf( stderr, "   parameters for z location \n" ) ;
			fprintf( stderr, "   of snapshot \n" ) ;

			*status = FLAG_ERROR ;

		}

		if( fscanf( stdin ,"field_comp=%s \n", &(snap->field_comp[0]) ) != 1 ) {

			fprintf( stderr, "Error in input file --- check \n" ) ;
			fprintf( stderr, "   field component for snapshots \n" ) ;

			*status = FLAG_ERROR ;

		}

		if( fscanf( stdin ,"snap_filename=%s \n", &(snap->filename[0]) ) != 1 ) {

			fprintf( stderr, "Error in input file --- check \n" ) ;
			fprintf( stderr, "   filename for snapshots \n" ) ;

			*status = FLAG_ERROR ;

		}

	}

	if( fscanf( stdin ,"voltage_filename=%s \n", &(general->voltage_filename[0]) ) != 1 ) {

		fprintf( stderr, "Error in input file --- check \n" ) ;
		fprintf( stderr, "   filename for voltage.\n" ) ;

		*status = FLAG_ERROR ;

	}

	if( fscanf( stdin ,"current_filename=%s \n", &(general->current_filename[0]) ) != 1 ) {

		fprintf( stderr, "Error in input file --- check \n" ) ;
		fprintf( stderr, "   the filename for the current. \n" ) ;

		*status = FLAG_ERROR ;

	}

}


/* Write the parameters from the input file. */

static void write_param ( GENERAL general, GRID grid, NORM norm, WAVELET wavelet, 
		  SCAN scan, SNAP snap )

{

	/* print the parameters in the input file */

	fprintf( stdout,"Name of the file with the specifications for\n" ) ;
	fprintf( stdout,"   the spacings of the grid points = %s\n", grid.spacing_filename ) ;

	if ( grid.image_plane[0] == 'y' )
	  fprintf( stdout,"The bottom boundary of the grid is an image plane.\n" ) ;
	else
	  fprintf( stdout,"The bottom boundary of the grid is a transmitting boundary.\n" ) ;

	if ( strcmp( grid.boundary, "mur" ) == 0 ){
		fprintf( stdout,"The transmitting boundaries will be\n" ) ;
		fprintf( stdout,"  computed with Mur's 2nd order\n" ) ;
		fprintf( stdout,"  approximation.\n\n" ) ;
	}
	else if ( strcmp( grid.boundary, "impedance" ) == 0 )
	  fprintf( stdout,"The transmitting boundaries will be computed with an impedance.\n\n" ) ;
	else {
		fprintf( stdout,"The specified boundary is not recognized.\n" ) ;
		fprintf( stdout,"The transmitting boundaries will be computed with an impedance.\n\n" ) ;
	}

	if ( general.src_type[0] == 'd' ) {
		fprintf( stdout, "The source is a delta gap.\n" ) ;
		fprintf( stdout, "Source location (r,z): (%-11.4g,%-11.4g)\n\n",
			general.src_loc_r, general.src_loc_z ) ;
	}
	else if ( general.src_type[0] == 'e' ) {
		fprintf( stdout, "The source is a electric conduction current density.\n" ) ;
		fprintf( stdout, "Source location (r,z): (%-11.4g,%-11.4g)\n\n",
			general.src_loc_r, general.src_loc_z ) ;
	}
	else if ( general.src_type[0] == 'c' ) {
		fprintf( stdout, "The source is a coaxial cable.\n" ) ;
		fprintf( stdout,"Name of the file specifying the coaxial cable = %s\n\n", 
			general.coax_filename ) ;
	}		

	fprintf( stdout,"Number of time steps = %-5ld\n", general.nt ) ;
	fprintf( stdout,"Time step size = %-11.4g\n\n", general.dt ) ;

	fprintf( stdout,"Time normalization = %-11.4g\n", norm.time ) ;
	fprintf( stdout,"Distance normalization = %-11.4g\n\n", 
		norm.distance ) ;

	fprintf( stdout,"Name of file with the electromagnetic regions: %s\n\n", 
		general.em_region_filename ) ;

	if( general.thin_layer[0] == 'y' ) {
		fprintf( stdout,"The model includes thin layers.\n" ) ;
		fprintf( stdout,"Name of file specifying the thin_layers: %s\n\n", 
			general.thin_layer_filename ) ;
	}
	else
	  fprintf( stdout,"The model does not include thin layers.\n\n" ) ;

	if( wavelet.option[0] == 's' ) 
	  fprintf( stdout, "Name of file with source wavelet: %s\n\n", wavelet.filename ) ;
	else if( wavelet.option[0] == 'a' ) {
		fprintf( stdout,"Type of source wavelet = %s\n", wavelet.type ) ;
		fprintf( stdout,"Time shift = %-11.4g\n", wavelet.time_shift ) ;
		fprintf( stdout,"Center frequency = %-11.4g\n", wavelet.center_freq ) ;
		fprintf( stdout,"Coefficient in exponent of e = %-11.4g\n", 
			wavelet.coef_exp_e ) ;
		fprintf( stdout,"Exponent of t = %-11.4g\n", wavelet.exp_t ) ;
		fprintf( stdout,"Phase shift = %-11.4g\n", wavelet.phase_shift ) ;
	}

	fprintf( stdout,"Wavelet strength = %-11.4g\n\n", wavelet.strength ) ;

	if( scan.flag[0] == 'y' ) {

		fprintf( stdout, "A file with scans will be written.\n" ) ;

		fprintf( stdout, "Time at start of scans = %-11.4g\n", 
			scan.time_start ) ;
		fprintf( stdout, "Increment = %-11.4g\n", 
			scan.time_step ) ;
		fprintf( stdout, "Time at end of scans = %-11.4g\n", 
			scan.time_end ) ;
		
 		fprintf( stdout, "Name of file with locations of scans: %s\n",
			scan.loc_filename ) ;

 		fprintf( stdout, "Name of file for scans: %s\n\n",
			scan.filename ) ;

	}
	else
	  fprintf( stdout, "A file with scans will not be written.\n\n" ) ;

	if ( snap.flag[0] == 'y' ) {

		fprintf( stdout, "A file with snapshots of the wavefield will be written.\n" ) ;
		fprintf( stdout, "Time at start of snapshots = %-11.4g\n",
			snap.time_start ) ;
		fprintf( stdout, "Increment = %-11.4g\n", snap.time_step ) ;
		fprintf( stdout, "Time at end of snapshots = %-11.4g\n", 
			snap.time_end ) ;

		fprintf( stdout, "Radial distance at left edge = %-11.4g\n", 
			snap.r_start ) ;
		fprintf( stdout, "Increment = %-11.4g\n", snap.r_step ) ;
		fprintf( stdout, "Radial distance at right edge = %-11.4g\n", 
			snap.r_end ) ;

		fprintf( stdout, "Axial distance at bottom edge = %-11.4g\n", 
			snap.z_start ) ;
		fprintf( stdout, "Increment = %-11.4g\n", snap.z_step ) ;
		fprintf( stdout, "Axial distance at top edge = %-11.4g\n", 
			snap.z_end ) ;

 		fprintf( stdout, "Field components for snapshots: %s\n",
			snap.field_comp ) ;
 		fprintf( stdout, "Name of file for snapshots: %s\n\n",
			snap.filename ) ;

	}
	else
	  fprintf( stdout, "Snapshots of the wavefield will not be printed\n\n" ) ;

	fprintf( stdout, "Name of the file for the voltage: %s\n\n", 
		general.voltage_filename ) ;

	fprintf( stdout, "Name of the file for the current: %s\n\n", 
		general.current_filename ) ;

	fflush( stdout ) ;

}

/* check the input parameters */

static void check_param ( GENERAL general, GRID grid, NORM norm, WAVELET wavelet, 
		  SCAN scan, SNAP snap, long int *status ) 

{

	double nptr ;

	*status = FLAG_OK ;

	/* check normalization constants */

	if ( norm.time <= 0. ) {

		fprintf( stderr, "ERROR: Normalization time is less than or equal to 0.\n" ) ;
		
		*status = FLAG_ERROR ;

	}

	if ( norm.distance <= 0. ) {

		fprintf( stderr, "ERROR: Normalization distance is less than or equal to 0.\n" ) ;
		
		*status = FLAG_ERROR ;

	}

	/* check time steps */

	if ( 0 < general.nt && general.nt < TIME_STEPS_MIN ) 
	  fprintf( stderr, "WARNING: The number of time steps seems too small.\n" ) ;

	if( general.nt <= 0 ) {

		fprintf( stderr, 
			"ERROR: The number of time steps is less or equal to zero\n" ) ;

		*status = FLAG_ERROR ; 

	}

	/* check time step size */

	if( general.dt <= 0. ) {

		fprintf( stderr, "ERROR: The time step is less or equal to zero.\n" ) ;

		*status = FLAG_ERROR ; 

	}

	/* check the wavelet */

	if ( wavelet.strength == 0. ) {

		fprintf( stderr, "ERROR: The source (wavelet) strength is zero.\n" ) ;

		*status = FLAG_ERROR ;

	}

	if ( general.src_type[0] == 'd' && grid.image_plane[0] == 'y' &&
	    wavelet.strength != 0.5 ) {

		fprintf( stderr, "WARNING: Because the source is a delta gap\n" ) ;
		fprintf( stderr, "   and the grid has an image plane, the source\n" ) ;
		fprintf( stderr, "   strength is usually 0.5.\n" ) ;

	}

	/* check that scans or snapshots or both are requested */

	if( !( scan.flag[0] == 'y' || snap.flag[0] == 'y' ) ) {

		fprintf( stderr, "ERROR: Scans, or snapshots or both were not\n" ) ;
		fprintf( stderr, "   requested.\n" ) ;

		*status = FLAG_ERROR ;

	}

	/* check the parameters that specify the scans */

	if ( scan.flag[0] == 'y' ) {

		if( scan.time_start < 0. ) {

			fprintf( stderr, "ERROR: The starting time for the scans must\n" ) ;
			fprintf( stderr, "   be greater than or equal to 0.\n" ) ;

			*status = FLAG_ERROR ;

		}


		if( scan.time_step <= 0. ) {

			fprintf( stderr, "ERROR: The time step for the scans must\n" ) ;
			fprintf( stderr, "   must be positive.\n" ) ;

			*status = FLAG_ERROR ;

		}

		if( scan.time_end <= scan.time_start ) {

			fprintf( stderr, "ERROR: For the scans, the ending time must\n" ) ;
			fprintf( stderr, "   be after the starting time.\n" ) ;

			*status = FLAG_ERROR ;

		}

		if( ( general.nt - 1 ) * general.dt <= scan.time_end  ) {

			fprintf( stderr, "WARNING: The ending time for the scans\n" ) ;
			fprintf( stderr, "   exceeds the computation time.\n" ) ;

		}

		if( ( scan.time_end - scan.time_start ) / scan.time_step < 
		   SCAN_SAMPLES_MIN ) {

			fprintf( stderr, "WARNING: The scans will have few samples.\n" ) ;

		}
		 
	}

	/* check the parameters that specifying the snapshots */

	if ( snap.flag[0] == 'y' ) {

		if( snap.time_start < 0. ) {

			fprintf( stderr, "ERROR: The starting time for the snapshots\n" ) ;
			fprintf( stderr, "  must be greater than or equal to 0.\n" ) ;

			*status = FLAG_ERROR ;

		}

		if( snap.time_step <= 0. ) {

			fprintf( stderr, "ERROR: The time step for the snapshots\n" ) ;
			fprintf( stderr, "  must be positive.\n" ) ;

			*status = FLAG_ERROR ;

		}

		if( snap.time_end <= snap.time_start ) {

			fprintf( stderr, "ERROR: For the snapshots, the ending time must\n" ) ;
			fprintf( stderr, "   be after the starting time.\n" ) ;

			*status = FLAG_ERROR ;

		}

		if( ( general.nt - 1 ) * general.dt < snap.time_end ) {

			fprintf( stderr, "WARNING: The ending time for the snapshots\n" ) ;
			fprintf( stderr, "  exceeds the computation time.\n" ) ;

		}

		if( ( snap.time_end - snap.time_start ) / snap.time_step < 
		   SNAPSHOTS_MIN ) {

			fprintf( stderr, "WARNING: Few snapshots have been requested.\n" ) ;

		}

		if( ( snap.time_end - snap.time_start ) / snap.time_step >
		   SNAPSHOTS_MAX ) {

			fprintf( stderr, "WARNING: Many snapshots have been requested.\n" ) ;

		}

		if( snap.r_start < 0. ) {

			fprintf( stderr, "ERROR: In the radial direction, the start of\n" ) ;
			fprintf( stderr, "   the snapshot (i.e., its left edge) must\n" ) ;
			fprintf( stderr, "   be greater than or equal to 0.\n" ) ;

			*status = FLAG_ERROR ;

		}

		if( snap.r_step <= 0. ) {

			fprintf( stderr, "ERROR: For the snapshot, the sample increment\n" ) ;
			fprintf( stderr, "   in the radial direction must be positive.\n" ) ;

			*status = FLAG_ERROR ;

		}

		if( snap.r_end < snap.r_start ) {

			fprintf( stderr, "ERROR: In the radial direction, the end of\n" ) ;
			fprintf( stderr, "   the snapshot must be to the left of\n" ) ;
			fprintf( stderr, "   the start.\n" ) ;

			*status = FLAG_ERROR ;

		}

		if( snap.z_step <= 0. ) {

			fprintf( stderr, "ERROR: For the snapshot, the sample increment\n" ) ;
			fprintf( stderr, "   in the axial direction must be positive.\n" ) ;

			*status = FLAG_ERROR ;

		}

		if( snap.z_end < snap.z_start ) {

			fprintf( stderr, "ERROR: In the axial direction, the end of\n" ) ;
			fprintf( stderr, "  the snapshot must be to the above the\n" ) ;
			fprintf( stderr, "  start.\n" ) ;

			*status = FLAG_ERROR ;

		}

	}

        fflush( stderr ) ;

}

/* Count the parameters, which specify the spacing of the
   grid points. */

static void count_space( GRID *grid, long int *status ) 

{

	long int i ;

	float distance, step_size ;

	FILE *fp ;

	*status = FLAG_OK ;

	fp = file_open2( grid->spacing_filename, "r" ) ;

	if( fscanf( fp, "nr_space=%ld\n", &(grid->nr_space) ) != 1 ) {

		fprintf( stderr, "Error in file %s \n", grid->spacing_filename ) ;
		fprintf( stderr, "   Check variable nr_space.\n" ) ;

		*status = FLAG_ERROR ;

		goto count_space_end ;
	}

	for ( i = 0 ; i < grid->nr_space ; ++ i ) 
	  if( fscanf( fp, "distance=%f   step_size=%f\n", &distance, &step_size ) != 2 ) {

		  fprintf( stderr, "Error in file %s \n", grid->spacing_filename ) ;
		  fprintf( stderr, "   Check variables distance and step_size\n" ) ;
		  fprintf( stderr, "   for the r direction.\n" ) ;
		  
		  *status = FLAG_ERROR ;

		  goto count_space_end ;

	  }

	if( fscanf( fp, "nz_space=%ld\n", &(grid->nz_space) ) != 1 ) {

		fprintf( stderr, "Error in file %s \n", grid->spacing_filename ) ;
		fprintf( stderr, "   Check variable nz_space.\n" ) ;

		*status = FLAG_ERROR ;

		goto count_space_end ;
	}

	for ( i = 0 ; i < grid->nz_space ; ++ i )
	  if( fscanf( fp, "distance=%f   step_size=%f\n", &distance, &step_size ) != 2 ) {

		  fprintf( stderr, "Error in file %s \n", grid->spacing_filename ) ;
		  fprintf( stderr, "   Check variables distance and step_size\n" ) ;
		  fprintf( stderr, "   for the z direction.\n" ) ;

		  *status = FLAG_ERROR ;

		  goto count_space_end ;
	  }

count_space_end:

	fclose( fp ) ;

	fflush( stderr ) ;

}

/* Read the parameters, which specify the spacing of the grid 
   points.  */

static void read_space( GRID *grid )

{
	long int i ;

	FILE *fp ;

	fp = file_open2( grid->spacing_filename, "r" ) ;

	fscanf( fp, "nr_space=%ld\n", &i ) ;

	for ( i = 0 ; i < grid->nr_space ; ++ i )
	  fscanf( fp, "distance=%f   step_size=%f\n", 
		 &(grid->r_distance[i]), &(grid->r_step_size[i]) ) ;

	fscanf( fp, "nz_space=%ld\n", &i ) ;

	for ( i = 0 ; i < grid->nz_space ; ++ i )
	  fscanf( fp, "distance=%f   step_size=%f\n", 
		 &(grid->z_distance[i]), &(grid->z_step_size[i]) ) ;

	fclose( fp ) ;

}


/* Write the parameters for the spacing of the grid points  */

static void write_space( GRID *grid )

{

	long int i ;

	fprintf( stdout, "\nParameters for the spacing of the grid points\n" ) ;

	fprintf( stdout, "Number of specified spacings for the r direction = %ld\n",
		grid->nr_space ) ;

	for ( i = 0 ; i < grid->nr_space ; ++ i )
	  fprintf( stdout, "  distance=%-11.4g   step_size=%-11.4g\n",
		  grid->r_distance[i], grid->r_step_size[i] ) ;

	fprintf( stdout, "Number of specified spacings for the z direction = %ld\n",
		grid->nz_space ) ;

	for ( i = 0 ; i < grid->nz_space ; ++ i )
	  fprintf( stdout, "  distance=%-11.4g   step_size=%-11.4g\n",
		  grid->z_distance[i], grid->z_step_size[i] ) ;

	fflush( stdout ) ;

}

/* Count the number of regions (defined by their electromagnetic
properties). */

static void count_em_regions ( char em_region_filename[], long int *n_em_region, 
			      long int *status ) 

{

	long int i ;

	EM_REGION a ;

	FILE *fp ;

	fp = file_open2( em_region_filename, "r" ) ; 

	*status = FLAG_OK ; 

	*n_em_region = 0 ;

	while( NULL != fgets( &(a.description[0]), STR_MAX, fp ) ) {

		if( fscanf( fp ,"n_vertices=%ld\n", &(a.n_vertices) ) != 1 ) {

			fprintf( stderr, "ERROR, FILE: %s\n", em_region_filename ) ;
			fprintf( stderr, "   Check variable n_vertices.\n" ) ;

			*status = FLAG_ERROR ;

			goto count_em_regions_end ;
		}

		if( a.n_vertices > VERTICES_MAX ) {

			fprintf( stderr, "Error in file %s \n", em_region_filename ) ;
			fprintf( stderr, "   Variable n_vertices is too large.\n" ) ;
			fprintf( stderr, "   Increase macro VERTICES_MAX.\n" ) ;

			*status = FLAG_ERROR ;

			goto count_em_regions_end ;
		}

		for ( i = 0 ; i < a.n_vertices ; ++ i ) 
		  if ( fscanf( fp ,"r=%f  z=%f\n", &(a.r[i]), &(a.z[i]) ) != 2 ) {

			  fprintf( stderr, "Error in file %s.\n", em_region_filename ) ;
			  fprintf( stderr, "   Check variables r and z.\n" ) ;

			  *status = FLAG_ERROR ;

			  goto count_em_regions_end ;

		  }

		if ( fscanf( fp ,"sigma=%f  eps_r=%f  mu_r=%f\n", 
			    &(a.sigma), &(a.eps_r), &(a.mu_r) ) != 3 ) {

			fprintf( stderr, "Error in file %s.\n", em_region_filename ) ;
			fprintf( stderr, "   Check variables sigma, eps_r, and mu.\n" ) ;

			*status = FLAG_ERROR ;

			goto count_em_regions_end ;

		}

		*n_em_region += 1 ;
	}

count_em_regions_end:

	fclose( fp ) ;

	fflush( stderr ) ;
}


/* For each em region, read the vertices of the polygon that
defines the region perimeter and read the em properties. */

static void read_em_regions ( char em_region_filename[], long int n_em_region, 
		      EM_REGION *em_region ) 

{

	long int i, j ;

	FILE *fp ;

	fp = file_open2( em_region_filename, "r" ) ;

	for( j = 0 ; j < n_em_region ; ++ j ) {

		fgets( &(em_region[j].description[0]), STR_MAX, fp ) ;

                fscanf( fp ,"n_vertices=%ld\n", &(em_region[j].n_vertices) ) ;

                for ( i = 0 ; i < em_region[j].n_vertices ; ++ i )
                  fscanf( fp ,"r=%f  z=%f\n", 
			 &(em_region[j].r[i]), &(em_region[j].z[i]) ) ;

		fscanf( fp ,"sigma=%f  eps_r=%f  mu_r=%f\n", 
		       &(em_region[j].sigma), &(em_region[j].eps_r), 
		       &(em_region[j].mu_r) ) ;
        }

	fclose( fp ) ;

}

/* For each em region, write the vertices of the polygon that
defines the region perimeter and write the em properties. */

static void write_em_regions ( long int n_em_region, EM_REGION *em_region ) 

{
	long int i, j ;

	fprintf( stdout, "\nParameters specifying the regions with different\n" ) ;
	fprintf( stdout, "electromagnetic properties.\n" ) ;

	fprintf( stdout, "Number of regions = %-5ld\n", n_em_region ) ;

	for ( j = 0 ; j < n_em_region ; ++ j ) {

		fprintf( stdout, "Region: %-4ld\n   Description: ", j+1 ) ;

		fputs( &(em_region[j].description[0]), stdout ) ;

		fprintf( stdout, "   Number of vertices: %-4ld\n", 
			em_region[j].n_vertices ) ;

		for ( i = 0 ; i < em_region[j].n_vertices ; ++i ) 
		  fprintf( stdout, "   r = %-11.4g    z = %-11.4g\n",
			  em_region[j].r[i], em_region[j].z[i] ) ;

		fprintf( stdout, "   sigma = %-11.4g   eps_r = %-11.4g   mu_r = %-11.4g\n", 
			em_region[j].sigma, em_region[j].eps_r, em_region[j].mu_r ) ;
	}

	fprintf( stdout ,"\n" ) ;

	fflush( stdout ) ;

}

/* Check the electromagnetic regions. */

static void check_em_regions ( long int n_em_region, EM_REGION *em_region, GRID *grid, 
		       long int *status )

{
	long int i, j ;

	double area, area_sum, area_grid ;

	*status = FLAG_OK ;

	for ( i = 0 ; i < n_em_region ; ++ i ) {

		/* check that the vertices of the polygon
		   (which defines the perimeter of the em region)
		   are within the grid */

		for ( j = 0 ; j < em_region[i].n_vertices ; ++j ) {

			if ( em_region[i].r[j] < grid->r_distance[0] ||
			    grid->r_distance[grid->nr_space-1] < em_region[i].r[j] ) {

				fprintf( stderr, "ERROR: For em region %-4ld,\n", i+1 ) ;
				fprintf( stderr, "  vertex %-4ld is outside the\n", j+1 ) ;
				fprintf( stderr, "  grid (in the r direction).\n" ) ;

				*status = FLAG_ERROR ;
			}

			if ( em_region[i].z[j] < grid->z_distance[0] ||
			    grid->z_distance[grid->nz_space-1] < em_region[i].z[j] ) {

				fprintf( stderr, "ERROR: For em region %-4ld,\n", i+1 ) ;
				fprintf( stderr, "  vertex %-4ld is outside the\n", j+1 ) ;
				fprintf( stderr, "  grid (in the z direction).\n" ) ;

				*status = FLAG_ERROR ;
			}

		}

		/* check the em properties of the em region */
		
		/* If the region is a metal, then its dielectric permittivity
		   and magnetic permeability are not used in the FDTD
		   simulation. Therefore, they do not need to be tested. */

		if( SIGMA_METAL <= em_region[i].sigma ) continue ;

		if( em_region[i].sigma < SIGMA_MIN || 
		   SIGMA_MAX < em_region[i].sigma ) {

			fprintf( stderr, "ERROR: The electrical conductivity\n" ) ;
			fprintf( stderr, "  for em region %-4ld is outside\n", i+1 ) ;
			fprintf( stderr, "  the prescribed range.\n" ) ;

			*status = FLAG_ERROR ;
		}

		if( em_region[i].eps_r < EPSILON_R_MIN || 
		   EPSILON_R_MAX < em_region[i].eps_r ) {

			fprintf( stderr, "ERROR: The relative dielectric permittivity\n" ) ;
			fprintf( stderr, "  for em region %-4ld is outside\n", i+1 ) ;
			fprintf( stderr, "  the prescribed range.\n" ) ;

			*status = FLAG_ERROR ;
		}

		if( em_region[i].mu_r < MU_R_MIN || 
		  MU_R_MAX < em_region[i].mu_r ) {

			fprintf( stderr, "ERROR: The relative magnetic permeability\n" ) ;
			fprintf( stderr, "  for em region %-4ld is outside\n", i+1 ) ;
			fprintf( stderr, "  the prescribed range.\n" ) ;

			*status = FLAG_ERROR ;
		}

	}

	area_sum = 0. ;

	fprintf( stdout, "\n\nEM Region    Area                Total Area\n" ) ;

	for ( i = 0 ; i < n_em_region ; ++ i ) {

		area = polygon_area( em_region[i].r, em_region[i].z, 
				    em_region[i].n_vertices ) ;

		area_sum += area ;

		fprintf( stdout, "%-4ld         %-16.9e    %-16.9e\n", 
			i+1, area, area_sum ) ;

	}
	
	area_grid = 
	  ( grid->r_distance[grid->nr_space-1] - grid->r_distance[0] ) *
	    ( grid->z_distance[grid->nz_space-1] - grid->z_distance[0] ) ;

	fprintf( stdout, "\nArea of grid = %-16.9e\n", area_grid ) ;	

	if( fabs( area_grid - area_sum ) / area_grid > AREA_ERR_MAX ) {

			fprintf( stderr, "ERROR: The total area of the em regions\n" ) ;
			fprintf( stderr, "  does not match the area of the grid.\n" ) ;

			*status = FLAG_ERROR ;
	}

	fflush( stderr ) ;

}

/* Count the number of thin_layers. */ 

static void count_thin_layers ( char thin_layer_filename[], long int *n_thin_layer, 
			      long int *status ) 

{

	long int i ;

	THIN_LAYER a ;

	FILE *fp ;

	fp = file_open2( thin_layer_filename, "r" ) ; 

	*status = FLAG_OK ; 

	*n_thin_layer = 0 ;

	while( NULL != fgets( &(a.description[0]), STR_MAX, fp ) ) {

		if( fscanf( fp ,"inner_radius=%f  radial_thickness=%f\n", 
			   &(a.inner_radius), &(a.radial_thickness) ) != 2 ) {

			fprintf( stderr, "ERROR, FILE: %s\n", thin_layer_filename ) ;
			fprintf( stderr, "   Check variables inner_radius and radial_thickness.\n" ) ;

			*status = FLAG_ERROR ;

			goto count_thin_layers_end ;
		}

		if( fscanf( fp ,"top=%f  bottom=%f\n", 
			   &(a.top), &(a.bottom) ) != 2 ) {

			fprintf( stderr, "ERROR, FILE: %s\n", thin_layer_filename ) ;
			fprintf( stderr, "   Check variables top and bottom.\n" ) ;

			*status = FLAG_ERROR ;

			goto count_thin_layers_end ;
		}


		if ( fscanf( fp ,"sigma=%f  eps_r=%f  mu_r=%f\n", 
			    &(a.sigma), &(a.eps_r), &(a.mu_r) ) != 3 ) {

			fprintf( stderr, "Error in file %s.\n", thin_layer_filename ) ;
			fprintf( stderr, "   Check variables sigma, eps_r, and mu.\n" ) ;

			*status = FLAG_ERROR ;

			goto count_thin_layers_end ;

		}

		*n_thin_layer += 1 ;
	}

count_thin_layers_end:

	fclose( fp ) ;

	fflush( stderr ) ;
}


/* Read the properties of each thin layer. */

static void read_thin_layers ( char thin_layer_filename[], long int n_thin_layer, 
		      THIN_LAYER *thin_layer ) 

{

	long int i, j ;

	FILE *fp ;

	fp = file_open2( thin_layer_filename, "r" ) ;

	for( j = 0 ; j < n_thin_layer ; ++ j ) {

		fgets( &(thin_layer[j].description[0]), STR_MAX, fp ) ;

		fscanf( fp ,"inner_radius=%f  radial_thickness=%f\n",
		       &(thin_layer[j].inner_radius), &(thin_layer[j].radial_thickness) ) ;

		fscanf( fp ,"top=%f  bottom=%f\n",
		       &(thin_layer[j].top), &(thin_layer[j].bottom) ) ;

		fscanf( fp ,"sigma=%f  eps_r=%f  mu_r=%f\n", 
		       &(thin_layer[j].sigma), &(thin_layer[j].eps_r), 
		       &(thin_layer[j].mu_r) ) ;
        }

	fclose( fp ) ;

}

/* Write the properties of each thin layer. */

static void write_thin_layers ( long int n_thin_layer, THIN_LAYER *thin_layer ) 

{
	long int i, j ;

	fprintf( stdout, "\nParameters specifying the thin layers.\n" ) ;

	fprintf( stdout, "Number of thin layers = %-5ld\n", n_thin_layer ) ;

	for ( j = 0 ; j < n_thin_layer ; ++ j ) {

		fprintf( stdout, "Layer: %-4ld\n   Description: ", j+1 ) ;

		fputs( &(thin_layer[j].description[0]), stdout ) ;

		fprintf( stdout, "   Inner Radius: %-11.4g\n", 
			thin_layer[j].inner_radius ) ;

		fprintf( stdout, "   Thickness (in radial direction): %-11.4g\n", 
			thin_layer[j].radial_thickness ) ;

		fprintf( stdout, "   Location of top edge: %-11.4g\n", 
			thin_layer[j].top ) ;

		fprintf( stdout, "   Location of bottom edge: %-11.4g\n", 
			thin_layer[j].bottom ) ;

		fprintf( stdout, "   sigma = %-11.4g   eps_r = %-11.4g   mu_r = %-11.4g\n", 
			thin_layer[j].sigma, thin_layer[j].eps_r, thin_layer[j].mu_r ) ;
	}

	fprintf( stdout ,"\n" ) ;

	fflush( stdout ) ;

}

/* Check the thin layers. */

static void check_thin_layers ( long int n_thin_layer, THIN_LAYER *thin_layer, GRID *grid, 
		       long int *status )

{
	long int i, j ;

	float r_tmp ;

	*status = FLAG_OK ;

	for ( i = 0 ; i < n_thin_layer ; ++ i ) {

		/* Check that the top is above the bottom. */

		if( thin_layer[i].top < thin_layer[i].bottom ) {

			r_tmp = thin_layer[i].top ;

			thin_layer[i].top = thin_layer[i].bottom ;

			thin_layer[i].bottom = r_tmp ;

		}

		/* Check that the thin layer is within the grid. */

		if ( thin_layer[i].inner_radius < grid->r_distance[0] ||
		    grid->r_distance[grid->nr_space-1] < thin_layer[i].inner_radius ) {

			fprintf( stderr, "ERROR: Thin layer %-4ld is outside the\n", i+1 ) ;
			fprintf( stderr, "  grid (in the r direction).\n" ) ;

			*status = FLAG_ERROR ;
		}

		if ( thin_layer[i].bottom < grid->z_distance[0] ||
		    grid->z_distance[grid->nz_space-1] < thin_layer[i].top ) {

			fprintf( stderr, "ERROR: Thin layer %-4ld is outside the\n", i+1 ) ;
			fprintf( stderr, "  grid (in the z direction).\n" ) ;

			*status = FLAG_ERROR ;
		}

		/* Check the em properties. */
		
		if( thin_layer[i].sigma < SIGMA_MIN || 
		   SIGMA_MAX < thin_layer[i].sigma ) {

			fprintf( stderr, "WARNING: The electrical conductivity\n" ) ;
			fprintf( stderr, "  for thin layer %-4ld is outside\n", i+1 ) ;
			fprintf( stderr, "  the prescribed range.\n" ) ;

		}

		if( thin_layer[i].eps_r < EPSILON_R_MIN || 
		   EPSILON_R_MAX < thin_layer[i].eps_r ) {

			fprintf( stderr, "ERROR: The relative dielectric permittivity\n" ) ;
			fprintf( stderr, "  for thin layer %-4ld is outside\n", i+1 ) ;
			fprintf( stderr, "  the prescribed range.\n" ) ;

			*status = FLAG_ERROR ;
		}

		if( thin_layer[i].mu_r < MU_R_MIN || 
		  MU_R_MAX < thin_layer[i].mu_r ) {

			fprintf( stderr, "ERROR: The relative magnetic permeability\n" ) ;
			fprintf( stderr, "  for thin layer %-4ld is outside\n", i+1 ) ;
			fprintf( stderr, "  the prescribed range.\n" ) ;

			*status = FLAG_ERROR ;
		}

	}

	fflush( stderr ) ;

}

/* Count the number of coaxial cables */

static void count_coax ( char coax_filename[], long int *n_coax, long int *status ) 

{

	COAX a ;

	FILE *fp ;

	fp = file_open2( coax_filename, "r" ) ; 

	*status = FLAG_OK ;

	*n_coax = 0 ;

	while( NULL != fgets( &(a.description[0]), STR_MAX, fp ) ) {

		if ( fscanf( fp, "nr=%ld   nz=%ld\n", &(a.nr), &(a.nz) ) != 2 ) {

			fprintf( stderr, "ERROR, FILE: %s\n", coax_filename ) ;
			fprintf( stderr, "   Check variables nr and nz.\n" ) ;

			*status = FLAG_ERROR ;

			goto count_coax_end ;

		}

		if ( fscanf( fp, "dr=%f  dz=%f\n", &(a.dr), &(a.dz) ) != 2 ) {

			fprintf( stderr, "ERROR, FILE: %s\n", coax_filename ) ;
			fprintf( stderr, "   Check variables dr and dz.\n" ) ;

			*status = FLAG_ERROR ;

			goto count_coax_end ;

		}

		if( fscanf( fp, "sigma=%f mu_r=%f epsilon_r=%f\n", 
			   &(a.sigma), &(a.mu_r), &(a.epsilon_r) ) != 3 ) {

			fprintf( stderr, "ERROR, FILE: %s\n", coax_filename ) ;
			fprintf( stderr, "   Check variables sigma, mu_r, and epsilon_r.\n" ) ;

			*status = FLAG_ERROR ;

			goto count_coax_end ;

		}

		if( fscanf( fp, "src_loc_z=%f  volt_cur_loc_z=%f\n", 
			   &(a.src_loc_z), &(a.volt_cur_loc_z) ) != 2 ) {

			fprintf( stderr, "ERROR, FILE: %s\n", coax_filename ) ;
			fprintf( stderr, "   Check variable src_loc_z and volt_cur_loc_z.\n" ) ;

			*status = FLAG_ERROR ;

			goto count_coax_end ;

		}

		if( fscanf( fp, "grid_attach_r=%f grid_attach_z=%f coax_attach=%s\n", 
			   &(a.grid_attach_r), &(a.grid_attach_z), &(a.coax_attach[0]) ) != 3 ) {

			fprintf( stderr, "ERROR, FILE: %s\n", coax_filename ) ;
			fprintf( stderr, "   Check variables grid_attach_r,\n" ) ; 
			fprintf( stderr, "   grid_attach_z, and coax_attach.\n" ) ;

			*status = FLAG_ERROR ;

			goto count_coax_end ;

		}

		*n_coax += 1 ;
	}

count_coax_end:

	fclose( fp ) ;

	fflush( stderr ) ;

}

/* Read the parameters for each coaxial cable. */

static void read_coax ( char coax_filename[], long int n_coax, COAX *coax ) 

{
	long int i ;

	FILE *fp ;

	fp = file_open2( coax_filename, "r" ) ; 

	for( i = 0 ; i < n_coax ; ++ i ) {

		fgets( &(coax[i].description[0]), STR_MAX, fp ) ;

		fscanf( fp, "nr=%ld   nz=%ld\n", &(coax[i].nr), &(coax[i].nz) ) ;

		fscanf( fp, "dr=%f  dz=%f\n", &(coax[i].dr), &(coax[i].dz) ) ;

		fscanf( fp, "sigma=%f mu_r=%f epsilon_r=%f\n",
		       &(coax[i].sigma), &(coax[i].mu_r), &(coax[i].epsilon_r) ) ;

		fscanf( fp, "src_loc_z=%f  volt_cur_loc_z=%f\n",
		       &(coax[i].src_loc_z), &(coax[i].volt_cur_loc_z) ) ;

		fscanf( fp, "grid_attach_r=%f grid_attach_z=%f coax_attach=%s\n",
		       &(coax[i].grid_attach_r), &(coax[i].grid_attach_z), 
		       &(coax[i].coax_attach[0]) ) ;

	}

	fclose( fp ) ;
}

/* Write the parameters for each coaxial cable. */

static void write_coax ( long int n_coax, COAX *coax ) 

{
	long int i ;

	fprintf( stdout, "\nParameters related to the coaxial cables.\n" ) ;
	fprintf( stdout, "Number of cables = %-5ld\n", n_coax ) ;

	for( i = 0 ; i < n_coax ; ++ i ) {

		fprintf( stdout, "\nCoaxial Cable %-5ld\n   Description: ", i+1 ) ;

		fputs( &(coax[i].description[0]), stdout ) ;
		
		fprintf( stdout, "   Number of grid points in the r direction = %-5ld\n", 
			coax[i].nr ) ;
		fprintf( stdout, "   Number of grid points in the z direction = %-5ld\n", 
			coax[i].nz ) ;
		fprintf( stdout, "   Grid spacing in the r direction = %-11.4g\n", 
			coax[i].dr ) ;
		fprintf( stdout, "   Grid spacing in the z direction = %-11.4g\n", 
			coax[i].dz ) ;
		fprintf( stdout, "   Electrical conductivity = %-11.4g\n", coax[i].sigma ) ;
		fprintf( stdout, "   Relative dielectric permittivity = %-11.4g\n",
			coax[i].epsilon_r ) ;
		fprintf( stdout, "   Relative magnetic permeability = %-11.4g\n", 
			coax[i].mu_r ) ;
		if( coax[i].src_loc_z > 0. ) {
			fprintf( stdout, "   In this cable, a source is applied.\n" ) ;
			fprintf( stdout, "   Location of source (in z direction) = %-11.4g\n",
				coax[i].src_loc_z ) ;
		}
		else
		  fprintf( stdout, "   In this cable, a source is not applied.\n" ) ;

		fprintf( stdout, "   Location (in z direction) at which voltage and current are calculated = %-11.4g\n", coax[i].volt_cur_loc_z ) ;

		fprintf( stdout, "   Attachment point of cable to the grid: (%-11.4g,%-11.4g)\n",
			coax[i].grid_attach_r, coax[i].grid_attach_z ) ;
		fprintf( stdout, "   The %s of the cable is attached to the grid.\n",
			coax[i].coax_attach ) ;

	}

}


static void check_coax( GENERAL *general, COAX *coax, GRID *grid, long int *status )

{

	long int k, count_coax_sources ;

	*status = FLAG_OK ;

	count_coax_sources = 0 ;

	for( k = 0 ; k < general->n_coax ; ++k ) {

		/* check the number of grid points */

		if( coax[k].nr <= 0 || coax[k].nz <= 0 ) {

			fprintf( stderr, "ERROR: For coaxial cable %-4ld\n", k+1 ) ;
			fprintf( stderr, "   the number of grid points in either\n" ) ;
			fprintf( stderr, "   the r or the z direction is less than\n" ) ;
			fprintf( stderr, "   or equal to 0.\n" ) ;

			*status = FLAG_ERROR ;
		}

		if( coax[k].nr+2 > GRID_PTS_MAX || coax[k].nz+2 > GRID_PTS_MAX ) {

			fprintf( stderr, "ERROR: For coaxial cable %-4ld\n", k+1 ) ;
			fprintf( stderr, "   the number of grid points in either\n" ) ;
			fprintf( stderr, "   the r or the z direction is greater than\n" ) ;
			fprintf( stderr, "   maximum allowable number.\n" ) ;

			*status = FLAG_ERROR ;
		}

		/* check the grid spacing */

		if( coax[k].dr <= 0. || coax[k].dz <= 0. ) {

			fprintf( stderr, "ERROR: For coaxial cable %-4ld\n", k+1 ) ;
			fprintf( stderr, "   the spacing between the grid points in\n" ) ;
			fprintf( stderr, "   either the r or the z direction is less\n" ) ;
			fprintf( stderr, "   or equal to 0.\n" ) ;

			*status = FLAG_ERROR ;
		}

		/* check the em properties */

		if( coax[k].sigma < SIGMA_MIN || SIGMA_MAX < coax[k].sigma ) {

			fprintf( stderr, "ERROR: The electrical conductivity\n" ) ;
			fprintf( stderr, "  for coaxial cable %-4ld is outside\n", k+1 ) ;
			fprintf( stderr, "  the prescribed range.\n" ) ;

			*status = FLAG_ERROR ;
		}

		if( coax[k].epsilon_r < EPSILON_R_MIN || EPSILON_R_MAX < coax[k].epsilon_r ) {

			fprintf( stderr, "ERROR: The relative dielectric permittivity\n" ) ;
			fprintf( stderr, "  for coaxial cable %-4ld is outside\n", k+1 ) ;
			fprintf( stderr, "  the prescribed range.\n" ) ;

			*status = FLAG_ERROR ;
		}

		if( coax[k].mu_r < MU_R_MIN || MU_R_MAX < coax[k].mu_r ) {

			fprintf( stderr, "ERROR: The relative magnetic permeability\n" ) ;
			fprintf( stderr, "  for coaxial cable %-4ld is outside\n", k+1 ) ;
			fprintf( stderr, "  the prescribed range.\n" ) ;

			*status = FLAG_ERROR ;
		}

		/* If this cable serves as the source, then record the index for the 
		   cable and check that the location (at which the field is applied)
		   is within the cable. */
		
		if( coax[k].src_loc_z > 0. ) {

			count_coax_sources += 1 ;

			general->coax_source = k ;

			if( coax[k].src_loc_z < coax[k].dz/2. ||
			   ( coax[k].nz - 0.5 ) * coax[k].dz < coax[k].src_loc_z ) {

				fprintf( stderr, "ERROR: For coaxial cable %-4ld\n", k+1 ) ;
				fprintf( stderr, "  an E field is applied as a source.\n" ) ;
				fprintf( stderr, "  Its specified location, however, is \n" ) ;
				fprintf( stderr, "  outside the cable.\n" ) ;

				*status = FLAG_ERROR ;
			}

		  }
		
		/* check that the attachment is within the grid */
	
		if( coax[k].grid_attach_r < grid->r_distance[0] || 
		   grid->r_distance[grid->nr_space-1] <= coax[k].grid_attach_r ) {

			fprintf( stderr, "ERROR: For coaxial cable %-4ld\n", k+1 ) ;
			fprintf( stderr, "  its specified attachment point is\n" ) ;
			fprintf( stderr, "  outside the grid (in the r direction).\n" ) ;

			*status = FLAG_ERROR ;
		}

		if( coax[k].grid_attach_z < grid->z_distance[0] || 
		   grid->z_distance[grid->nz_space-1] < coax[k].grid_attach_z ) {

			fprintf( stderr, "ERROR: For coaxial cable %-4ld\n", k+1 ) ;
			fprintf( stderr, "  its specified attachment point is\n" ) ;
			fprintf( stderr, "  outside the grid (in the z direction).\n" ) ;

			*status = FLAG_ERROR ;
		}

		/* check that the attachment for the coaxial cable is either
		   its top or its bottom */

		if( strcmp( coax[k].coax_attach, "bottom" ) != 0  &&
		   strcmp( coax[k].coax_attach, "top" ) != 0 ) {

			fprintf( stderr, "ERROR: For coaxial cable %-4ld\n", k+1 ) ;
			fprintf( stderr, "  its specified attachment is neither\n" ) ;
			fprintf( stderr, "  its top or its bottom.\n" ) ;

			*status = FLAG_ERROR ;
		}
	
	}

	if( count_coax_sources < 1 ) {

		fprintf( stderr, "WARNING: Although the model has %-4ld coaxial\n",
			general->n_coax ) ;
		fprintf( stderr, "   cables, no cables acts as a source.\n" ) ;

		*status = FLAG_ERROR ;

	}

	if( count_coax_sources > 1 ) {

		fprintf( stderr, "ERROR: %-4ld coaxial cables act as sources.\n",
			count_coax_sources ) ;

		*status = FLAG_ERROR ;

	}

	fflush( stderr ) ;
}

/* Count the scan locations. */

static void count_scan_loc ( char loc_filename[], long int *n_scan, long int *status ) 

{

	long int i ;

	SCAN_LOC a ;

	FILE *fp ;

	fp = file_open2( loc_filename, "r" ) ; 

	*status = FLAG_OK ; 

	*n_scan = 0 ;

	while( ( i = fscanf( fp ,"r=%f  z=%f  field_comp=%s\n",
			    &(a.r), &(a.z), a.field_comp ) ) == 3 ) 
	  *n_scan += 1 ;

	if( i != 3 && i != EOF ) {

		fprintf( stderr, "Error in file %s.\n", loc_filename ) ;
		fprintf( stderr, "   Check variables r, z, or field_comp.\n" ) ;
		*status = FLAG_ERROR ;

	}

	fclose( fp ) ;

}

/* Read the scan locations. */

static void read_scan_loc ( char loc_filename[], long int n_scan, SCAN_LOC *scan_loc ) 

{

	long int i ;

	FILE *fp ;

	fp = file_open2( loc_filename, "r" ) ; 

	for ( i = 0 ; i < n_scan ; ++i ) 
	  fscanf( fp ,"r=%f  z=%f  field_comp=%s\n", 
		 &(scan_loc[i].r), &(scan_loc[i].z), scan_loc[i].field_comp ) ;

	fclose( fp ) ;

}

/* Write the locations of the scans. */

static void write_scan_loc ( long int n_scan, SCAN_LOC *scan_loc ) 

{

	long int i ;

	fprintf( stdout, "\nNumber of scans = %-4ld\n", n_scan ) ;

	for ( i = 0 ; i < n_scan ; ++i ) 
	  fprintf( stdout ,"   r = %-11.4g   z = %-11.4g    Field component = %s\n", 
		  scan_loc[i].r, scan_loc[i].z, scan_loc[i].field_comp ) ;

	fprintf( stdout ,"\n" ) ;

	fflush( stdout ) ;

}

/* Check the scan locations. */

static void check_scan_loc ( long int n_scan, SCAN_LOC *scan_loc, GRID *grid, long int *status )

{
	long int i ;

	*status = FLAG_OK ;

	for ( i = 0 ; i < n_scan ; ++ i ) {

		if ( scan_loc[i].r < grid->r_distance[0] ||
		    grid->r_distance[grid->nr_space-1] < scan_loc[i].r ) {

			fprintf( stderr, "ERROR: Function check_scan_loc\n" ) ;
			fprintf( stderr, "  Scan %-4ld is\n", i+1 ) ;
			fprintf( stderr, "  outside the grid\n" ) ;
			fprintf( stderr, "  (in the r direction).\n" ) ;

			*status = FLAG_ERROR ;
		}

		if ( scan_loc[i].z < grid->z_distance[0] ||
		    grid->z_distance[grid->nz_space-1] < scan_loc[i].z ) {

			fprintf( stderr, "ERROR: Function check_scan_loc\n" ) ;
			fprintf( stderr, "  Scan %-4ld is\n", i+1 ) ;
			fprintf( stderr, "  outside the grid\n" ) ;
			fprintf( stderr, "  (in the z direction).\n" ) ;

			*status = FLAG_ERROR ;
		}

		if( strcmp( scan_loc[i].field_comp, "ez" ) != 0 &&
		   strcmp( scan_loc[i].field_comp, "er" ) != 0 &&
		   strcmp( scan_loc[i].field_comp, "h" ) != 0 ) {

			fprintf( stderr, "ERROR: Function check_scan_loc\n" ) ;
			fprintf( stderr, "  Requested component of the\n" ) ;
			fprintf( stderr, "  field for scan %-3ld is\n", i+1 ) ;
			fprintf( stderr, "  unrecognizable.\n" ) ;

			*status = FLAG_ERROR ;

		}
	}

	fflush( stderr ) ;

}

/* read a source wavelet (voltage) from a binary file. */

static void read_wavelet( char wavelet_filename[], long int nt, float *source, long int *status ) 

{

	FILE *fp ;

	*status = FLAG_OK ;

	fp = file_open2( wavelet_filename, "r" ) ; 

	if( nt != file_read_float( fp, source, nt )  ) {

		fprintf( stderr, "ERROR: Function read_wavelet\n" ) ;
		fprintf( stderr, "  Could not read wavelet (source)\n" ) ;
		fprintf( stderr, "  from file %s\n", wavelet_filename ) ;

		*status = FLAG_ERROR ;

	}

	fclose( fp ) ;

}

/* check the wavelet */

static void check_wavelet( GENERAL general ) 

{
	long int i ;

	float max_value, ratio ;

	max_value = 0. ;

	for( i = 0 ; i < general.nt ; ++ i ) 
	  max_value = MAX( max_value, fabs( general.source[i] ) ) ;

	if( general.source[0] != 0. ) {

		ratio = max_value / fabs( general.source[0] ) ;

		fprintf( stdout, "\nSource Wavelet\n" ) ;
		fprintf( stdout, "   Ratio of maximum value to initial value = %11.4g\n", 
			ratio ) ;

		if( ratio < RATIO_MIN ) 
		  fprintf( stderr, "WARNING: The ratio appears to be small.\n" ) ;

	}
	else {
		fprintf( stdout, "\nSource Wavelet\n" ) ;
		fprintf( stdout, "   Initial value of the wavelet is 0.\n" ) ;
		fprintf( stdout, "   Check wavelet carefully.\n" ) ;
	}

	fflush( stdout ) ;
}

/* Calculate the additional source wavelet for the coaxial cable. */

static void calc_extra_wavelet( GENERAL *general, WAVELET wavelet, COAX *coax ) 

{

	float shift ;

	long int k, status ;

	k = general->coax_source ;

	/* If the coaxial cable is attached at its top (to the main part of the grid),
	   then the incident wave propagates up the cable (in the +z direction).
	   On the other hand, if the coaxial cable is attached at its bottom, then
	   the incident wave propagates down the cable (in the -z direction). */

	if( strcmp( coax[k].coax_attach, "top" ) == 0 )
	  shift = 0.5 * general->dt -
	    0.5 * coax[k].dz * sqrt( coax[k].mu_r * coax[k].epsilon_r ) / 3.e8 ;
	else
	  shift = 0.5 * general->dt +
	    0.5 * coax[k].dz * sqrt( coax[k].mu_r * coax[k].epsilon_r ) / 3.e8 ;

	calc_wavelet( general->nt, general->dt, wavelet.type, (wavelet.time_shift-shift),
		     wavelet.center_freq, wavelet.coef_exp_e, wavelet.exp_t,
		     wavelet.phase_shift, wavelet.strength, &(general->source_c[0]), &status ) ;

	/* The variable status is set to -1 (the value indicating that an error has occured)
	   if the function calc_wavelet does not recognize the wavelet type. When the
	   program enters this function, calc_wavelet has already been called once, and
	   so the wavelet type has already been checked. Thus, there is no need to check 
	   status. */
}


/* calculate the locations of the grid points */

static void calc_grid_pts( GRID *grid, long int *status )

{

	long int i, ip, in, np, nn, index_zero ;

	float slope, intercept, delta, *zp, *zn ;

	zp = (float *) mem_alloc2( GRID_PTS_MAX/2 * sizeof( float ), "zp" ) ;

	zn = (float *) mem_alloc2( GRID_PTS_MAX/2 * sizeof( float ), "zn" ) ;

	ip = 0 ;

	grid->r[ip] = grid->r_distance[0] ;

	for( i = 0 ; i < grid->nr_space-1 ; ++ i ) {

		slope = ( grid->r_step_size[i+1] - grid->r_step_size[i] ) /
		  ( grid->r_distance[i+1] - grid->r_distance[i] ) ;

		intercept = grid->r_step_size[i+1] - grid->r_distance[i+1] * slope ;

		while( grid->r[ip] < grid->r_distance[i+1] ) {

			delta = slope * grid->r[ip] + intercept ;

			ip += 1 ;

			if ( ip >= GRID_PTS_MAX ) {

				fprintf( stderr, "ERROR: function calc_grid_pts.\n" ) ;
				fprintf( stderr, "  Exceeded the maximum number of\n" ) ;
				fprintf( stderr, "  grid points in the radial direction.\n" ) ;

				*status = FLAG_ERROR ;

				return ;

			}

			grid->r[ip] = grid->r[ip-1] + delta ;

		}
	}

	np = ip + 1 ;

	grid->nr = np - 2 ;

	index_zero = -1 ;

	for ( i = 0 ; i < grid->nz_space ; ++ i )
	  if( grid->z_distance[i] == 0. )
	    index_zero = i ;

	if( index_zero == -1 ) {

		fprintf( stderr, "ERROR: function calc_grid_pts\n" ) ;
		fprintf( stderr, "  One value of z_distance must be zero.\n" ) ;

		*status = FLAG_ERROR ;

		return ;
	}

	in = 0 ;

	zn[in] = grid->z_distance[index_zero] ;

	for( i = index_zero ; i > 0 ; -- i ) {

		/* when there is an image plane, index_zero = 0. So, this
		   for-loop will not be executed. */

		slope = ( grid->z_step_size[i-1] - grid->z_step_size[i] ) /
		  ( grid->z_distance[i-1] - grid->z_distance[i] ) ;

		intercept = grid->z_step_size[i-1] - grid->z_distance[i-1] * slope ;

		while( grid->z_distance[i-1] < zn[in] ) {

			delta = slope * zn[in] + intercept ;

			in += 1 ;

			if ( in >= GRID_PTS_MAX/2 ) {

				fprintf( stderr, "ERROR: function calc_grid_pts.\n" ) ;
				fprintf( stderr, "  Exceeded the maximum number of\n" ) ;
				fprintf( stderr, "  grid points in the negative \n" ) ;
				fprintf( stderr, "  axial direction.\n" ) ;

				*status = FLAG_ERROR ;

				return ;

			}

			zn[in] = zn[in-1] - delta ;
			
		}
	}

	nn = in + 1 ;

	ip = 0 ;

	zp[ip] = grid->z_distance[index_zero] ;

	/* if there is an image plane, z_distance[index_zero] = 0. 
	   this code will start zp in the right place. */

	if ( grid->image_plane[0] == 'y' )
	  zp[ip] = grid->z_distance[index_zero] - grid->z_step_size[index_zero]/2. ;

	for( i = index_zero ; i < grid->nz_space-1 ; ++ i ) {

		slope = ( grid->z_step_size[i+1] - grid->z_step_size[i] ) /
		  ( grid->z_distance[i+1] - grid->z_distance[i] ) ;

		intercept = grid->z_step_size[i+1] - grid->z_distance[i+1] * slope ;

		while( zp[ip] < grid->z_distance[i+1] ) {

			delta = slope * zp[ip] + intercept ;

			ip += 1 ;

			if ( ip >= GRID_PTS_MAX/2 ) {

				fprintf( stderr, "ERROR: function calc_grid_pts.\n" ) ;
				fprintf( stderr, "  Exceeded the maximum number of\n" ) ;
				fprintf( stderr, "  grid points in the positive\n" ) ;
				fprintf( stderr, "  axial direction.\n" ) ;

				*status = FLAG_ERROR ;

				return ;

			}

			zp[ip] = zp[ip-1] + delta ;

		}
	}

	np = ip + 1 ;

	if ( grid->image_plane[0] != 'y' ) {

		/* this for-loop does not write zn[0] (which is zero) into
		   z */

		for ( i = 0 ; i < nn-1 ; ++ i )
		  grid->z[i] = zn[nn-1-i] ;

		for ( i = 0 ; i < np ; ++ i )
		  grid->z[i+nn-1] = zp[i] ;

		grid->nz = ( nn - 1 ) + np - 2 ;
	}
	else {

		for ( i = 0 ; i < np ; ++ i )
		  grid->z[i] = zp[i] ;

		grid->nz = np - 2 ;
	}

	free( zn ) ;
	free( zp ) ;

}


/* check the grid points */

static void check_grid_pts( GENERAL general, GRID *grid, long int *status ) 

{

	long int i, j ;

	float ratio, max_ratio, v_min, v_max, delta_r, delta_z, r_temp, v ;

	if ( grid->nr < GRID_PTS_MIN ) {
		fprintf( stderr, "WARNING: The number of grid points in\n" ) ;
		fprintf( stderr, "  in the radial direction seems small.\n" ) ;
	}

	if( grid->nr <= 0 ) {

		fprintf( stderr, "ERROR: The number of grid points in\n" ) ;
		fprintf( stderr, "  in the radial direction is less\n" ) ;
		fprintf( stderr, "  than or equal to zero\n" ) ;

		*status = FLAG_ERROR ; 

	}

	if ( grid->nz < GRID_PTS_MIN ) {
		fprintf( stderr, "WARNING: The number of grid points in\n" ) ;
		fprintf( stderr, "  in the axial direction seems small.\n" ) ;
	}

	if( grid->nz <= 0 ) {

		fprintf( stderr, "ERROR: The number of grid points in\n" ) ;
		fprintf( stderr, "  in the axial direction is less\n" ) ;
		fprintf( stderr, "  than or equal to zero\n" ) ;

		*status = FLAG_ERROR ; 

	}

	/* calculate the maximum change between the length of
	   adjacent cells. the maximum change is expressed as
	   a percentage. */

	max_ratio = 0. ;

	for ( i = 0 ; i < grid->nr ; ++ i ) {

		ratio = ( grid->r[i+2] - grid->r[i+1] ) / 
		  ( grid->r[i+1] - grid->r[i] ) ;

		if ( ratio < 1. )
		  ratio = 1. / ratio ;

		max_ratio = MAX( max_ratio, ratio ) ;
	}

	fprintf( stdout, "\nMaximum percent change in cell length\n" ) ;
	fprintf( stdout, "   Radial direction: %-11.4g\n", ( max_ratio - 1. ) * 100. ) ;

	max_ratio = 0. ;

	for ( i = 0 ; i < grid->nz ; ++ i ) {

		ratio = fabs( grid->z[i+2] - grid->z[i+1] ) / 
		  fabs( grid->z[i+1] - grid->z[i] ) ;

		if ( ratio < 1. )
		  ratio = 1. / ratio ;

		max_ratio = MAX( max_ratio, ratio ) ;
	}

	fprintf( stdout, "   Axial direction: %-11.4g\n", ( max_ratio - 1. ) * 100. ) ;

	/* calculate the maximum velocity at which the stability
	   condition will be satisfied. */

	v_min = FLT_MAX ;
	v_max = - FLT_MAX ;

	for( i = 0 ; i < grid->nr+1 ; ++i ) {

		delta_r = grid->r[i+1] - grid->r[i] ;

		for( j = 0 ; j < grid->nz+1 ; ++j ) {

			delta_z = grid->z[j+1] - grid->z[j] ;

			r_temp = 1./(delta_r*delta_r) + 1./(delta_z*delta_z) ;

			v = 1. / ( general.dt * sqrt( r_temp ) ) ;

			v_min = MIN( v_min, v ) ;

			v_max = MAX( v_max, v ) ;

		}

	}

        fprintf( stdout, "\nThe stability condition will be satisfied\n" ) ;
        fprintf( stdout, "for velocities between %-11.4g and %-11.4g\n", v_min, v_max ) ;

	fflush( stdout ) ;

}

/* For each point in the grid, determine which em region it is in.
   Assign the index (for the em region) to that grid point. 
*/

static void calc_em_indices( GRID *grid, long int n_em_region, EM_REGION *em_region, 
		       long int *status )

{
	long int i, j, index, *n_vertices ;

	float dr_lft, dr_rht, dz_bot, dz_top, *r_vertices, *z_vertices, r_pt, z_pt ;

	/* put the vertices into (column major) arrays. 

	   The non-uniform grid extends slightly beyond what was specified for the 
	   spacing of the grid points (file grid.spacing_filename). This is 
	   because of the method that was used to calculate the non-uniform grid. 
	   Consequently, the grid extends slighly beyond the area covered by the 
	   em regions. To account for this, the em regions (adjacent to the 
	   left, right, top, and bottom boundarys) are extended slightly. 
	   This is done by adjusting the vertices adjacent to the boundaries.
	   */

	dr_lft = grid->r[1] - grid->r[0] ;
	dr_rht = grid->r[grid->nr+1] - grid->r[grid->nr] ;
	dz_bot = grid->z[1] - grid->z[0] ;
	dz_top = grid->z[grid->nz+1] - grid->z[grid->nz] ;

	i = n_em_region * sizeof( long int ) ;
	n_vertices = (long int *) mem_alloc2( i, "n_vertices" ) ;

	i = ( VERTICES_MAX + 1 ) * n_em_region * sizeof( float ) ;
	r_vertices = (float *) mem_alloc2( i, "r_vertices" ) ;

	z_vertices = (float *) mem_alloc2( i, "z_vertices" ) ;

	for ( i = 0 ; i < n_em_region ; ++ i ) {

		n_vertices[i] = em_region[i].n_vertices ;

		for ( j = 0 ; j < n_vertices[i] ; ++j ) {

			R_VERTICES(i,j) = em_region[i].r[j] ;
			Z_VERTICES(i,j) = em_region[i].z[j] ;

			/* check if r corresponds to the left boundary */
			if ( R_VERTICES(i,j) == grid->r_distance[0] )
			  R_VERTICES(i,j) -= dr_lft ;

			/* check if r corresponds to the right boundary */
			if ( R_VERTICES(i,j) == grid->r_distance[grid->nr_space-1] )
			  R_VERTICES(i,j) += dr_rht ;

			/* check if z corresponds to the bottom boundary */
			if ( Z_VERTICES(i,j) == grid->z_distance[0] )
			  Z_VERTICES(i,j) -= dz_bot ;

			/* check if z corresponds to the top boundary */
			if ( Z_VERTICES(i,j) == grid->z_distance[grid->nz_space-1] )
			  Z_VERTICES(i,j) += dz_top ;

		}

		R_VERTICES(i,j) = R_VERTICES(i,0) ;
		Z_VERTICES(i,j) = Z_VERTICES(i,0) ;

	}

	/* grid points associated with Ez */

	for ( i = 0 ; i < grid->nr+1 ; ++ i ) {

		r_pt = ( grid->r[i+1] + grid->r[i] ) / 2. ;

		for ( j = 0 ; j < grid->nz ; ++ j ) {

			z_pt = grid->z[j+1] ;

			find_em_region( &r_pt, &z_pt, n_em_region, n_vertices, 
				       r_vertices, z_vertices, &index, status ) ;

			if ( *status == FLAG_ERROR ) {
				fprintf( stderr, "ERROR: Function calc_em_indices.\n" ) ;
				fprintf( stderr, "  The Ez grid point at %-11.4g, %-11.4g\n", 
					r_pt, z_pt ) ;
				fprintf( stderr, "  is not within an em region.\n" ) ;
				return ;
			}

			INDEX_EZ(i,j) = (short int) index ;
		}
	}

	/* grid points associated with Er */

	for ( i = 0 ; i < grid->nr ; ++ i ) {

		r_pt = grid->r[i+1] ;

		for ( j = 0 ; j < grid->nz+1 ; ++ j ) {

			z_pt = ( grid->z[j+1] + grid->z[j] ) / 2. ;

			find_em_region( &r_pt, &z_pt, n_em_region, n_vertices, 
    				    r_vertices, z_vertices, &index, status ) ;

			if ( *status == FLAG_ERROR ) {
				fprintf( stderr, "ERROR: Function calc_em_indices.\n" ) ;
				fprintf( stderr, "  The Er grid point at %-11.4g, %-11.4g\n",
					r_pt, z_pt ) ;
				fprintf( stderr, "  is not within an em region.\n" ) ;
				return ;
			}

			INDEX_ER(i,j) = (short int) index ;
		}
	}

	/* grid points associated with H */

	for ( i = 0 ; i < grid->nr ; ++ i ) {

		r_pt = grid->r[i+1] ;

		for ( j = 0 ; j < grid->nz ; ++ j ) {

			z_pt = grid->z[j+1] ;

			find_em_region( &r_pt, &z_pt, n_em_region, n_vertices, 
				    r_vertices, z_vertices, &index, status ) ;

			if ( *status == FLAG_ERROR ) {
				fprintf( stderr, "ERROR: Function calc_em_indices.\n" ) ;
				fprintf( stderr, "  The H grid point at %-11.4g, %-11.4g\n",
					r_pt, z_pt ) ;
				fprintf( stderr, "  is not within an em region.\n" ) ;
				return ;
			}

			INDEX_H(i,j) = (short int) index ;

			/* If H is inside a region that is metal, then the surrounding
			   EZ and ER are also assigned to that region. */

			if( SIGMA_METAL <= em_region[index].sigma ) {
				INDEX_EZ(i,j) = INDEX_EZ(i+1,j) = (short int) index ;
				INDEX_ER(i,j) = INDEX_ER(i,j+1) = (short int) index ;
			}
		}
	}

	free( n_vertices ) ;
	free( r_vertices ) ;
	free( z_vertices ) ;
	  
}


/* Find the region (defined by its electromagnetic properties)
in which the grid point is located. */

static void find_em_region( float *r, float *z, long int n_em_region, long int *n_vertices, 
		    float *r_vertices, float *z_vertices, long int *index, 
		    long int *status ) 

{
	long int flag ;

	*status = FLAG_OK ;

	for ( *index = 0 ; *index < n_em_region ; ++ *index ) {

		poly_pts( &(n_vertices[*index]), &(R_VERTICES(*index,0)),
			 &(Z_VERTICES(*index,0)), r, z, &flag ) ;

		if ( flag != 0 ) return ;
			
	}

	*status = FLAG_ERROR ;

}

/* For each thin layer, check that its thickness is small compared
   to the spacing of the grid points. Count the number of grid
   points associated with the thin layer and determine the indices
   of the grid points. */

static void calc_tl_indices( GRID *grid, long int n_thin_layer, THIN_LAYER *thin_layer, 
			    long int *status )

{
	long int i, j, k ;

	float layer_center, diff, r_tmp, r_pt1, r_pt2 ;

	for( k = 0 ; k < n_thin_layer ; ++k ) {

		/* determine the i index of the grid points (for H 
		   and ER) that are associated with the thin layer. */

		layer_center = thin_layer[k].inner_radius + 
		  0.5 * thin_layer[k].radial_thickness ;

		diff = FLT_MAX ;

		for( i = 0 ; i < grid->nr ; ++ i ) {

			r_tmp = fabs( grid->r[i+1] - layer_center ) ;

			if( r_tmp < diff ) {
				diff = r_tmp ;
				thin_layer[k].i_index_h = i ;
			}
		}
		
		/* determine the i index of the grid points (for Ez)
		   that are associated with the thin layer. 
		   
		   this index equals either i_index_h or i_index_h+1, and so the
		   search is restricted to these two points. 
		   */

		i = thin_layer[k].i_index_h ;

		r_pt1 = ( grid->r[i+1] + grid->r[i] ) / 2. ;

		r_pt2 = ( grid->r[i+2] + grid->r[i+1] ) / 2. ;

		if ( fabs( layer_center - r_pt1 ) <= fabs( layer_center - r_pt2 ) )
		  thin_layer[k].i_index_ez = i ;
		else
		  thin_layer[k].i_index_ez = i + 1 ;

		/* Check the thickness. */

		r_tmp = REL_THICK_MAX * ( grid->r[i+1] - grid->r[i] ) ;

		if( thin_layer[k].radial_thickness > r_tmp ) {

			fprintf( stderr, "ERROR: The thickness of thin layer %3ld\n", k+1 ) ;
			fprintf( stderr, "   is too big compared to the grid spacing.\n" ) ;

			*status = FLAG_ERROR ;
			
			return ;
		}

		/* determine the j index of the grid point (for H) that is
		   associated with the bottom of the thin layer. */
		
		diff = FLT_MAX ;

		for( j = 0 ; j < grid->nz ; ++ j ) {

			r_tmp = fabs( grid->z[j+1] - thin_layer[k].bottom ) ;

			if( r_tmp < diff ) {
				diff = r_tmp ;
				thin_layer[k].j_index = j ;
			}
		}

		/* count the number of grid points (for H) that are
		   associated with the thin layer. */

		thin_layer[k].n_index = 0 ;

		for( j = thin_layer[k].j_index ; j < grid->nz ; ++ j ) 
		  if( thin_layer[k].bottom <=  grid->z[j+1] &&  
		     grid->z[j+1] <= thin_layer[k].top )
		    thin_layer[k].n_index ++ ;

	}

}

/* calculate time and space (grid) indices for the scans */

static void calc_scan_indices( SCAN *scan, SCAN_LOC *scan_loc, GRID *grid, GENERAL general )

{

	long int i, j, k ;

	float diff, r_pt, z_pt ;

	scan->time_index_start = round_off ( scan->time_start / general.dt ) ;
	scan->time_index_step = round_off ( scan->time_step / general.dt ) ;
	scan->time_index_end = round_off ( scan->time_end / general.dt ) ;

	fprintf( stdout, "\nScan Information\n" ) ;

	fprintf( stdout, "The sample interval is %-11.4g,\n", 
		scan->time_index_step * general.dt ) ;
	fprintf( stdout, "which corresponds to %-4ld * dt.\n", scan->time_index_step ) ;
	fprintf( stdout, "The time indices for the first and last samples are %-5ld and %-5ld\n",
		scan->time_index_start, scan->time_index_end ) ;

	fprintf( stdout, "\nNearest Grid Points For Scan Location\n\n" ) ;
	fprintf( stdout, "Scan    I      J\n" ) ;
	fprintf( stdout, "Number  Index  Index  R            Z\n" ) ;

	for ( k = 0 ; k < scan->n_scan ; ++k ) {

		if ( strcmp( scan_loc[k].field_comp, "ez" ) == 0 ) {

			diff = FLT_MAX ;
			for ( i = 0 ; i < grid->nr+1 ; ++ i ) {
				r_pt = ( grid->r[i+1] + grid->r[i] ) / 2. ;
				if( fabs( scan_loc[k].r - r_pt ) < diff ) {
					diff = fabs( scan_loc[k].r - r_pt ) ;
					scan_loc[k].i_index = i ;
				}
			}

			diff = FLT_MAX ;
			for ( j = 0 ; j < grid->nz ; ++ j ) {
				z_pt = grid->z[j+1] ;
				if( fabs( scan_loc[k].z - z_pt ) < diff ) {
					diff = fabs( scan_loc[k].z - z_pt ) ;
					scan_loc[k].j_index = j ;
				}
			}

			i = scan_loc[k].i_index ;
			j = scan_loc[k].j_index ;
			r_pt = ( grid->r[i+1] + grid->r[i] ) / 2. ;
			z_pt = grid->z[j+1] ;

		}

		else if ( strcmp( scan_loc[k].field_comp, "er" ) == 0 ) {

			diff = FLT_MAX ;
			for ( i = 0 ; i < grid->nr ; ++ i ) {
				r_pt = grid->r[i+1] ;
				if( fabs( scan_loc[k].r - r_pt ) < diff ) {
					diff = fabs( scan_loc[k].r - r_pt ) ;
					scan_loc[k].i_index = i ;
				}
			}

			diff = FLT_MAX ;
			for ( j = 0 ; j < grid->nz+1 ; ++ j ) {
				z_pt = ( grid->z[j+1] + grid->z[j] ) / 2. ;
				if( fabs( scan_loc[k].z - z_pt ) < diff ) {
					diff = fabs( scan_loc[k].z - z_pt ) ;
					scan_loc[k].j_index = j ;
				}
			}

			i = scan_loc[k].i_index ;
			j = scan_loc[k].j_index ;
			r_pt = grid->r[i+1] ;
			z_pt = ( grid->z[j+1] + grid->z[j] ) / 2. ;

		}

		else if ( strcmp( scan_loc[k].field_comp, "h" ) == 0 ) {

			diff = FLT_MAX ;
			for ( i = 0 ; i < grid->nr ; ++ i ) {
				r_pt = grid->r[i+1] ;
				if( fabs( scan_loc[k].r - r_pt ) < diff ) {
					diff = fabs( scan_loc[k].r - r_pt ) ;
					scan_loc[k].i_index = i ;
				}
			}

			diff = FLT_MAX ;
			for ( j = 0 ; j < grid->nz ; ++ j ) {
				z_pt = grid->z[j+1] ;
				if( fabs( scan_loc[k].z - z_pt ) < diff ) {
					diff = fabs( scan_loc[k].z - z_pt ) ;
					scan_loc[k].j_index = j ;
				}
			}

			i = scan_loc[k].i_index ;
			j = scan_loc[k].j_index ;
			r_pt = grid->r[i+1] ;
			z_pt = grid->z[j+1] ;
		}

		fprintf( stdout, "%-3ld     %-4ld   %-4ld   %-11.4g  %-11.4g\n", 
			k+1, i, j, r_pt, z_pt ) ;
			
	}

}

/* Check the location of the snapshots and the requested field
   component. Calculate the time and space (grid) indices for the
   snapshots.
*/

static void calc_snap_indices( SNAP *snap, GRID *grid, GENERAL general, long int *status )

{

	long int i, j, is, js ;

	float rs, zs, diff, r_pt, z_pt ;

	*status = FLAG_OK ;

	snap->time_index_start = round_off ( snap->time_start / general.dt ) ;
	snap->time_index_step = round_off ( snap->time_step / general.dt ) ;
	snap->time_index_end = round_off ( snap->time_end / general.dt ) ;

	fprintf( stdout, "\nSnapshot Information\n" ) ;

	fprintf( stdout, "The time interval between snapshots is %-11.4g,\n", 
		snap->time_index_step * general.dt ) ;
	fprintf( stdout, "which corresponds to %-4ld * dt.\n", snap->time_index_step ) ;
	fprintf( stdout, "The time indices for the first and last snapshots are %-5ld and %-5ld\n",
		snap->time_index_start, snap->time_index_end ) ;

	if ( snap->r_start < grid->r[0] || snap->r_end >= grid->r[grid->nr+1] ) {

		fprintf( stderr, "ERROR: Function calc_snap_indices.\n" ) ;
		fprintf( stderr, "  The snapshot is outside\n" ) ;
		fprintf( stderr, "  the grid (in the r direction).\n" ) ;

		*status = FLAG_ERROR ;

		return ;

	}

	if ( snap->z_start <= grid->z[0] || snap->z_end >= grid->z[grid->nz+1] ) {

		fprintf( stderr, "ERROR: Function calc_snap_indices.\n" ) ;
		fprintf( stderr, "  The snapshot is outside\n" ) ;
		fprintf( stderr, "  the grid (in the z direction).\n" ) ;

		*status = FLAG_ERROR ;

		return ;

	}

	snap->n_i = ( snap->r_end - snap->r_start ) / snap->r_step + 1 ;

	snap->n_j = ( snap->z_end - snap->z_start ) / snap->z_step + 1 ;
	
	fprintf( stdout, "   Number of pixels in r direction = %-5ld\n",
		snap->n_i );

	fprintf( stdout, "   Number of pixels in z direction = %-5ld\n\n",
		snap->n_j );

	if ( snap->n_i > INDICES_MAX || snap->n_j > INDICES_MAX ) {

		fprintf( stderr, "ERROR: Function calc_snap_indices.\n" ) ;
		fprintf( stderr, "  Too many locations are requested for\n" ) ;
		fprintf( stderr, "  the snapshot.\n" ) ;

		*status = FLAG_ERROR ;

		return ;
	}

	if ( strcmp( snap->field_comp, "ez" ) == 0 ) {

		for ( is = 0 ; is < snap->n_i ; ++ is ) {
			
			rs = snap->r_step * is + snap->r_start ;

			diff = FLT_MAX ;
			for ( i = 0 ; i < grid->nr+1 ; ++ i ) {
				r_pt = ( grid->r[i+1] + grid->r[i] ) / 2. ;
				if( fabs( rs - r_pt ) < diff ) {
					diff = fabs( rs - r_pt ) ;
					snap->i_index[is] = i ;
				}
			}
		}

		for ( js = 0 ; js < snap->n_j ; ++ js ) {

			zs = snap->z_step * js + snap->z_start ;

			diff = FLT_MAX ;
			for ( j = 0 ; j < grid->nz ; ++ j ) {
				z_pt = grid->z[j+1] ;
				if( fabs( zs - z_pt ) < diff ) {
					diff = fabs( zs - z_pt ) ;
					snap->j_index[js] = j ;
				}
			}

		}
	}

	else if ( strcmp( snap->field_comp, "er" ) == 0 ) {

                for ( is = 0 ; is < snap->n_i ; ++ is ) {

                        rs = snap->r_step * is + snap->r_start ;

			diff = FLT_MAX ;
			for ( i = 0 ; i < grid->nr ; ++ i ) {
				r_pt = grid->r[i+1] ;
				if( fabs( rs - r_pt ) < diff ) {
					diff = fabs( rs - r_pt ) ;
					snap->i_index[is] = i ;
				}
			}
		}

		for ( js = 0 ; js < snap->n_j ; ++ js ) {

                        zs = snap->z_step * js + snap->z_start ;

			diff = FLT_MAX ;
			for ( j = 0 ; j < grid->nz+1 ; ++ j ) {
				z_pt = ( grid->z[j+1] + grid->z[j] ) / 2. ;
				if( fabs( zs - z_pt ) < diff ) {
					diff = fabs( zs - z_pt ) ;
					snap->j_index[js] = j ;
				}
			}
		}
	}
		
	else if ( strcmp( snap->field_comp, "h" ) == 0 ||
		 strcmp( snap->field_comp, "e_mag" ) == 0 ||
		 strcmp( snap->field_comp, "ezer" ) == 0 ||
		 strcmp( snap->field_comp, "ezerh" ) == 0 ) {

                for ( is = 0 ; is < snap->n_i ; ++ is ) {

                        rs = snap->r_step * is + snap->r_start ;

			diff = FLT_MAX ;
			for ( i = 0 ; i < grid->nr ; ++ i ) {
				r_pt = grid->r[i+1] ;
				if( fabs( rs - r_pt ) < diff ) {
					diff = fabs( rs - r_pt ) ;
					snap->i_index[is] = i ;
				}
			}
		}

                for ( js = 0 ; js < snap->n_j ; ++ js ) {

                        zs = snap->z_step * js + snap->z_start ;

			diff = FLT_MAX ;
			for ( j = 0 ; j < grid->nz ; ++ j ) {
				z_pt = grid->z[j+1] ;
				if( fabs( zs - z_pt ) < diff ) {
					diff = fabs( zs - z_pt ) ;
					snap->j_index[js] = j ;
				}
			}

		}
	}
		
	else {

		fprintf( stderr, "ERROR:   Function calc_snap_indices.\n" ) ;
		fprintf( stderr, "  Requested component of the field for\n" ) ;
		fprintf( stderr, "  the snapshot is unknown.\n" ) ;
		*status = FLAG_ERROR ;

	}

}

/* This function is used if the source is either a delta
   gap or a electric conduction current density.
   Check that its location is within the grid.
   Calculate the space (grid) indices for its location.
*/

static void calc_src_indices( GRID *grid, GENERAL *general, long int *status )

{

	long int i, j ;

	float diff, r_pt, z_pt ;

	*status = FLAG_OK ;

	if ( general->src_loc_r < grid->r[0] || general->src_loc_r > grid->r[grid->nr+1] ) {

		fprintf( stderr, "ERROR: Function calc_src_indices.\n" ) ;
		fprintf( stderr, "  Location of the source is outside\n" ) ;
		fprintf( stderr, "  the grid (in the r direction).\n" ) ;

		*status = FLAG_ERROR ;

		return ;

	}

	if ( general->src_loc_z < grid->z[0] || general->src_loc_z > grid->z[grid->nz+1] ) {

		fprintf( stderr, "ERROR: Function calc_src_indices.\n" ) ;
		fprintf( stderr, "  Location of the source is outside\n" ) ;
		fprintf( stderr, "  the grid (in the z direction).\n" ) ;

		*status = FLAG_ERROR ;

		return ;

	}

	diff = FLT_MAX ;
	for ( i = 0 ; i < grid->nr+1 ; ++ i ) {
		r_pt = ( grid->r[i+1] + grid->r[i] ) / 2. ;
		if( fabs( general->src_loc_r - r_pt ) < diff ) {
			diff = fabs( general->src_loc_r - r_pt ) ;
			general->src_loc_i = i ;
		}
	}

	diff = FLT_MAX ;
	for ( j = 0 ; j < grid->nz ; ++ j ) {
		z_pt = grid->z[j+1] ;
		if( fabs( general->src_loc_z - z_pt ) < diff ) {
			diff = fabs( general->src_loc_z - z_pt ) ;
			general->src_loc_j = j ;
		}
	}

}

static void calc_coax_grid( long int n_coax, COAX *coax, GRID *grid, long int *status )

{

	long int i, j, jj, k ;

	float smallest_diff, diff ;

	*status = FLAG_OK ;

	for( k = 0 ; k < n_coax ; ++ k ) {
		
		/* Calculate the space indices for the leftmost attachment point 
		   to the main grid. This attachment point is associated with Er. */

		smallest_diff = FLT_MAX ;
		for ( i = 0 ; i < grid->nr ; ++ i ) {
			diff = fabs ( grid->r[i+1] - coax[k].grid_attach_r ) ;
			if( diff < smallest_diff ) {
				smallest_diff = diff ;
				coax[k].grid_attach_i = i ;
			}
		}

		smallest_diff = FLT_MAX ;
		for ( j = 0 ; j < grid->nz+1 ; ++ j ) {
			diff = fabs ( ( grid->z[j+1] + grid->z[j] ) / 2. - coax[k].grid_attach_z ) ;
			if( diff < smallest_diff ) {
				smallest_diff = diff ;
				coax[k].grid_attach_j = j ;
			}
		}

		/* Calculate the locations of the grid points (for H). These locations
		 are in the coordinate system for the main grid. (It turns out that,
		 in the coordinate systems for both the main grid and the coaxial cable,
		 the r axes coincide.) */

		coax[k].r[0] = 0. ;

		i = coax[k].grid_attach_i ;

		coax[k].r[1] = grid->r[i+1] ;

		for( i = 2 ; i < coax[k].nr+2 ; ++ i )
		  coax[k].r[i] = coax[k].r[i-1] + coax[k].dr ;

		jj = coax[k].grid_attach_j ;

		if( strcmp( coax[k].coax_attach, "top" ) == 0 ) {

			/* The top of the coaxial cable is attached to the main grid.
			   Note that these computed locations are in reverse order --- 
			   from largest to smallest. */

			for( j = 0 ; j < coax[k].nz+2 ; ++ j )
			  coax[k].z[j] = grid->z[jj+1] - coax[k].dz * j ;

		}
		else {

			/* The bottom of the coaxial cable is attached to the main grid. */

			for( j = 0 ; j < coax[k].nz+2 ; ++ j )
			  coax[k].z[j] = grid->z[jj+1] - coax[k].dz * ( 1 - j ) ;

		}

		/* Check that, at the attachment point, the grid
		   for the cable matches the main grid. */

		for( i = coax[k].grid_attach_i ; i < coax[k].grid_attach_i + coax[k].nr - 1 ; ++ i ) {

			diff = grid->r[i+2] - grid->r[i+1] ;

			if( fabs( ( coax[k].dr - diff ) / diff ) > GRID_DIF_MAX ){

				fprintf( stderr, "ERROR: At the attachment point\n" ) ;
				fprintf( stderr, "   for coaxial cable %-4ld\n", k+1 ) ;
				fprintf( stderr, "   the cable grid spacing (in the r\n" ) ;
				fprintf( stderr, "   direction) does not match the\n" ) ;
				fprintf( stderr, "   main grid spacing.\n" ) ;

			*status = FLAG_ERROR ;

			}
		}

		j = coax[k].grid_attach_j ;

		diff = ( grid->z[j+2] - grid->z[j] ) / 2. ;

		if( fabs( ( coax[k].dz - diff ) / diff ) > GRID_DIF_MAX ){

			fprintf( stderr, "ERROR: At the attachment point\n" ) ;
			fprintf( stderr, "   for coaxial cable %-4ld\n", k+1 ) ;
			fprintf( stderr, "   the cable grid spacing (in the z\n" ) ;
			fprintf( stderr, "   direction) does not match the\n" ) ;
			fprintf( stderr, "   main grid spacing.\n" ) ;

			*status = FLAG_ERROR ;

		}

		/* Calculate the space index for the location of the source
		   (if it is a source cable). */

		if( coax[k].src_loc_z > 0. ) {

			smallest_diff = FLT_MAX ;
			for( j = 1 ; j < coax[k].nz-1 ; ++ j ) {

				diff = fabs ( (j+0.5) * coax[k].dz - coax[k].src_loc_z ) ;

				if( diff < smallest_diff ) {
					smallest_diff = diff ;
					coax[k].src_loc_j = j ;
				}
			}
		}
		else
		  coax[k].src_loc_j = -1 ;

		/* Calculate the space indices for the location at which the voltage 
		   and the current are calculated. */

		smallest_diff = FLT_MAX ;
		for( j = 0 ; j < coax[k].nz ; ++ j ) {

			diff = fabs ( j * coax[k].dz - coax[k].volt_cur_loc_z ) ;

			if( diff < smallest_diff ) {
				smallest_diff = diff ;
				coax[k].volt_cur_loc_j = j ;
			}
		}
	}
}


static void write_idl_file( char filename[], GRID *grid, GENERAL general, COAX *coax, 
			   SCAN scan, SCAN_LOC *scan_loc, SNAP snap,
			   EM_REGION *em_region, THIN_LAYER *thin_layer )

{

	long int i, j, ii, jj, count, one=1 ;

	FILE *fp ;

	/*
	  ---------------------------------------------------------------
	  If this code is compiled on a UNIX system, the second argument
	  must be "w".

	  If this code is compiled on a PC, the second argument
	  must be "wb".
	  ---------------------------------------------------------------
	  */

	fp = file_open2( filename, "wb" ) ;

	/*
	  ---------------------------------------------------------------
	  The number of grid points that are in the r direction and are
	  associated with H.
	  ---------------------------------------------------------------
	  */

	file_write_long( fp, &(grid->nr), one ) ;

	/*
	  ---------------------------------------------------------------
	  The grid points in the r direction.
	  ---------------------------------------------------------------
	  */

	file_write_float( fp, &(grid->r[0]), grid->nr+2 ) ;

	/*
	  ---------------------------------------------------------------
	  The number of grid points that are in the z direction and are
	  associated with H.
	  ---------------------------------------------------------------
	  */

	file_write_long( fp, &(grid->nz), one ) ;

	/*
	  ---------------------------------------------------------------
	  The grid points in the z direction.
	  ---------------------------------------------------------------
	  */

	file_write_float( fp, &(grid->z[0]), grid->nz+2 ) ;

	/*
	  ---------------------------------------------------------------
	  The type of source:
	  1 means delta gap
	  2 means electric conduction current density
	  3 means coaxial cable
	  ---------------------------------------------------------------
	  */

	switch( general.src_type[0] ) {

	    case 'd': 
		i = 1 ;
		break ;
	    case 'e':
		i = 2 ;
		break ;
	    case 'c':
		i = 3 ;
		break ;
	}

	file_write_long( fp, &i, one ) ;

	/*
	  ---------------------------------------------------------------
	  The source location. 
	  (The source location is defined only if the source is a 
	  delta gap or an electric conduction current density. That is, 
	  if general->src_type[0] = 'd' or 'e'. The source location is 
	  undefined if the source is a coaxial cable. That is, if 
	  general->src_type[0] = 'c'. In this case, the location is
	  meaningless.)
	  ---------------------------------------------------------------
	  */

	file_write_float( fp, &(general.src_loc_r), one ) ;	

	file_write_float( fp, &(general.src_loc_z), one ) ;	

	/*
	  ---------------------------------------------------------------
	  The number of coaxial cables. This number is zero, if there are
	  no coaxial cables.
	  ---------------------------------------------------------------
	  */

	file_write_long( fp, &(general.n_coax), one ) ;	

	for( i = 0 ; i < general.n_coax ; ++ i ) {

		/*
		  -------------------------------------------------------
		  The number of grid points that are in the r direction
		  and are associated with H.
		  -------------------------------------------------------
		  */

		file_write_long( fp, &(coax[i].nr), one ) ;

		/*
		  -------------------------------------------------------
		  The spacing between the grid points in the r direction.
		  -------------------------------------------------------
		  */

		file_write_float( fp, &(coax[i].dr), one ) ;	

		/*
		  -------------------------------------------------------
		  The grid points in the r direction.
		  -------------------------------------------------------
		  */

		file_write_float( fp, &(coax[i].r[0]), coax[i].nr+2 ) ;

		/*
		  -------------------------------------------------------
		  The number of grid points that are in the z direction
		  and are associated with H.
		  -------------------------------------------------------
		  */

		file_write_long( fp, &(coax[i].nz), one ) ;

		/*
		  -------------------------------------------------------
		  The spacing between the grid points in the z direction.
		  -------------------------------------------------------
		  */

		file_write_float( fp, &(coax[i].dz), one ) ;	

		/*
		  -------------------------------------------------------
		  The grid points in the z direction.
		  -------------------------------------------------------
		  */

		file_write_float( fp, &(coax[i].z[0]), coax[i].nz+2 ) ;

		/*
		  -------------------------------------------------------
		  The index (corresponding to the z direction) at which 
		  the source is applied. This index applies to the grid 
		  for the coaxial cable. if a source is not applied in 
		  this cable, then this index is -1.
		  -------------------------------------------------------
		  */

		file_write_long( fp, &(coax[i].src_loc_j), one ) ;	

		/*
		  -------------------------------------------------------
		  The index (corresponding to the z direction) at which 
		  the voltage and the current are calculated. this index 
		  applies to the grid for the coaxial cable.
		  -------------------------------------------------------
		  */

		file_write_long( fp, &(coax[i].volt_cur_loc_j), one ) ;	
		
		/*
		  -------------------------------------------------------
		  The indices at which the coaxial cable is attached to 
		  the main grid. These indices apply to the main grid. 
		  Furthermore, the indices refer only to the left-most 
		  grid point (because there are coax[i].nr points at 
		  which the cable is attached to the main grid).
		  -------------------------------------------------------
		  */

		file_write_long( fp, &(coax[i].grid_attach_i), one ) ;	

		file_write_long( fp, &(coax[i].grid_attach_j), one ) ;	

		/*
		  -------------------------------------------------------
		  A flag indicating whether the top or the bottom of the 
		  coaxial cable is attached to the main grid. 
		  1 means the top
		  2 means the bottom
		  -------------------------------------------------------
		  */

		switch( coax[i].coax_attach[0] ) {

		    case 't': 
			j = 1 ;
			break ;
		    case 'b':
			j = 2 ;
			break ;
		}

		file_write_long( fp, &j, one ) ;

	}

	/*
	  ---------------------------------------------------------------
	  The number of scans. If scans are not calculated, the number is
	  zero.
	  ---------------------------------------------------------------
	  */

	file_write_long( fp, &(scan.n_scan), one ) ;	

	/*
	  ---------------------------------------------------------------
	  The scan locations.
	  ---------------------------------------------------------------
	  */

	for ( i = 0 ; i < scan.n_scan ; ++i ) {
		file_write_float( fp, &(scan_loc[i].r), one ) ;
		file_write_float( fp, &(scan_loc[i].z), one ) ;
	}

	/*
	  ---------------------------------------------------------------
	  Flag indicating whether snapshots are calculated. 
	  0 means no
	  1 means yes
	  ---------------------------------------------------------------
	  */

	switch( snap.flag[0] ) {

	    case 'n': 
		i = 0 ;
		break ;
	    case 'y':
		i = 1 ;
		break ;
	}

	file_write_long( fp, &i, one ) ;

	/*
	  ---------------------------------------------------------------
	  The snapshot location. If there are no snapshots, then these 
	  numbers are meaningless.
	  ---------------------------------------------------------------
	  */

	file_write_float( fp, &(snap.r_start), one ) ;
	file_write_float( fp, &(snap.r_end), one ) ;
	file_write_float( fp, &(snap.z_start), one ) ;
	file_write_float( fp, &(snap.z_end), one ) ;

	/*
	  ---------------------------------------------------------------
	  The number of em_regions.
	  ---------------------------------------------------------------
	  */

	file_write_long( fp, &(general.n_em_region), one ) ;

	/*
	  ---------------------------------------------------------------
	  For each em region, the number of vertices, the coordinates of
	  the vertices, and a flag indicating whether the region is 
	  metal.
	  ---------------------------------------------------------------
	  */

	for ( i = 0 ; i < general.n_em_region ; ++i ) {

		file_write_long( fp, &(em_region[i].n_vertices), one ) ;

		for ( j = 0 ; j < em_region[i].n_vertices ; ++j ) {

			file_write_float( fp, &(em_region[i].r[j]), one ) ;
			file_write_float( fp, &(em_region[i].z[j]), one ) ;

		}

		if( SIGMA_METAL <= em_region[i].sigma )
		  j = 1 ;
		else 
		  j = 0 ;

		file_write_long( fp, &j, one ) ;

	}

	/*
	  ---------------------------------------------------------------
	  The number of grid points (for H) that are within em regions 
	  that are metal. This number must be computed first.
	  ---------------------------------------------------------------
	  */

	count = 0 ;

	for ( i = 0 ; i < general.n_em_region ; ++i ) {

		if( SIGMA_METAL <= em_region[i].sigma ) {

			for ( ii = 0 ; ii < grid->nr ; ++ ii )
			  for ( jj = 0 ; jj < grid->nz ; ++ jj ) 
			    if( INDEX_H(ii,jj) == i ) count += 1 ;
		}

	}

	file_write_long( fp, &count, one ) ;

	/*
	  ---------------------------------------------------------------
	  The indices of the grid points (for H) that are within em 
	  regions that are metal. 
	  ---------------------------------------------------------------
	  */
	
	for ( i = 0 ; i < general.n_em_region ; ++i ) {

		if( SIGMA_METAL <= em_region[i].sigma ) {

			for ( ii = 0 ; ii < grid->nr ; ++ ii )
			  for ( jj = 0 ; jj < grid->nz ; ++ jj )
			    if( INDEX_H(ii,jj) == i ) {
				    file_write_long( fp, &ii, one ) ;
				    file_write_long( fp, &jj, one ) ;
			    }
		}
	}

	/*
	  ---------------------------------------------------------------
	  The number of thin layers. This number is 0, if there are no 
	  thin layers.
	  ---------------------------------------------------------------
	  */

	file_write_long( fp, &(general.n_thin_layer), one ) ;

	for ( i = 0 ; i < general.n_thin_layer ; ++i ) {

		/*
		  -------------------------------------------------------
		  Inner radius.
		  -------------------------------------------------------
		  */

		file_write_float( fp, &(thin_layer[i].inner_radius), one ) ;

		/*
		  -------------------------------------------------------
		  Thickness in the radial direction.
		  -------------------------------------------------------
		  */

		file_write_float( fp, &(thin_layer[i].radial_thickness), one ) ;

		/*
		  -------------------------------------------------------
		  z coordinate of the top edge.
		  -------------------------------------------------------
		  */

		file_write_float( fp, &(thin_layer[i].top), one ) ;

		/*
		  -------------------------------------------------------
		  z coordinate of the bottom edge.
		  -------------------------------------------------------
		  */

		file_write_float( fp, &(thin_layer[i].bottom), one ) ;

		/*
		  -------------------------------------------------------
		  i index of the grid points (for H and ER) that are
		  associated with the thin layer.
		  -------------------------------------------------------
		  */

		file_write_long( fp, &(thin_layer[i].i_index_h), one ) ;

		/*
		  -------------------------------------------------------
		  i index of the grid points (for Ez) that are associated
		  with the thin layer.
		  -------------------------------------------------------
		  */

		file_write_long( fp, &(thin_layer[i].i_index_ez), one ) ;

		/*
		  -------------------------------------------------------
		  j index of the grid point (for H) that is associated
		  with the bottom of the thin layer.
		  -------------------------------------------------------
		  */

		file_write_long( fp, &(thin_layer[i].j_index), one ) ;

		/*
		  -------------------------------------------------------
		  Number of grid points (for H) that are associated with
		  the thin layer.
		  -------------------------------------------------------
		  */

		file_write_long( fp, &(thin_layer[i].n_index), one ) ;
	}

	fclose( fp ) ;
}


/* calculate the normalization constants that will be used to normalize the
variables. */

static void calc_norms( NORM *norm )

{

	/* The definition of charge_norm and mass_norm are somewhat
           arbitrary. I have chosen these definitions because they
           seem to work well. */

 	norm->charge = norm->time * norm->distance ;

	norm->mass = norm->time * norm->time * norm->charge / norm->distance ;

	norm->epsilon = norm->time * norm->time * 
	  norm->charge * norm->charge / 
	    ( norm->distance * norm->distance * norm->distance * norm->mass ) ;

	norm->mu = norm->distance * norm->mass /
	  ( norm->charge * norm->charge ) ;
	  
	norm->sigma = norm->time * norm->charge * norm->charge / 
	  ( norm->distance * norm->distance * norm->distance * norm->mass ) ;

	norm->p = norm->distance * norm->distance * norm->distance * 
	  norm->mass / ( norm->time * norm->time * norm->charge ) ;

	norm->e = norm->distance * norm->mass /
	  ( norm->time * norm->time * norm->charge ) ;

	norm->h = norm->charge / ( norm->time * norm->distance ) ;

        norm->current = norm->charge / norm->time ;

        norm->voltage = norm->distance * norm->distance * norm->mass /
          ( norm->time * norm->time * norm->charge ) ;

	fprintf( stdout, "Normalization for charge = %-11.4g\n", 
		norm->charge ) ;
	fprintf( stdout, "Normalization for mass = %-11.4g\n", norm->mass ) ;
	fprintf( stdout, "Normalization for dielectric permittivity = %-11.4g\n", norm->epsilon ) ;
	fprintf( stdout, "Normalization for magnetic permeability = %-11.4g\n", norm->mu ) ;
	fprintf( stdout, "Normalization for conductivity = %-11.4g\n", 
		norm->sigma ) ;
	fprintf( stdout, "Normalization for Hertz potential = %-11.4g\n", 
		norm->p ) ;
	fprintf( stdout, "Normalization for electic field intensity = %-11.4g\n", norm->e ) ;
	fprintf( stdout, "Normalization for magnetic field intensity = %-11.4g\n", norm->h ) ;
        fprintf( stdout, "Normalization for current = %-11.4g\n", 
		norm->current ) ;
        fprintf( stdout, "Normalization for voltage = %-11.4g\n\n", 
		norm->voltage
 ) ;

	fflush( stdout ) ;
}


/* normalize the variables. */

static void apply_norms( NORM norm, GRID *grid, GENERAL *general, COAX *coax, SCAN *scan, 
		 SCAN_LOC *scan_loc, SNAP *snap, EM_REGION *em_region, 
		 THIN_LAYER *thin_layer )

{
	long int i, j, k ;

	general->dt /= norm.time ;
	general->src_loc_r /= norm.distance ;
	general->src_loc_z /= norm.distance ;

	if( general->src_type[0] == 'e' ) {

		/* the vector "source" represents an electrical current 
		   density with units of amps/m^2 */

		for ( i = 0 ; i < general->nt ; ++ i )
		  general->source[i] /= ( norm.current / ( norm.distance * norm.distance ) ) ;
	}
	else if( general->src_type[0] == 'd' ) {

		/* the vector "source" represents a voltage with units of volts */

		for ( i = 0 ; i < general->nt ; ++ i )
		  general->source[i] /= norm.voltage ;

	}
	else if ( general->src_type[0] == 'c' ) {

		/* the vectors "source" and "source_c" represent voltage 
		   with units of volts */

		for ( i = 0 ; i < general->nt ; ++ i ) {
			general->source[i] /= norm.voltage ;
			general->source_c[i] /= norm.voltage ;
		}

	}

	for ( j = 0 ; j < grid->nr_space ; ++ j ) {
		grid->r_distance[j] /= norm.distance ;
		grid->r_step_size[j] /= norm.distance ;
	}

	for ( j = 0 ; j < grid->nz_space ; ++ j ) {
		grid->z_distance[j] /= norm.distance ;
		grid->z_step_size[j] /= norm.distance ;
	}

	for ( j = 0 ; j < grid->nr+2 ; ++ j )
	  grid->r[j] /= norm.distance ;

	for ( j = 0 ; j < grid->nz+2 ; ++ j )
	  grid->z[j] /= norm.distance ;

	for ( k = 0 ; k < general->n_coax ; ++ k ) {
		coax[k].dr /= norm.distance ;
		coax[k].dz /= norm.distance ;
		for( i = 0 ; i < coax[k].nr+2 ; ++ i )
		  coax[k].r[i] /= norm.distance ;
		for( j = 0 ; j < coax[k].nz+2 ; ++ j )
		  coax[k].z[j] /= norm.distance ;
		coax[k].sigma /= norm.sigma ;
		coax[k].mu = ( coax[k].mu_r * MU_0 ) / norm.mu ;
		coax[k].epsilon = ( coax[k].epsilon_r * EPSILON_0 ) / norm.epsilon ;
		coax[k].src_loc_z /= norm.distance ;
		coax[k].volt_cur_loc_z /= norm.distance ;
		coax[k].grid_attach_r /= norm.distance ;
		coax[k].grid_attach_z /= norm.distance ;
	}

	scan->time_start /= norm.time ;
	scan->time_step /= norm.time ;
	scan->time_end /= norm.time ;

	for ( i = 0 ; i < scan->n_scan ; ++ i ) {
		scan_loc[i].r /= norm.distance ;
		scan_loc[i].z /= norm.distance ;
	}

	snap->time_start /= norm.time ;
	snap->time_step /= norm.time ;
	snap->time_end /= norm.time ;

	snap->r_start /= norm.distance ;
	snap->r_step /= norm.distance ;
	snap->r_end /= norm.distance ;

	snap->z_start /= norm.distance ;
	snap->z_step /= norm.distance ;
	snap->z_end /= norm.distance ;

	for ( j = 0 ; j < general->n_em_region ; ++ j ) {
		for ( i = 0 ; i < em_region[j].n_vertices ; ++ i ) {
			em_region[j].r[i] /= norm.distance ;
			em_region[j].z[i] /= norm.distance ;
		}
		em_region[j].sigma /= norm.sigma ;
		em_region[j].epsilon = ( em_region[j].eps_r * EPSILON_0 ) / norm.epsilon ;
		em_region[j].mu = ( em_region[j].mu_r * MU_0 ) / norm.mu ;
	}

	for ( k = 0 ; k < general->n_thin_layer ; ++ k ) {
		thin_layer[k].inner_radius /= norm.distance ;
		thin_layer[k].radial_thickness /= norm.distance ;
		thin_layer[k].top /= norm.distance ;
		thin_layer[k]. bottom /= norm.distance ;
		thin_layer[k].sigma /= norm.sigma ;
		thin_layer[k].epsilon = ( thin_layer[k].eps_r * EPSILON_0 ) / norm.epsilon ;
		thin_layer[k].mu = ( thin_layer[k].mu_r * MU_0 ) / norm.mu ;
	}
}
