/*   

     propagate

     description: 

        Simulates the propagation of radar waves for the borehole geometry.

The FDTD grid:

        axis of (azimuthal) symmetry
        |
^       |
|       |
| j     |
|       |
        |
   nz   |         Er        Er        Er        Er           Er         Er     
        |         |         |         |         |            |          |      
nz-1    |    Ez - H  - Ez - H  - Ez - H  - Ez - H  -  ...  - H  - Ez -  H  - Ez
        |         |         |         |         |            |          |      
   nz-1 |         Er        Er        Er        Er           Er         Er 
        |         |         |         |         |            |          | 
nz-2    |    Ez - H  - Ez - H  - Ez - H  - Ez - H  -  ...  - H  - Ez -  H  - Ez
        |         |         |         |         |            |          |
   nz-2 |         Er        Er        Er        Er           Er         Er
        |         |         |         |         |            |          |
nz-3    |    Ez - H  - Ez - H  - Ez - H  - Ez - H  -  ...  - H  - Ez -  H  - Ez
        |         |         |         |         |            |          |     
        |		            	             	              
        |         .         .         .         .            .          .
        |         .         .         .         .            .          .
        |         .         .         .         .            .          .
        |		            	             	              
        |         |         |         |         |            |          | 
4       |    Ez - H  - Ez - H  - Ez - H  - Ez - H  -  ...  - H  - Ez -  H  - Ez
        |         |         |         |         |            |          |
   4    |         Er        Er        Er        Er           Er         Er
        |         |         |         |         |            |          |
3       |    Ez - H  - Ez - H  - Ez - H  - Ez - H  -  ...  - H  - Ez -  H  - Ez
        |         |         |         |         |            |          |
   3    |         Er        Er        Er        Er           Er         Er
        |         |         |         |         |            |          | 
2       |    Ez - H  - Ez - H  - Ez - H  - Ez - H  -  ...  - H  - Ez -  H  - Ez
        |         |         |         |         |            |          | 
   2    |         Er        Er        Er        Er           Er         Er 
        |         |         |         |         |            |          | 
1       |    Ez - H  - Ez - H  - Ez - H  - Ez - H  -  ...  - H  - Ez -  H  - Ez
        |         |         |         |         |            |          |
   1    |         Er        Er        Er        Er           Er         Er
        |         |         |         |         |            |          |
0       |    Ez - H  - Ez - H  - Ez - H  - Ez - H  -  ...  - H  - Ez -  H  - Ez
        |         |         |         |         |            |          |
   0    |         Er        Er        Er        Er           Er         Er
        | 
        |    0         1         2         3                      nr-1       nr
        |         0         1         2         3            nr-2       nr-1
        |
        |                         ----->
        |                           i
           
H: phi component of the magnetic intensity;  size of array is nr * nz
Ez: z component of the electric intensity;   size of array is (nr+1) * nz
Er: r component of the electric intensity;   size of array is nr * (nz+1)


axis of (azimuthal) symmetry
|
|                         ----->
|                           i
|
|    0         1         2         3                      nr-1       nr
|         0         1         2         3            nr-2       nr-1
|
|
|    Ez - H  - Ez - H  - Ez - H  - Ez - H  -  ...  - H  - Ez -  H  - Ez
|         |         |         |         |            |          |
|         Er        Er        Er        Er           Er         Er
| 
|         |         |         |         |     ...    |          |       |
r[0]      r[1]      r[2]      r[3]      r[4]         r[nr-1]    r[nr]   r[nr+1]
|
|---------|---------|---------|---------|     ...    |----------|-------|
|  dr[0]     dr[1]     dr[2]     dr[3]                 dr[nr-1]   dr[nr]
|
|    |---------|---------|---------|---------|     ...    |----------|
|       pr[0]     pr[1]     pr[2]     pr[3]                 pr[nr-1]    

r: location of vertix (r direction);              size of vector is nr+2
   r[0] is zero. This extra value is used because
   it makes the arrays for r similar to those for z. 
dr: distance between vertices (r directon);       size of vector is nr+1
pr: distance between cell centers (r directon);   size of vector is nr


        axis of (azimuthal) symmetry
        |
^       |
|       |
| j     |
|       |              - z[nz+1]  -                       
        |                         |      
   nz   |         Er              |dz[nz]     -
        |         |               |           |
nz-1    |    Ez - H    - z[nz]    -           |pz[nz-1]           
        |         |               |           |                 
   nz-1 |         Er              |dz[nz-1]   -           
        |         |               |           |           
nz-2    |    Ez - H    - z[nz-1]  -           |pz[nz-2]   
        |         |               |           |            
   nz-2 |         Er              |dz[nz-2]   -           
        |         |               |           |            
nz-3    |    Ez - H    - z[nz-2]  -           |pz[nz-3]   
        |         |                           |
        |                                     -
        |         .    		                          
        |         .    		                          
        |         .    		                          
        |	       		                          
        |         |    		                          
4       |    Ez - H    - z[5]     -
        |         |               |
   4    |         Er              |dz[4]      -                   
        |         |               |           |           
3       |    Ez - H    - z[4]     -           |pz[3]      
        |         |               |           |           
   3    |         Er              |dz[3]      -           
        |         |               |           |           
2       |    Ez - H    - z[3]     -           |pz[2]      
        |         |               |           |           
   2    |         Er              |dz[2]      -           
        |         |               |           |           
1       |    Ez - H    - z[2]     -           |pz[1]      
        |         |               |           |           
   1    |         Er              |dz[1]      -           
        |         |               |           |           
0       |    Ez - H    - z[1]     -           |pz[0]      
        |         |               |           |         
   0    |         Er              |dz[0]      -         
        | 	                  |                     
        |	       - z[0]     -                     

z: location of vertix (z direction);              size of vector is nz+2
dz: distance between vertices (z directon);       size of vector is nz+1
pz: distance between cell centers (z directon);   size of vector is nz

    required functions: calc_coef, setup_mur_bndr, 
           calc_mur_bndr, setup_imp_bndr, calc_imp_bndr, 
	   calc_cx_coef, calc_coax_bndr, write_scans, write_snaps,
           mem_alloc2, file_open2

    references: none


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "defs.h"

FILE *file_open2 ( char *name, char *mode ) ;
void *mem_alloc2 ( size_t nbytes, char array_name[] ) ;
long int file_write_float( FILE *fp, float *array, long int n ) ;

static void calc_coef( NORM norm, GENERAL general, GRID *grid, EM_REGION *em_region,
		      FIELD *h, FIELD *ez, FIELD *er ) ; 
static void setup_mur_bndr( GENERAL general, GRID *grid, EM_REGION *em_region,
			   BOUNDARY *mur_top, BOUNDARY *mur_rht, BOUNDARY *mur_bot ) ; 
static void calc_mur_bndr( GRID *grid, BOUNDARY *mur_top, BOUNDARY *mur_rht, 
			  BOUNDARY *mur_bot, FIELD *ez, FIELD *er ) ; 
static void setup_imp_bndr( EM_REGION *em_region, GRID *grid, float *imp_top,
			   float *imp_rht, float *imp_bot ) ; 
static void calc_imp_bndr( GRID *grid, float *imp_top, float *imp_rht, float *imp_bot,
			  FIELD *ez, FIELD *er, FIELD *h ) ;
static void calc_cc_coef( GENERAL general, NORM norm, COAX *coax, CC *cc ) ; 
static void calc_cc_bndr( GRID *grid, long int n_coax, COAX *coax, CC *cc,
			 FIELD *h, FIELD *er ) ;
static void calc_tl_coef( GENERAL general, GRID *grid, EM_REGION *em_region,
			 FIELD *ez, THIN_LAYER *thin_layer ) ; 
static void calc_tl_h( GENERAL general, GRID *grid, FIELD *ez, FIELD *er,
		      FIELD *h, THIN_LAYER *thin_layer ) ;
static void calc_tl_er( GENERAL general, GRID *grid,  FIELD *h, THIN_LAYER *thin_layer ) ;
static void write_scans( GRID *grid, SCAN scan, SCAN_LOC *scan_loc, NORM norm,
			FIELD *ez, FIELD *er, FIELD *h, FILE *fp_scan ) ;
static void write_snaps( GRID *grid, SNAP snap, NORM norm, FIELD *ez, FIELD *er,
			FIELD *h, FILE *fp_snap ) ;


void propagate ( NORM norm, GRID *grid, GENERAL general, COAX *coax, 
		SCAN scan, SCAN_LOC *scan_loc, SNAP snap, EM_REGION *em_region,
		THIN_LAYER *thin_layer )

{

	long int i, j, k, time_index ;

	float *imp_top, *imp_rht, *imp_bot, src_coef, i_coef ;

	double seconds, total_minutes ;

	clock_t start, finish ;

	FILE *fp_scan, *fp_snap, *fp_voltage, *fp_current ;

	CC *cc ;

	FIELD *h, *er, *ez ;

	BOUNDARY *mur_top, *mur_rht, *mur_bot ;

	/*
	  ---------------------------------------------------------------
	  Although most variables used in this C function are described 
	  in "fd_bh.c" and "defs.h", a few important variables are not. 
	  These are described in the following list:

	  imp_top,       Impedances along the top, right edge, and bottom
	  imp_rht, &     of the grid. The impedances are used to
          imp_bot        implement absorbing boundaries. The sizes of
                         these vectors are grid.nr, grid.nz, and
                         grid.nr.
	  src_coef       Coefficents used in the FDTD equations for a
	                 delta gap and electric conduction current, 
			 which are types of sources.
	  i_coef         Coefficent needed to calculate the electric
	                 conduction current.
	  ---------------------------------------------------------------
	  */


	/*
	  ---------------------------------------------------------------
	  Allocate arrays for grid distances --- dr, pr, dz, pz ---
	  and then compute them.
	  ---------------------------------------------------------------
	  */

	grid->dr = (float *) mem_alloc2( ( grid->nr + 1 ) * sizeof( float ), "dr" ) ;
	grid->pr = (float *) mem_alloc2( grid->nr * sizeof( float ), "pr" ) ;
	grid->dz = (float *) mem_alloc2( ( grid->nz + 1 ) * sizeof( float ), "dz" ) ;
	grid->pz = (float *) mem_alloc2( grid->nz * sizeof( float ), "pz" ) ;

	for ( i = 0 ; i < grid->nr+1 ; ++ i )
	  grid->dr[i] = grid->r[i+1] - grid->r[i] ;

	for ( i = 0 ; i < grid->nr ; ++ i )
	  grid->pr[i] = ( grid->dr[i+1] + grid->dr[i] ) / 2. ;

	for ( j = 0 ; j < grid->nz+1 ; ++ j )
	  grid->dz[j] = grid->z[j+1] - grid->z[j] ;

	for ( j = 0 ; j < grid->nz ; ++ j )
	  grid->pz[j] = ( grid->dz[j+1] + grid->dz[j] ) / 2. ;


	/*
	  ---------------------------------------------------------------
	  1. calculate the coefficients needed for the FD simulation.
	  (for h, er, and ez there are two coefficients each.)
	  2. set h, er, and ez themselves to zero. 
	  ---------------------------------------------------------------
	  */

	j = grid->nr * grid->nz ;
	h = (FIELD *) mem_alloc2( j * sizeof( FIELD ), "h" ) ;

	j = ( grid->nr + 1 ) * grid->nz ;
	ez = (FIELD *) mem_alloc2( j * sizeof( FIELD ), "ez" ) ;

	j = grid->nr * ( grid->nz + 1 ) ;
	er = (FIELD *) mem_alloc2( j * sizeof( FIELD ), "er" ) ;

	calc_coef ( norm, general, grid, em_region, h, ez, er ) ;

	/*
	  ---------------------------------------------------------------
	  calculate the arrays needed for the absorbing boundary conditions
	  ---------------------------------------------------------------
	  */

	if ( strcmp( grid->boundary, "mur" ) == 0 ) {

		/* boundary conditions based on Mur's 2nd order approximation */

		mur_top = (BOUNDARY *) 
		  mem_alloc2( grid->nr * sizeof( BOUNDARY ), "mur_top" ) ;

		mur_rht = (BOUNDARY *) 
		  mem_alloc2( grid->nz * sizeof( BOUNDARY ), "mur_rht" ) ;

		if ( grid->image_plane[0] != 'y' ) 
		  mur_bot = (BOUNDARY *) 
		    mem_alloc2( grid->nr * sizeof( BOUNDARY ), "mur_bot" ) ;

		setup_mur_bndr( general, grid, em_region, mur_top, mur_rht, mur_bot ) ;

	}
	else {

		/* boundary conditions using impedances */

		imp_top = (float *) 
		  mem_alloc2( grid->nr * sizeof( float ), "imp_top" ) ;

		imp_rht = (float *) 
		  mem_alloc2( grid->nz * sizeof( float ), "imp_rht" ) ;

		if ( grid->image_plane[0] != 'y' ) 
		  imp_bot = (float *) 
		    mem_alloc2( grid->nr * sizeof( float ), "imp_bot" ) ;

		setup_imp_bndr( em_region, grid, imp_top, imp_rht, imp_bot ) ;

	}

	if ( general.src_type[0] == 'e' ) {

		/*
		  -------------------------------------------------------
		  The source is an electrical conduction current.
		  Calculate the source and the current coefficients for 
		  the FDTD equation.
		  -------------------------------------------------------
		  */

		float epsilon, sigma ;

		i = general.src_loc_i ;

		j = general.src_loc_j ;

		epsilon = em_region[ INDEX_EZ(i,j) ].epsilon ;

		sigma = em_region[ INDEX_EZ(i,j) ].sigma ;

		src_coef = 
		  ( general.dt / epsilon ) / ( 1. + general.dt * sigma / ( 2. * epsilon ) ) ;

		i_coef = norm.current / ( norm.distance * norm.distance ) ;

	}

	else if ( general.src_type[0] == 'd' ) {

		/*
		  -------------------------------------------------------
		  The source is a delta gap.
		  Calculate the source and the current coefficients for 
		  the FDTD equations.
		  
		  For this source, the voltage is multiplied by
		  -1 and divided by delta z (the spacing between
		  the grid points in the z direction).
		  -------------------------------------------------------
		  */

		float f_temp ;

		i = general.src_loc_i ;

		j = general.src_loc_j ;

		f_temp = ( grid->z[j+2] - grid->z[j] ) / 2. ;

		src_coef = - 1. / f_temp ;

		i_coef = 2. * 3.14159 * grid->r[i+1] * norm.current ;

	}

	else if ( general.src_type[0] == 'c' ) {

		/*
		  -------------------------------------------------------
		  The model has coaxial cable(s). Allocate the arrays and 
		  calculate the coefficients that are needed to simulate 
		  propagation in the cables.
		  -------------------------------------------------------
		  */

		cc = (CC *) mem_alloc2( general.n_coax * sizeof( CC ), "cc" ) ;

		for( k = 0 ; k < general.n_coax ; ++ k ) {

			j = coax[k].nr * coax[k].nz ;

			cc[k].hc = (FIELD *) mem_alloc2( j * sizeof( FIELD ), "hc" ) ;

			j = ( coax[k].nr + 1 ) * coax[k].nz ;

			cc[k].ezc = (FIELD *) mem_alloc2( j * sizeof( FIELD ), "ezc" ) ;

			j = coax[k].nr * ( coax[k].nz + 1 ) ;

			cc[k].erc = (FIELD *) mem_alloc2( j * sizeof( FIELD ), "erc" ) ;

			if( coax[k].src_loc_z > 0. ) 
			  cc[k].src_coef = 
			    (float *) mem_alloc2( coax[k].nr * sizeof( float ), "src_coef" ) ;
 
		}

		/* calculate 
		   1. the radial distances to the grid points for H
		   2. the coefficients needed to simulate propagation.
		   3. the source coefficients (if the cable contains a source)
		   4. the coefficients for the voltage and current
		   5. the impedance of the boundary */

		calc_cc_coef( general, norm, coax, cc ) ;

	}

	if( general.thin_layer[0] == 'y' ) {

		/*
		  -------------------------------------------------------
		  The model has thin layer(s). Allocate the arrays and 
		  calculate the coefficients that are needed to simulate 
		  propagation in the thin layers.

		  For those Ez associated with the thin layers, the 
		  coefficients are recomputed.
		  -------------------------------------------------------
		  */

		for( k = 0 ; k < general.n_thin_layer ; ++ k ) {

			i = thin_layer[k].n_index ;

			thin_layer[k].erth = (FIELD *) 
			  mem_alloc2( (i+1)  * sizeof( FIELD ), "erth" ) ;

			thin_layer[k].hth = (FIELD *) 
			  mem_alloc2( i * sizeof( FIELD ), "hth" ) ;

			thin_layer[k].c3 = (float *) 
			  mem_alloc2( i * sizeof( FIELD ), "coef" ) ;

		}

		calc_tl_coef( general, grid, em_region, ez, thin_layer ) ;
	}

		/*
	  ---------------------------------------------------------------
	  For fp_scan and fp_snap:

	  If this code is compiled on a UNIX system, the second argument
	  must be "w".

	  If this code is compiled on a PC, the second argument
	  must be "wb".
	  ---------------------------------------------------------------
	  */

	if( scan.flag[0] == 'y' ) 
	  fp_scan = file_open2( scan.filename , "wb" ) ;

	if( snap.flag[0] == 'y' ) 
	  fp_snap = file_open2( snap.filename , "wb" ) ;


	fp_current = file_open2( general.current_filename , "w" ) ;

	if( general.src_type[0] == 'd' || general.src_type[0] == 'c' ) 
	  fp_voltage = file_open2( general.voltage_filename , "w" ) ;

        start = clock() ;

        total_minutes = 0. ;

        fprintf( stdout, "                          Number of\n" ) ;
        fprintf( stdout, "Iteration    Simulation   Seconds for          Total Number\n" ) ;
        fprintf( stdout, "Number       Time         Current Iteration    of Minutes\n" ) ;

	/*
	  ---------------------------------------------------------------
	  perform the FDTD simulation.

	  For each time step two sets of calculations are performed.
	  After half of the time step, the magnetic field is calculated.
	  After the entire time step, the electric field is calculated.
	  ---------------------------------------------------------------
	  */


	for ( time_index = 0 ; time_index < general.nt ; ++ time_index ) {

		/* Calculate the magnetic field on the grid. */

		for ( j = 0 ; j < grid->nz ; ++ j ) 
		  for ( i = 0 ; i < grid->nr ; ++ i )
		    H(i,j).val += 
		      H(i,j).c1 * ( EZ(i+1,j).val - EZ(i,j).val ) -
			H(i,j).c2 * ( ER(i,j+1).val - ER(i,j).val ) ;

		/* Calcuate the magnetic field at those grid points associated
		   with thin layers. */

		if( general.thin_layer[0] == 'y' )
		  calc_tl_h( general, grid, ez, er, h, thin_layer ) ;

		/* If there are coaxial cables, calculate the magnetic field in them and
		   apply the source. */

		if ( general.src_type[0] == 'c' ) {

			float sign ;

			for( k = 0 ; k < general.n_coax ; ++ k )
			  for ( j = 0 ; j < coax[k].nz ; ++ j ) 
			    for ( i = 0 ; i < coax[k].nr ; ++ i )
			      HC(i,j).val += 
				HC(i,j).c1 * ( EZC(i+1,j).val - EZC(i,j).val ) -
				  HC(i,j).c2 * ( ERC(i,j+1).val - ERC(i,j).val ) ;

			k = general.coax_source ;

			j = coax[k].src_loc_j ;

			if( strcmp( coax[k].coax_attach, "top" ) == 0 )
			  sign = 1. ;
			else
			  sign = -1. ;

			for ( i = 0 ; i < coax[k].nr ; ++ i )
			  HC(i,j).val += 
			    sign * HC(i,j).c2 * general.source[time_index] * cc[k].src_coef[i] ;

		}

		/* Calculate the electric field on the grid. 
		   To calculate Ez for i = 0, both the value of H on the z axis and the
		   radial distance on the z axis are used. Because of the model, both
		   values are 0. Thus the FDTD equation is slightly different. */

		i = 0 ;

		for ( j = 0 ; j < grid->nz ; ++ j ) 
		    EZ(i,j).val = EZ(i,j).c1 * EZ(i,j).val +
		      EZ(i,j).c2 * ( grid->r[i+1]*H(i,j).val ) ;

		for ( j = 0 ; j < grid->nz ; ++ j ) 
		  for ( i = 1 ; i < grid->nr ; ++ i ) 
		    EZ(i,j).val = EZ(i,j).c1 * EZ(i,j).val +
		      EZ(i,j).c2 * ( grid->r[i+1]*H(i,j).val - grid->r[i]*H(i-1,j).val ) ;

		for ( j = 1 ; j < grid->nz ; ++ j ) 
		  for ( i = 0 ; i < grid->nr ; ++ i )
		    ER(i,j).val = ER(i,j).c1 * ER(i,j).val -
		      ER(i,j).c2 * ( H(i,j).val - H(i,j-1).val ) ;

		/* Calcuate the electric field (Er) inside the thin layers. */

		if( general.thin_layer[0] == 'y' )
		  calc_tl_er( general, grid, h, thin_layer ) ;

		/* If there are coaxial cables, calculate the electric field in them
		   and apply the source. */

		if ( general.src_type[0] == 'c' ) {

			for( k = 0 ; k < general.n_coax ; ++ k ) 
			  for ( j = 0 ; j < coax[k].nz ; ++ j )
			    for ( i = 1 ; i < coax[k].nr ; ++ i )
			      EZC(i,j).val = EZC(i,j).c1 * EZC(i,j).val +
				EZC(i,j).c2 * ( coax[k].r[i+1]*HC(i,j).val - coax[k].r[i]*HC(i-1,j).val ) ;

			for( k = 0 ; k < general.n_coax ; ++ k )
			  for ( j = 1 ; j < coax[k].nz ; ++ j ) 
			    for ( i = 0 ; i < coax[k].nr ; ++ i )
			      ERC(i,j).val = ERC(i,j).c1 * ERC(i,j).val -
				ERC(i,j).c2 * ( HC(i,j).val - HC(i,j-1).val ) ;

			k = general.coax_source ;

			j = coax[k].src_loc_j ;

			for ( i = 0 ; i < coax[k].nr ; ++ i )
			  ERC(i,j).val += ERC(i,j).c2 * general.source_c[time_index] * 
			    cc[k].src_coef[i] / cc[k].impedance ;

		}

		/* If the source is an electric conduction current density, apply it. */

		if ( general.src_type[0] == 'e' ) {

			i = general.src_loc_i ;

			j = general.src_loc_j ;

			EZ(i,j).val -= src_coef * general.source[time_index] ;

		}

		/* if the source is a delta gap, apply it. */

		if ( general.src_type[0] == 'd' ) {

			i = general.src_loc_i ;

			j = general.src_loc_j ;

			EZ(i,j).val += src_coef * general.source[time_index] ;

		}

		/* calculate the electric field at the boundaries. */

		if ( strcmp( grid->boundary, "mur" ) == 0 ) 
		  calc_mur_bndr( grid, mur_top, mur_rht, mur_bot, ez, er ) ;
		else
		  calc_imp_bndr( grid, imp_top, imp_rht, imp_bot, ez, er, h ) ;

		/* if there are coaxial cables, calculate the electric field 
		   at the boundaries. merge the fields in the cables and the 
		   main grid. */

		if ( general.src_type[0] == 'c' ) 
		  calc_cc_bndr( grid, general.n_coax, coax, cc, h, er ) ;

		/* print the voltage and the current. */

		if ( general.src_type[0] == 'e' ) {

			fprintf( fp_current, "%-13.6e  %-13.6e\n",
				time_index * general.dt * norm.time,
				general.source[time_index] * i_coef ) ;

			fflush( fp_current ) ;

		}
		else if ( general.src_type[0] == 'd' ) {

			fprintf( fp_voltage, "%-13.6e  %-13.6e\n",
				time_index * general.dt * norm.time,
				general.source[time_index] * norm.voltage ) ;

			fflush( fp_voltage ) ; 

			i = general.src_loc_i ;

			j = general.src_loc_j ;

			fprintf( fp_current, "%-13.6e  %-13.6e\n",
				time_index * general.dt * norm.time,
				H(i,j).val * i_coef ) ;

			fflush( fp_current ) ; 
		}

		else if ( general.src_type[0] == 'c' ) {

			fprintf( fp_voltage, "%-13.6e",
				time_index * general.dt * norm.time ) ;

			fprintf( fp_current, "%-13.6e",
				time_index * general.dt * norm.time ) ;

			for( k = 0 ; k < general.n_coax ; ++ k ) {

				j = coax[k].volt_cur_loc_j ;

				fprintf( fp_voltage, "  %-13.6e", 
					ERC(0,j).val * cc[k].v_coef ) ;

				fprintf( fp_current, "  %-13.6e", 
					HC(0,j).val * cc[k].i_coef ) ;
			}

			fprintf( fp_voltage, "\n" ) ;
			fprintf( fp_current, "\n" ) ;

			fflush( fp_voltage ) ; 
			fflush( fp_current ) ; 

		}

		/* write the scan for the current time step. */

		if ( scan.flag[0] == 'y' && 
		    scan.time_index_start <= time_index  &&
		    time_index <= scan.time_index_end &&
		    ( time_index - scan.time_index_start ) % 
		    scan.time_index_step == 0 ) 
		  write_scans( grid, scan, scan_loc, norm, ez, er, h, fp_scan ) ;

		/* write the snapshot for the current time step. */

		if ( snap.flag[0] == 'y' &&
		    time_index >= snap.time_index_start && 
		    time_index <= snap.time_index_end &&
		    ( time_index - snap.time_index_start ) % 
		    snap.time_index_step == 0 ) 
		  write_snaps( grid, snap, norm, ez, er, h, fp_snap ) ;

		/* print information about the timing */

                finish = clock() ;

                seconds = ( (double) ( finish - start ) ) / ( (double) CLOCKS_PER_SEC ) ;

                total_minutes += seconds / 60. ;

		if( time_index % 10 == 0 || time_index == general.nt-1 )
		  fprintf(stdout,"%-6ld       %-10.3e   %-10.2e           %-10.2f\n",
			  time_index, time_index * general.dt * norm.time,
			  seconds, total_minutes ) ;

                fflush(stdout) ;

                start = finish ;

	}

	fprintf(stdout,"\nFinished Simulation\n" ) ;
	fflush(stdout) ;

	if( scan.flag[0] == 'y' ) 
	  fclose( fp_scan ) ;

	if( snap.flag[0] == 'y' ) 
	  fclose( fp_snap ) ;

	fclose( fp_current ) ;

	if( general.src_type[0] == 'd' || general.src_type[0] == 'c' )
	  fclose( fp_voltage ) ;

	free( grid->dr ) ;
	free( grid->pr ) ;
	free( grid->dz ) ;
	free( grid->pz ) ;

	free( h ) ;
	free( ez ) ;
	free( er ) ;

	if ( strcmp( grid->boundary, "mur" ) == 0 ) {
		free( mur_top ) ;
		free( mur_rht ) ;
		if ( grid->image_plane[0] != 'y' ) 
		  free( mur_bot ) ;
	}
	else {
		free( imp_top ) ;
		free( imp_rht ) ;
		if ( grid->image_plane[0] != 'y' ) 
		  free( imp_bot ) ;
	}

	for( k = 0 ; k < general.n_coax ; ++ k ) {
		free( cc[k].hc ) ;
		free( cc[k].ezc ) ;
		free( cc[k].erc ) ;
		if( coax[k].src_loc_z > 0. )
		  free( cc[k].src_coef ) ;
	}

	if ( general.src_type[0] == 'c' ) 
	  free( cc ) ;

	for( k = 0 ; k < general.n_thin_layer ; ++ k ) {
		free( thin_layer[k].erth ) ;
		free( thin_layer[k].hth ) ;
		free( thin_layer[k].c3 ) ;
	}
}

/* Calculates the coefficients used in the FDTD equations. For
those grid points that are either inside or at the boundary of metallic
region, the coefficients are set to zero. */

static void calc_coef ( NORM norm, GENERAL general, GRID *grid, EM_REGION *em_region, 
		FIELD *h, FIELD *ez, FIELD *er ) 

{

	long int i, j, k ;

	float f_temp, sigma, epsilon, mu ;

	/* Ez 
	   To calculate Ez for i = 0, both the value of H on the z axis and the
	   radial distance on the z axis are used. Because of the model, both
	   values are 0. Thus the FDTD equation is slightly different. */

	i = 0 ;

	for ( j = 0 ; j < grid->nz ; ++ j ) {
		EZ(i,j).val = 0. ;
		sigma = em_region[ INDEX_EZ(i,j) ].sigma ;
		epsilon = em_region[ INDEX_EZ(i,j) ].epsilon ;
		f_temp = general.dt * sigma / ( 2. * epsilon ) ;

		if( SIGMA_METAL <= sigma * norm.sigma ) {
			EZ(i,j).c1 = 0. ;
			EZ(i,j).c2 = 0. ;
		}
		else {
			EZ(i,j).c1 = ( 1. - f_temp ) / ( 1. + f_temp ) ;
			EZ(i,j).c2 = 
			  ( 2. * general.dt / epsilon ) /
			    ( ( 1. + f_temp ) * ( grid->r[i+1] * grid->r[i+1] ) ) ;
		}
	}

	for ( j = 0 ; j < grid->nz ; ++ j )
	  for ( i = 1 ; i < grid->nr+1 ; ++ i ) {
		  EZ(i,j).val = 0. ;
		  sigma = em_region[ INDEX_EZ(i,j) ].sigma ;
		  epsilon = em_region[ INDEX_EZ(i,j) ].epsilon ;
		  f_temp = general.dt * sigma / ( 2. * epsilon ) ;

		  if( SIGMA_METAL <= sigma * norm.sigma ) {
			  EZ(i,j).c1 = 0. ;
			  EZ(i,j).c2 = 0. ;
		  }
		  else {
			  EZ(i,j).c1 = ( 1. - f_temp ) / ( 1. + f_temp ) ;
			  EZ(i,j).c2 = 
			    ( 2. * general.dt / epsilon ) /
			      ( ( 1. + f_temp ) * 
			       ( grid->r[i+1] * grid->r[i+1] - grid->r[i] * grid->r[i] ) ) ;
		  }
	  }

	/* Er */

	for ( j = 0 ; j < grid->nz+1 ; ++ j ) 
	  for ( i = 0 ; i < grid->nr ; ++ i ) {
		  ER(i,j).val = 0. ;
		  sigma = em_region[ INDEX_ER(i,j) ].sigma ;
		  epsilon = em_region[ INDEX_ER(i,j) ].epsilon ;
		  f_temp = general.dt * sigma / ( 2. * epsilon ) ;

		  if( SIGMA_METAL <= sigma * norm.sigma ) {
			  ER(i,j).c1 = 0. ;
			  ER(i,j).c2 = 0. ;
		  }
		  else {
			  ER(i,j).c1 = ( 1. - f_temp ) / ( 1. + f_temp ) ;
			  ER(i,j).c2 = 
			    general.dt / ( epsilon * grid->dz[j] ) / ( 1. + f_temp ) ;
		  }
	  }

	/* H */

	for ( j = 0 ; j < grid->nz ; ++ j )
	  for ( i = 0 ; i < grid->nr ; ++ i ) {
		  H(i,j).val = 0. ;
		  mu = em_region[ INDEX_H(i,j) ].mu ;
		  sigma = em_region[ INDEX_H(i,j) ].sigma ;

		  if( SIGMA_METAL <= sigma * norm.sigma ) {
			  H(i,j).c1 = 0. ;
			  H(i,j).c2 = 0. ;
		  }
		  else {
			  H(i,j).c1 = general.dt / ( mu * grid->pr[i] ) ;
			  H(i,j).c2 = general.dt / ( mu * grid->pz[j] ) ;
		  }
	  }
		
}

/* Calculates the coefficients used for the Mur boundary. */

static void setup_mur_bndr( GENERAL general, GRID *grid, EM_REGION *em_region, 
		    BOUNDARY *mur_top, BOUNDARY *mur_rht, BOUNDARY *mur_bot )

{

	long int i, j ;

	float vel, epsilon, mu ;

	j = grid->nz ;

	for( i = 0 ; i < grid->nr ; ++ i ) {

		mur_top[i].b_now = mur_top[i].b1_now = 
		  mur_top[i].b_prev = mur_top[i].b1_prev = 0. ;

		epsilon = em_region[ INDEX_ER(i,j) ].epsilon ;
		
		mu = em_region[ INDEX_H(i,j-1) ].mu ;
		
		vel = 1. / sqrt( mu * epsilon ) ;

		mur_top[i].coef_1 = 
		  ( vel * general.dt - grid->pz[grid->nz-1] ) / 
		    ( vel * general.dt + grid->pz[grid->nz-1] ) ;

		mur_top[i].coef_2 = 
		  2. * grid->pz[grid->nz-1] / ( vel * general.dt + grid->pz[grid->nz-1] ) ;

		mur_top[i].coef_3 = 
		  vel * vel * general.dt * general.dt * grid->pz[grid->nz-1] / 
		    ( 2. * grid->pr[i] * grid->pr[i] * 
		     ( vel * general.dt + grid->pz[grid->nz-1] ) ) ;
	}

	i = grid->nr ;

	for( j = 0 ; j < grid->nz ; ++ j ) {

		mur_rht[j].b_now = mur_rht[j].b1_now = 
		  mur_rht[j].b_prev = mur_rht[j].b1_prev = 0. ;

		epsilon = em_region[ INDEX_EZ(i,j) ].epsilon ;

		mu = em_region[ INDEX_H(i-1,j) ].mu ;

		vel = 1. / sqrt( mu * epsilon ) ;

		mur_rht[j].coef_1 = 
		  ( vel * general.dt - grid->pr[grid->nr-1] ) / 
		    ( vel * general.dt + grid->pr[grid->nr-1] ) ;
		mur_rht[j].coef_2 = 
		  2. * grid->pr[grid->nr-1] / ( vel * general.dt + grid->pr[grid->nr-1] ) ;
		mur_rht[j].coef_3 = 
		  vel * vel * general.dt * general.dt * grid->pr[grid->nr-1] / 
		    ( 2. * grid->pz[j] * grid->pz[j] * 
		     ( vel * general.dt + grid->pr[grid->nr-1] ) ) ;
	}


	if ( grid->image_plane[0] != 'y' ) {

		j = 0 ;

		for( i = 0 ; i < grid->nr ; ++ i ) {

			mur_bot[i].b_now = mur_bot[i].b1_now = 
			  mur_bot[i].b_prev = mur_bot[i].b1_prev = 0. ;

			epsilon = em_region[ INDEX_ER(i,j) ].epsilon ;

			mu = em_region[ INDEX_H(i,j) ].mu ;

			vel = 1. / sqrt( mu * epsilon ) ;

			mur_bot[i].coef_1 = 
			  ( vel * general.dt - grid->pz[0] ) / 
			    ( vel * general.dt + grid->pz[0] ) ;
			mur_bot[i].coef_2 = 
			  2. * grid->pz[0] / ( vel * general.dt + grid->pz[0] ) ;
			mur_bot[i].coef_3 = 
			  vel * vel * general.dt * general.dt * grid->pz[0] / 
			    ( 2. * grid->pr[i] * grid->pr[i] * 
			     ( vel * general.dt + grid->pz[0] ) ) ;
		}

	}

}

/* Implements the Mur boundary during the simulation. */

static void calc_mur_bndr( GRID *grid, BOUNDARY *mur_top, BOUNDARY *mur_rht, 
			  BOUNDARY *mur_bot, FIELD *ez, FIELD *er )

{

	long int i, j ;

	/* right boundary */

	i = grid->nr ;

	/* first order */
	for ( j = 0 ; j < grid->nz ; ++ j ) 
	  EZ(i,j).val = - mur_rht[j].b1_prev + 
	    mur_rht[j].coef_1 * ( EZ(i-1,j).val + mur_rht[j].b_prev ) +
	      mur_rht[j].coef_2 * ( mur_rht[j].b_now + mur_rht[j].b1_now ) ;

	/* second order */
	for ( j = 1 ; j < grid->nz-1 ; ++ j ) 
	  EZ(i,j).val += mur_rht[j].coef_3 * 
	    ( mur_rht[j+1].b_now - 2. * mur_rht[j].b_now + mur_rht[j-1].b_now +
	     mur_rht[j+1].b1_now - 2. * mur_rht[j].b1_now + mur_rht[j-1].b1_now ) ;

	/* update */
	for ( j = 0 ; j < grid->nz ; ++ j ) {
		mur_rht[j].b_prev = mur_rht[j].b_now ;
		mur_rht[j].b_now = EZ(i,j).val ;
		mur_rht[j].b1_prev = mur_rht[j].b1_now ;
		mur_rht[j].b1_now = EZ(i-1,j).val ;
	}

	/* bottom boundary */

	j = 0 ;

	if ( grid->image_plane[0] == 'y' )
	  for ( i = 0 ; i < grid->nr ; ++ i )
	    ER(i,j).val = 0. ;
	else {

		/* first order */
		for ( i = 0 ; i < grid->nr ; ++ i ) 
		  ER(i,j).val = - mur_bot[i].b1_prev + 
		    mur_bot[i].coef_1 * ( ER(i,j+1).val + mur_bot[i].b_prev ) +
		      mur_bot[i].coef_2 * ( mur_bot[i].b_now + mur_bot[i].b1_now ) ;

		/* second order */
		for (i = 1 ; i < grid->nr-1 ; ++ i )
		  ER(i,j).val += mur_bot[i].coef_3 * 
		    ( mur_bot[i+1].b_now - 2. * mur_bot[i].b_now + 
		     mur_bot[i-1].b_now + mur_bot[i+1].b1_now 
		     - 2. * mur_bot[i].b1_now + mur_bot[i-1].b1_now ) ;

		/* update */
		for ( i = 0 ; i < grid->nr ; ++ i ) {
			mur_bot[i].b_prev = mur_bot[i].b_now ;
			mur_bot[i].b_now = ER(i,j).val ;
			mur_bot[i].b1_prev = mur_bot[i].b1_now ;
			mur_bot[i].b1_now = ER(i,j+1).val ;
		}

	}

	/* top boundary */

	j = grid->nz ;

	/* first order */
	for ( i = 0 ; i < grid->nr ; ++ i ) 
	  ER(i,j).val = - mur_top[i].b1_prev + 
	    mur_top[i].coef_1 * ( ER(i,j-1).val + mur_top[i].b_prev ) +
	      mur_top[i].coef_2 * ( mur_top[i].b_now + mur_top[i].b1_now ) ;

	/* second order */
	for (i = 1 ; i < grid->nr-1 ; ++ i )
	  ER(i,j).val += mur_top[i].coef_3 * 
	    ( mur_top[i+1].b_now - 2. * mur_top[i].b_now + mur_top[i-1].b_now +
	     mur_top[i+1].b1_now - 2. * mur_top[i].b1_now + mur_top[i-1].b1_now ) ;

	/* update */
	for ( i = 0 ; i < grid->nr ; ++ i ) {
		mur_top[i].b_prev = mur_top[i].b_now ;
		mur_top[i].b_now = ER(i,j).val ;
		mur_top[i].b1_prev = mur_top[i].b1_now ;
		mur_top[i].b1_now = ER(i,j-1).val ;
	}

}

/* Calculates the coefficients used for the impedance boundary. */

static void setup_imp_bndr( EM_REGION *em_region, GRID *grid, float *imp_top, 
		    float *imp_rht, float *imp_bot )

{

	long int i, j ;

	float epsilon, mu ;

	/* top boundary */

	j = grid->nz ;

	for ( i = 0 ; i < grid->nr ; ++ i ) {

		epsilon = em_region[ INDEX_ER(i,j) ].epsilon ;

		mu = em_region[ INDEX_H(i,j-1) ].mu ;

		imp_top[i] =  sqrt( mu / epsilon ) ; 

	}

	/* right boundary */

	i = grid->nr ;

	for ( j = 0 ; j < grid->nz ; ++ j ) {

		epsilon = em_region[ INDEX_EZ(i,j) ].epsilon ;

		mu = em_region[ INDEX_H(i-1,j) ].mu ;

		imp_rht[j] = sqrt( mu / epsilon ) ;

	}

	/* bottom boundary */

	j = 0 ;

	if ( grid->image_plane[0] != 'y' )
	  for ( i = 0 ; i < grid->nr ; ++ i ) {

		  epsilon = em_region[ INDEX_ER(i,j) ].epsilon ;

		  mu = em_region[ INDEX_H(i,j) ].mu ;

		  imp_bot[i] = sqrt( mu / epsilon ) ;

	  }

}

/* Implements the impedance boundary during the simulation. */

static void calc_imp_bndr( GRID *grid, float *imp_top, float *imp_rht, float *imp_bot, 
		   FIELD *ez, FIELD *er, FIELD *h )

{

	long int i, j ;

	/* top boundary */

	j = grid->nz ;

	for ( i = 0 ; i < grid->nr ; ++ i )
	  ER(i,j).val = H(i,j-1).val * imp_top[i] ;

	/* right boundary */

	i = grid->nr ;

	for ( j = 0 ; j < grid->nz ; ++ j ) 
	  EZ(i,j).val = - H(i-1,j).val * imp_rht[j] ;

	/* bottom boundary */

	j = 0 ;

	if ( grid->image_plane[0] == 'y' )
	  for ( i = 0 ; i < grid->nr ; ++ i )
	    ER(i,j).val = 0. ;
	else
	  for ( i = 0 ; i < grid->nr ; ++ i )
	    ER(i,j).val = - H(i,j).val * imp_bot[i] ;

}

static void calc_cc_coef( GENERAL general, NORM norm, COAX *coax, CC *cc ) 

{

	long int i, j, k ;

	float f_temp ;

	for( k = 0 ; k < general.n_coax ; ++k ) {

		/* calculate the coefficients needed to simulate propagation. */

		f_temp = general.dt * coax[k].sigma / ( 2. * coax[k].epsilon ) ;

		for ( j = 0 ; j < coax[k].nz ; ++ j )
		  for ( i = 1 ; i < coax[k].nr ; ++ i ) {
			  EZC(i,j).val = 0. ;
			  EZC(i,j).c1 = ( 1. - f_temp ) / ( 1. + f_temp ) ;
			  EZC(i,j).c2 = 
			    ( 2. * general.dt / coax[k].epsilon ) /
			      ( ( 1. + f_temp ) * 
			       ( coax[k].r[i+1] * coax[k].r[i+1] - coax[k].r[i] * coax[k].r[i] ) ) ;
		  }

		for ( j = 0 ; j < coax[k].nz ; ++ j ) {

			i = 0 ;
			EZC(i,j).val = EZC(i,j).c1 = EZC(i,j).c2 = 0. ;

			i = coax[k].nr ;
			EZC(i,j).val = EZC(i,j).c1 = EZC(i,j).c2 = 0. ;

		}

		for ( j = 0 ; j < coax[k].nz+1 ; ++ j ) 
		  for ( i = 0 ; i < coax[k].nr ; ++ i ) {
			  ERC(i,j).val = 0. ;
			  ERC(i,j).c1 = ( 1. - f_temp ) / ( 1. + f_temp ) ;
			  ERC(i,j).c2 = 
			    ( general.dt / ( coax[k].epsilon * coax[k].dz ) ) / 
			      ( 1. + f_temp ) ;
		  }

		for ( j = 0 ; j < coax[k].nz ; ++ j )
		  for ( i = 0 ; i < coax[k].nr ; ++ i ) {
			  HC(i,j).val = 0. ;
			  HC(i,j).c1 = general.dt / ( coax[k].mu * coax[k].dr ) ;
			  HC(i,j).c2 = general.dt / ( coax[k].mu * coax[k].dz ) ;
		  }


		/* the argument of the logarithm is the ratio
		   of the radii of the outer and the inner
		   conductors */

		i = coax[k].nr ;

		f_temp = log( ( coax[k].r[i] + coax[k].dr/2. ) / 
			     ( coax[k].r[1] - coax[k].dr/2. ) ) ;

		/* calculate the source coefficients (if the cable contains a source) */

		if( coax[k].src_loc_z > 0. ) 
		  for ( i = 0 ; i < coax[k].nr ; ++ i )
		    cc[k].src_coef[i] = 1. / ( coax[k].r[i+1] * f_temp ) ;

		/* calculate the coefficients for the voltage and current */

		cc[k].v_coef = coax[k].r[1] * f_temp * norm.voltage ;

		cc[k].i_coef = 2. * 3.14159 * coax[k].r[1] * norm.current ;

		/* calculate the impedance of the coaxial cable */

		cc[k].impedance = sqrt( coax[k].mu / coax[k].epsilon ) ;

	}

}

/* in the coaxial cables, calculate the electric field
at the boundaries. merge the fields in the cables and the
main grid. */

static void calc_cc_bndr( GRID *grid, long int n_coax, COAX *coax, CC *cc, 
		  FIELD *h, FIELD *er )

{

	long int i, j, k, si, sj ;

	for( k = 0 ; k < n_coax ; ++ k ) {

		/* Ez on the the left and right boundaries is initially
		   set to zero and is never updated in the FDTD equations. 
		   Thus, it should always be zero. Just to be sure ...
		   */

		/* left boundary (inner conductor) */

		i = 0 ;

		for( j = 0 ; j < coax[k].nz ; ++ j )
		  EZC(i,j).val = 0. ;

		/* right boundary (outer conductor) */

		i = coax[k].nr ;

		for( j = 0 ; j < coax[k].nz ; ++ j )
		  EZC(i,j).val = 0. ;

		if( strcmp( coax[k].coax_attach, "top" ) == 0 ) {

			/* top boundary --- merged with main grid */

			j = coax[k].nz ;

			si = coax[k].grid_attach_i ;
			sj = coax[k].grid_attach_j ;

			for ( i = 0 ; i < coax[k].nr ; ++ i ) {
				ERC(i,j).val = ERC(i,j).c1 * ERC(i,j).val - 
				  ERC(i,j).c2 * ( H(i+si,sj).val - HC(i,j-1).val ) ;
			
				ER(i+si,sj).val = ERC(i,j).val ;
			}

			/* bottom (absorbing) boundary */

			j = 0 ;

			for ( i = 0 ; i < coax[k].nr ; ++ i ) 
			  ERC(i,j).val = - HC(i,j).val * cc[k].impedance ;

		}
		else {

			/* top (absorbing) boundary */

			j = coax[k].nz ;

			for ( i = 0 ; i < coax[k].nr ; ++ i ) 
			  ERC(i,j).val = HC(i,j-1).val * cc[k].impedance ;

			/* bottom boundary --- merged with main grid */

			j = 0 ;

			si = coax[k].grid_attach_i ;
			sj = coax[k].grid_attach_j ;

			for ( i = 0 ; i < coax[k].nr ; ++ i ) {
				ERC(i,j).val = ERC(i,j).c1 * ERC(i,j).val - 
				  ERC(i,j).c2 * ( HC(i,j).val - H(i+si,sj-1).val ) ;
			
				ER(i+si,sj).val = ERC(i,j).val ;
			}

		}

	}
}

/* Calculates the coefficients needed for the thin layers. */

static void calc_tl_coef( GENERAL general, GRID *grid, EM_REGION *em_region,
			 FIELD *ez, THIN_LAYER *thin_layer )

{
	long int i, j, j1, k ;

	float r1, r2, weight, sigma, epsilon, mu, f_temp ;

	for( k = 0 ; k < general.n_thin_layer ; ++ k ) {

		/* Ez */

		i = thin_layer[k].i_index_ez ;

		r1 = thin_layer[k].inner_radius ;
		r2 = r1 + thin_layer[k].radial_thickness ;

		weight = ( r2 * r2 - r1 * r1 ) /
		  ( grid->r[i+1] * grid->r[i+1] - grid->r[i] * grid->r[i] ) ;

		for( j1 = 0 ; j1 < thin_layer[k].n_index ; ++ j1 ) {

			j = thin_layer[k].j_index + j1 ;

			sigma = em_region[ INDEX_EZ(i,j) ].sigma * ( 1. - weight ) +
			  thin_layer[k].sigma * weight ;

			epsilon = em_region[ INDEX_EZ(i,j) ].epsilon * ( 1. - weight ) +
			  thin_layer[k].epsilon * weight ;

			f_temp = general.dt * sigma / ( 2. * epsilon ) ;

			EZ(i,j).val = 0. ;

			EZ(i,j).c1 = ( 1. - f_temp ) / ( 1. + f_temp ) ;

			EZ(i,j).c2 =
			  ( 2. * general.dt / epsilon ) /
			    ( ( 1. + f_temp ) *
			     ( grid->r[i+1] * grid->r[i+1] - grid->r[i] * grid->r[i] ) ) ;

		}

		/* Er in the thin layer */

		sigma = thin_layer[k].sigma ;

		epsilon = thin_layer[k].epsilon ;

		f_temp = general.dt * sigma / ( 2. * epsilon ) ;

		for( j1 = 0 ; j1 < thin_layer[k].n_index+1 ; ++ j1 ) {

			j = thin_layer[k].j_index + j1 ;

			thin_layer[k].erth[j1].val = 0. ;

			thin_layer[k].erth[j1].c1 = ( 1. - f_temp ) / ( 1. + f_temp ) ;

			thin_layer[k].erth[j1].c2 = 
			  general.dt / ( epsilon * grid->dz[j] ) / ( 1. + f_temp ) ;

		}

		/* H */

		i = thin_layer[k].i_index_h ;

		weight = thin_layer[k].radial_thickness / grid->pr[i] ; 

		for( j1 = 0 ; j1 < thin_layer[k].n_index ; ++ j1 ) {

			j = thin_layer[k].j_index + j1 ;

			mu = em_region[ INDEX_H(i,j) ].mu * ( 1. - weight ) +
			  thin_layer[k].sigma * weight ;

			thin_layer[k].hth[j1].val = 0. ;

			thin_layer[k].hth[j1].c1 = general.dt / ( mu * grid->pr[i] ) ;

			thin_layer[k].hth[j1].c2 = general.dt *
			  ( 1. - weight ) / ( mu * grid->pz[j] ) ;

			thin_layer[k].c3[j1] = general.dt * weight /
			  ( mu * grid->pz[j] ) ;

		}

	}

}

/* Calculate the magnetic field at those grid points associated 
with the thin layers. */

static void calc_tl_h( GENERAL general, GRID *grid, FIELD *ez, FIELD *er,
		      FIELD *h, THIN_LAYER *thin_layer )

{
	long int i, j, j1, k ;

	for( k = 0 ; k < general.n_thin_layer ; ++ k ) {

		i = thin_layer[k].i_index_h ;

		for( j1 = 0 ; j1 < thin_layer[k].n_index ; ++ j1 ) {

			j = thin_layer[k].j_index + j1 ;

			thin_layer[k].hth[j1].val += 
			  thin_layer[k].hth[j1].c1 * ( EZ(i+1,j).val - EZ(i,j).val ) -
			    thin_layer[k].hth[j1].c2 * ( ER(i,j+1).val - ER(i,j).val ) -
			      thin_layer[k].c3[j1] * ( thin_layer[k].erth[j1+1].val -
						      thin_layer[k].erth[j1].val ) ;

			H(i,j).val = thin_layer[k].hth[j1].val ;
		}
	}
}

/* Calculate the electric field (Er) inside the thin layers. */

static void calc_tl_er( GENERAL general, GRID *grid, FIELD *h, THIN_LAYER *thin_layer )

{
	long int i, j, j1, k ;

	for( k = 0 ; k < general.n_thin_layer ; ++ k ) {

		i = thin_layer[k].i_index_h ;

		for( j1 = 0 ; j1 < thin_layer[k].n_index+1 ; ++ j1 ) {

			j = thin_layer[k].j_index + j1 ;

			thin_layer[k].erth[j1].val = 
			  thin_layer[k].erth[j1].c1 * thin_layer[k].erth[j1].val -
			    thin_layer[k].erth[j1].c2 * ( H(i,j).val - H(i,j-1).val ) ;

		}
	}
}

/* Writes scans to a binary file. */

static void write_scans( GRID *grid, SCAN scan, SCAN_LOC *scan_loc, NORM norm, 
		 FIELD *ez, FIELD *er, FIELD *h, FILE *fp_scan )

{
	long int i, j, k ; 

	float f_temp ;

	for ( k = 0 ; k < scan.n_scan ; ++ k ) {

		i = scan_loc[k].i_index ;
		j = scan_loc[k].j_index ;

		if ( strcmp( scan_loc[k].field_comp, "ez" ) == 0 )
		  f_temp = norm.e * EZ(i,j).val ;
		
		else if ( strcmp( scan_loc[k].field_comp, "er" ) == 0 )
		  f_temp = norm.e * ER(i,j).val ;

		else if ( strcmp( scan_loc[k].field_comp, "h" ) == 0 )
		  f_temp = norm.h * H(i,j).val ;

		else
		  fprintf( stderr, "WARNING: Field component for scan %-4ld is unknown.\n", k+1 ) ;
				
		file_write_float( fp_scan, &f_temp, 1L ) ;

	}

	fflush( fp_scan ) ; 
	
}

/* Writes snapshots of the electromagnetic field
to a binary file. */

static void write_snaps( GRID *grid, SNAP snap, NORM norm, FIELD *ez, FIELD *er, 
		 FIELD *h, FILE *fp_snap )

{
	long int i, j, i1, j1 ; 

	float f_temp ;

	if ( strcmp( snap.field_comp, "ez" ) == 0 ||
	    strcmp( snap.field_comp, "ezer" ) == 0 ||
	    strcmp( snap.field_comp, "ezerh" ) == 0 ) {
		/* snapshot for Ez */
		for ( j1 = 0 ; j1 < snap.n_j ; ++j1 ) {
			j = snap.j_index[j1] ;
			for ( i1 = 0 ; i1 < snap.n_i ; ++i1 ) {
				i = snap.i_index[i1] ;
				f_temp = norm.e * EZ(i,j).val ;
				file_write_float( fp_snap, &f_temp, 1L ) ;
			}
		}
	}
	if ( strcmp( snap.field_comp, "er" ) == 0 ||
	    strcmp( snap.field_comp, "ezer" ) == 0 ||
	    strcmp( snap.field_comp, "ezerh" ) == 0 ) {
		/* snapshot for Er */
		for ( j1 = 0 ; j1 < snap.n_j ; ++j1 ) {
			j = snap.j_index[j1] ;
			for ( i1 = 0 ; i1 < snap.n_i ; ++i1 ) {
				i = snap.i_index[i1] ;
				f_temp = norm.e * ER(i,j).val ;
				file_write_float( fp_snap, &f_temp, 1L ) ;
			}
		}
	}
	if ( strcmp( snap.field_comp, "e_mag" ) == 0 ) {
		/* snapshot for |E| */
		for ( j1 = 0 ; j1 < snap.n_j ; ++j1 ) {
			j = snap.j_index[j1] ;
			for ( i1 = 0 ; i1 < snap.n_i ; ++i1 ) {
				i = snap.i_index[i1] ;
				f_temp = norm.e * 
				  sqrt( ER(i,j).val * ER(i,j).val + 
				       EZ(i,j).val * EZ(i,j).val ) ;
				file_write_float( fp_snap, &f_temp, 1L ) ;
			}
		}
	}
	if ( strcmp( snap.field_comp, "h" ) == 0 ) {
		/* snapshot for H */
		for ( j1 = 0 ; j1 < snap.n_j ; ++j1 ) {
			j = snap.j_index[j1] ;
			for ( i1 = 0 ; i1 < snap.n_i ; ++i1 ) {
				i = snap.i_index[i1] ;
				f_temp = norm.e * H(i,j).val ;
				file_write_float( fp_snap, &f_temp, 1L ) ;
			}
		}
	}
			
	fflush( fp_snap ) ; 

}

